/* global controlTypes, instrumentTypes, timeAdjustObjectValue, time, netSetLocalValue, WebViewUI, log, localStorage, zoomIn:true, showZoomIcon, postponeResize, paused:true, getInstrumentValue, setInstrumentValue, graphics, ImageLoader, config, coordScale */
/* exported DP_DAWN, FLAPS_Steps, imagePostfix, imagePostfix, imagesFolder, altImagePrefix, netInterval, stillInstruments, dontExtrapolate, extrapolateCustomThresholds, localValues, controls, onConnectionEstablished, dontSend, viewLimits, onReceiveData */
/* jshint proto: true */
/* jshint curly: false */


var ROW1_Y = 254,
	ROW2_Y = 442,
	ROW3_Y = 647,
	COL1_X = 368,
	COL2_X = 555,
	COL3_X = 743,
	COL4_X = 931,
	AUX_COL1_X = 67,
	AUX_COL2_X = 198,
	AUX_ROW1_Y = 265,
	AUX_ROW2_Y = 417,
	AUX_ROW3_Y = 577,
	KNOB_OFFSET_X = 69,
	KNOB_OFFSET_Y = 72,
	ICON1_X = 985,
	ICON_Y = 134,
	ICON_OFFSET = -42;

var DP_DAWN = 0, DP_DAY = 1, DP_DUSK = 2, DP_NIGHT = 3,	
	CLOCK_UT = 0, CLOCK_LT=1, CLOCK_FT=2, CLOCK_ET = 3,
	FLAPS_LeverTravel = 50, FLAPS_Steps = 4,
	AngleRef = config.useDegrees?180:Math.PI;
	
//logical view size limits, multiply by coordScale to get the physical limits (in pixels)
var viewLimits = {  		
	minWidth : 980, minHeight : 640,//Many devices are 1080×1920, Asus Transformer, KindleFireHD & Nexus7 are 1200×1920 and our minimal retina size is 1280×1960. 1080×1920 is popular too
	maxWidth : 1280, maxHeight : 840,//the actual max height is maxHeight
	normalWidth : 1024, normalHeight : 768,
	blackPixelsOnTop : true,
	//sideBlackStripes : 0.04,//to avoid switching to hiRes texture or scaling up the canvas, add black stripes on the sides, but only if they are a few percent of the width 
	addTop : 12,//pixels added on top of normalHeight for maxHeight
	trimTop : -108,//top pixels trimmed from normalHeight for minHeight
	zoomInThreshold : 1.1,//zoom in only if the zoom ratio is greater than this
	zoomOutThreshold : 1.15,//zoom out only if the zoom ratio is greater than this, zooming out is less useful than zooming in => use greater threshold
	hiRezCoordScale : 2,
};

var hiRezPostfix = "@2x.png", loRezPostfix = ".png",
	hiRez = false,
	imagePostfix = hiRez?hiRezPostfix:loRezPostfix,
	imagesFolder = "c172/images",//coordScale==2?"images@2x":"images",
	altImagePrefix = "night_",
	netInterval = 100,//100 ms between updates, the init packet may override it - data.netInterval
	gpsPanel = true,
	radioPanel =true,
	lever;
	
var stillInstruments = [//instruments that move rarely or slowly. They can be drawn in an offscreen canvas which is updated only when a value changes above certain threshold
						"airspeed_calibration_angle","ai_bars_position","kohlsman_setting","plane_heading_bug","gyro_drift_angle","plane_heading_bug","adf1_card","nav1_obs","nav2_obs",
						"fuel_qty_left_tank","fuel_qty_right_tank","engine_egt","engine_fuel_flow","engine_oil_temp","engine_oil_pressure","vacuum","amp"],
	dontExtrapolate = {//instruments which values are not extrapolated, the stillInstruments are added to it
						zulu_time:true, local_time:true, time_of_day:true, outside_air_temp:true, bus_voltage:true, clock_mode:true, clock_voltage_oat_mode:true, 
						ai_fail:true, airspeed_fail:true, altitude_fail:true, vertical_fail:true, gyro_fail:true, nav1_available: true, nav_fail:true, 
						ai_bars_position: true, gyro_drift_angle: true,
						nav2_available:true, adf_fail:true, /*engine_elapsed_time:true,*/ engine_instruments_fail:true, fuel_indicator_fail:true, vacuum_fail:true, electrical_fail:true,
						low_fuel_l:true, low_fuel_r:true, low_oil_pressure:true, low_vac_l:true, low_vac_r:true, low_voltage:true, flaps_handle_ratio:true},
						//whiskey_compass:true, plane_heading_magnetic:true},//caged?
	occasionalProperties = {},
	netEventNames = {},
	extrapolateCustomThresholds = { indicated_altitude:2000 },
	localValues = {egt_ref:1100};
				  
var knobControl = {//a prototype for all knob controls
	__proto__: controlTypes.drag,
	x: 0, y: 0,
	w:60, h:60,
	onDrag : function( x,y, dx,dy, value, min,max ){
		var delta = (dx+dy)*(max-min)/300;//from max to min in 300 pixels
		if( config.useDegrees )
			delta = delta*180/Math.PI;
		if( this.knobSpeed )
			delta *= this.knobSpeed;
		return value + (this.invert?-delta:delta);
	}
};

/*function onReceiveData(data){
	data._gyro_heading_bug = (data.plane_heading_magnetic+data.gyro_drift_angle-data.plane_heading_bug);
	data._gyro_heading = (data.plane_heading_magnetic+data.gyro_drift_angle);
	console.log(data.plane_heading_magnetic);
}*/

function onConnectionEstablished(){
	gpsPanel = radioPanel = true;
	//netSetLocalValue( radioPanel, "radio_panel" );
	//netSetLocalValue( gpsPanel,  "gps_panel" );
}

function onReceiveData(data){
	//the tacho extrapolation is different from the arrow, it uses cycle => copy it
	data.engine_rpm_tacho = data.engine_rpm;

	//remove eng1_ before all engine values, thus keeeping compatibility with the old server
	for( var p in data ){
		if( p.indexOf("eng1_")==0 ){
			var value = data[p];
			delete data[p];
			p = p.substring(5);
			data[p] = value;
		}
	}
}

var instruments = [
	{//Day/Night
		__proto__ : instrumentTypes.base,
		draw: function(){
			ImageLoader.showAltImages( instruments, this.value==DP_NIGHT || this.value==DP_DUSK );
		},
		value: ImageLoader.usingAltImages()?DP_NIGHT:DP_DAY,
		listen:"time_of_day",
		interpolateSpeed:0
	},
	{//ASI_Tas
		__proto__ : instrumentTypes.rotate,
		x: COL1_X,
		y: ROW1_Y,
		listen:"airspeed_calibration_angle",
		interpolateSpeed: 0,
		setValue: function(val){
			if( config.useDegrees )
				this.value = -(val-4)*1.9;
			else{
				val = -(val-0.07)*2;//in radians
				this.value = val*360/(2*Math.PI);//degrees
			}
		},
		src:"asi/asi_tas"
	},
	{//ASI_Temp
		//@todo - the font sucks
		//@todo - the iOS version is not centered, what is the exact formula? I have mapped the min/max of TAS on iOS to -30:+30 on the TEMP scale while retaining the TAS formula
		__proto__ : instrumentTypes.slideTexture,
		x: COL1_X+2,
		y: ROW1_Y-50,
		horizontal: true,
		listen: "airspeed_calibration_angle",
		//inputValueRange: [-AngleRef/6.66, AngleRef/6.66],//+/- 0.139 : make the min/max input range map to -30,+30 on the scales
		inputValueRange: [-30,30],//+/- 0.139 : make the min/max input range map to -30,+30 on the scales
		setValue: function(val){
			//val = 0.000019706884723346*val*val*val +     0.000088680981255058*val*val +     0.998162332999548*val +     1.4971104780274394;
			val+=1.5;
			if( val<-26 )
				val = -26;
			if( val>29 )
				val = 29;
			val =  ( this.valueOffset ? this.valueOffset : 0 ) + ((this.valueScale!==undefined ) ? this.valueScale*val:val );
			this.value = val;
		},
		peekSize:31,
		src:"asi/h5_asi_temp",
		control: {
			__proto__: knobControl,
			x: -2-KNOB_OFFSET_X,
			y: 50+KNOB_OFFSET_Y,
			knobSpeed: 0.1
		}		
	},
	{//ASI_Dial
		__proto__ : instrumentTypes.static,
		x: COL1_X,
		y: ROW1_Y,
		src:"asi/asi_dial",
	},
	{//ASI_Arrow
		__proto__ : instrumentTypes.rotate,
		x: COL1_X,
		y: ROW1_Y,
		src:"asi/asi_arrow",
		listen:"indicated_airspeed",
		inputValueRange: [0,210],
		setValue: function(val){
			if( val<40 )
				val = val*3/4;//in the 0->30 range
			else{
				if( val>200 )
					val = 317.5 + (val-200)*1.1;//extrapolate from 200 and onwards
				else{
					//see the 'polynominal' folder for a webapp that calculates the coefficients
					//the original is here, I've modified it cause it rounded the results too much: http://www.had2know.com/academics/cubic-through-4-points.html
					//200 -> 317.5
					//120 -> 207                           
					//80  -> 112.5
					//40  -> 30
					val = -0.0000745443*val*val*val + 0.021640625*val*val + 0.3005208333*val - 11.875;
				}
			}
			this.value = val;
		}
	},
	{//AI_bg
		__proto__ : instrumentTypes.rotate,
		x: COL2_X,
		y: ROW1_Y,
		inputValueRange: [ -AngleRef, AngleRef ],
		repeat: true,
		listen: "ai_bank",
		src:"ai/ai_bg"
	},
	{//AI_dial
		__proto__ : instrumentTypes.rotate,
		x: COL2_X,
		y: ROW1_Y,
		listen: ["ai_bank","ai_pitch"],
		defaultValue: "ai_bank",//the property for rotate, as we have multiple values, so this.value is not used 
		ai_bank:{ inputValueRange: [ -AngleRef, AngleRef ], repeat: true },
		ai_pitch:{ valueScale:95*Math.PI/AngleRef/*190 - like in MSFS, 120 - like in the iOS version*/, inputValueRange: [ -AngleRef/8, AngleRef/8 ] },//inputValueRange is not used to determine scale/offset, we don't know the property type to do it(slide,rotate,etc)
		/*setValue: function(val,key){				
			return instrumentTypes.base.setValue.call(this,val,key);
		},*/
		drawImage: function( image,x,y ){//move the image up/down after rotating
			var pitch =  timeAdjustObjectValue( this.ai_pitch, this.ai_pitch.value )*coordScale;//@todo - use getDisplayedValue!
			graphics.drawImage( image, x, y-pitch ); 
		},
		src:"ai/ai_dial"
	},
	{//AI_shadow
		__proto__ : instrumentTypes.static,
		x: COL2_X,
		y: ROW1_Y,
		src:"ai/ai_bank_ring_shadow"
	},
	{//AI_bars
		__proto__ : instrumentTypes.slide,
		x: COL2_X,
		y: ROW1_Y,
		inputValueRange:[-1,1],
		valueScale:-8,
		listen:"ai_bars_position",
		control:{
			__proto__: knobControl,
			x: 0, y: KNOB_OFFSET_Y,
			knobSpeed: 0.1,
			sendDelta: true,
			intValue: true,
			deltaScale: -20,
			invert:true
		},
		src:"ai/ai_bars",
	},
	{//AI_bank_ring
		__proto__ : instrumentTypes.rotate,
		x: COL2_X,
		y: ROW1_Y,
		inputValueRange: [ -AngleRef, AngleRef ],
		repeat: true,
		listen: "ai_bank",
		src:"ai/ai_bank_ring",	
	},
	{//AI_fail_flag
		//@todo - casts shadow when hiiden
		__proto__ : instrumentTypes.rotate,
		x: COL2_X-55,
		y: ROW1_Y-58,
		inputValueRange:[0,1],
		outputValueRange:[180/5,0],
		listen:"ai_fail",
		src:"ai/ai_fail_flag"
	},
	{//ALT_pressure
		__proto__ : instrumentTypes.slideTexture,
		x: COL3_X+55,
		y: ROW1_Y+1,
		inputValueRange:[ 28.0, 31.5 ],
		peekSize:35,
		listen: "kohlsman_setting",
		control: {
			__proto__: knobControl,
			knobSpeed:0.03,
			invert:true,
			x: -55-KNOB_OFFSET_X,
			y: -1+KNOB_OFFSET_Y
		},		
		src:"altimeter/h5_altimeter_pressure"
	},
	{//ALT_bg
		__proto__ : instrumentTypes.static,
		x: COL3_X,
		y: ROW1_Y,
		src:"altimeter/altimeter_bg"
	},
	{//ALT_x10000
		__proto__ : instrumentTypes.rotate,
		x: COL3_X,
		y: ROW1_Y,
		repeat: true,
		nocycle: true,//the values are not cyclic
		inputValueRange : [ 0,10000*10 ],
		listen: "indicated_altitude",
		src:"altimeter/altimeter_x10000_arrow"
	},
	{//ALT_x1000
		__proto__ : instrumentTypes.rotate,
		x: COL3_X,
		y: ROW1_Y,
		repeat: true,
		nocycle: true,//the values are not cyclic
		inputValueRange : [ 0, 1000*10 ],
		listen: "indicated_altitude",
		src:"altimeter/altimeter_x1000_arrow"
	},
	{//ALT_x100
		__proto__ : instrumentTypes.rotate,
		x: COL3_X,
		y: ROW1_Y,
		repeat: true,
		nocycle: true,//the values are not cyclic
		inputValueRange : [ 0, 100*10 ],
		listen: "indicated_altitude",
		src:"altimeter/altimeter_x100_arrow"
	},
	{//TURN_SLIP_back
		__proto__ : instrumentTypes.static,
		x: COL1_X,
		y: ROW2_Y,
		src:"turn_slip/turn_slip_back"
	},
	{//TURN_SLIP_back
		__proto__ : instrumentTypes.static,
		x: COL1_X,
		y: ROW2_Y,
		src:"turn_slip/turn_slip_back"
	},
	{//TURN_SLIP_ball
		__proto__ : instrumentTypes.slide,
		x: COL1_X,
		y: ROW2_Y,
		horizontal: true,
		inputValueRange:[-127,127 ], 
		outputValueRange:[-127*0.27,127*0.27],
		listen:"turn_coordinator_ball",
		getY: function(){
			return this.y - Math.abs(this.getDisplayedValue(this.value))*0.06;
		},
		src:"turn_slip/turn_slip_ball"
	},
	{//TURN_SLIP_front
		//@todo - trim
		__proto__ : instrumentTypes.static,
		x: COL1_X,
		y: ROW2_Y,
		src:"turn_slip/turn_slip_front"
	},
	{//TURN_SLIP_arrow (plane shape)
		__proto__ : instrumentTypes.rotate,
		x: COL1_X,
		y: ROW2_Y,
		inputValueRange:[-AngleRef/4,AngleRef/4 ], 
		outputValueRange:[-1.75*180,1.75*180],
		listen:"turn_coordinator_bank",
		src:"turn_slip/turn_slip_arrow"
	},
	{//DIR_GYRO_dial
		__proto__ : instrumentTypes.rotate,
		x: COL2_X,
		y: ROW2_Y,
		repeat: true,
		inputValueRange:[0,2*AngleRef],
		plane_heading_magnetic:0, gyro_drift_angle:0,
		getDisplayedValue : function(){
			return this.timeAdjustValue( -(this.plane_heading_magnetic+this.gyro_drift_angle), 360 );  
		},
		listen:["gyro_drift_angle","plane_heading_magnetic"],	
		src:"dir_gyro/dir_gyro_dial"
	},
	{//GYRO_reset
		__proto__ : controlTypes.base,
		x: COL2_X, y: ROW2_Y, w: 183,h: 183,  
		onClick : function(){
			setInstrumentValue(true,'gyro_drift_reset');
			/*var counter = 0;
			sendIncrement();
			function sendIncrement(){
				counter++;
				setInstrumentValue(counter,'gyro_drift_reset');
				if( counter<90 )
					setTimeout( sendIncrement,10);
			}
			//setInstrumentValue(true,'gyro_drift_reset');*/
		}
	},
	{//DIR_GYRO_glass
		__proto__ : instrumentTypes.static,
		x: COL2_X,
		y: ROW2_Y,
		src:"dir_gyro/dir_gyro_glass"
	},
	{//DIR_GYRO_heading_bug  
		//@todo - test gyro_drift_angle                  
		__proto__ : instrumentTypes.rotate,
		x: COL2_X,
		y: ROW2_Y,
		repeat: true,
		inputValueRange:[0,2*AngleRef],
		plane_heading_magnetic:0, gyro_drift_angle:0, plane_heading_bug:0,//gets converted to 0..360 in setValue
		getDisplayedValue : function(){
			return this.timeAdjustValue( -(this.plane_heading_magnetic+this.gyro_drift_angle-this.plane_heading_bug), 360 );  
		},
		listen:["plane_heading_bug","plane_heading_magnetic","gyro_drift_angle"],
		control: { __proto__: knobControl, x: KNOB_OFFSET_X },//controls the plane_heading_bug
		src:"dir_gyro/dir_gyro_heading_bug"
	},
	{//VSI_bg                     
		__proto__ : instrumentTypes.static,
		x: COL3_X,
		y: ROW2_Y,
		src:"vsi/vsi_bg"
	},
	{//VSI_arrow                     
		__proto__ : instrumentTypes.rotate,
		x: COL3_X,
		y: ROW2_Y,
		inputValueRange:[2050,-2050 ], //2000
		outputValueRange:[177.3,-177.3],//173  
		listen:"vertical_speed",
		setValue: function(val){
			var sign = val>0;//we'll invert it in the end
			val = Math.abs(val);
			if( val<1500 ){
				//see the 'polynominal' folder for a webapp that calculates the coefficients
				//0 -> 0
				//500 -> 405
				//1000  -> 923 
				//1500 -> 1500
				val = -7.2e-8*val*val*val + 0.000334*val*val + 0.661*val + 0;
			}
			//else - linear above 1500
			if( !sign )
				val = -val;
			this.value = this.convertValue(val);
		},
		src:"vsi/vsi_arrow"
	},
	
	{//ADF_bg             
		//@todo - black rect        
		__proto__ : instrumentTypes.static,
		x: COL4_X,
		y: ROW3_Y,
		src:"adf/adf_bg"
	},
	
	{//ADF_arrow                     
		__proto__ : instrumentTypes.rotate,
		x: COL4_X,
		y: ROW3_Y,
		repeat:true,
		inputValueRange : [0,2*AngleRef],
		listen:"adf1_radial",
		src:"adf/adf_arrow"
	} ,
	{//ADF_dial_shadow                     
		__proto__ : instrumentTypes.static,
		x: COL4_X,
		y: ROW3_Y,
		src:"adf/adf_dial_shadow"
	},
	{//ADF_dial                     
		__proto__ : instrumentTypes.rotate,
		x: COL4_X,
		y: ROW3_Y,
		repeat:true,
		inputValueRange : [0,-2*AngleRef],
		listen:"adf1_card",
		src:"adf/adf_dial"
	},
	{//ADF_glass
		//@todo - merge with dial shadow?                     
		__proto__ : instrumentTypes.static,
		x: COL4_X,
		y: ROW3_Y,
		src:"adf/adf_glass"
	},
	{//TACHOMETER_hours_bg                     
		__proto__ : instrumentTypes.static,
		x: COL3_X,
		y: ROW3_Y,
		src:"tachometer/tachometer_hours_bg"
	},
	{//TACHOMETER numerator group
		__proto__ : instrumentTypes.group,
		x: COL3_X,
		y: ROW3_Y,
		listen:"engine_elapsed_time",
		setChildValues: true,
		instruments: [
			{//TACHOMETER_numbers
				//@done - cropped
				//@todo - the digits texture is not pixel-correct                  
				__proto__ : instrumentTypes.rollCounter,
				x:-4,
				y:31,
				leading:11,
				numDigits: 5,
				height:14,
				fract:true,
				interpolateSpeed: 0,		
				src: ["tachometer/h5_tachometer_numbers_white", "tachometer/h5_tachometer_numbers_black"]
			},
			{//TACHOMETER_symbol
				__proto__ : instrumentTypes.slideTexture,
				x:28,
				y:31,
				peekSize: 14,
				repeat: true,
				listen:"engine_rpm_tacho",
				totalTime: 0,
				inputValueRange:[0,1],
				getDisplayedValue:function(){
					if( this.prevTime!==undefined && this.value>100/*engine_rpm*/){
						var deltaTime =  time-this.prevTime;
						if( deltaTime<500 )//e.g. has not been paused
							this.totalTime += deltaTime;
					}
					this.prevTime = time;
					return -this.image.height*this.totalTime/15000;//1 symbol per 15 seconds
				},
				src: "tachometer/h5_tachometer_symbol"
			}
		]
	},
	{//TACHOMETER_dial                     
		__proto__ : instrumentTypes.static,
		x: COL3_X,
		y: ROW3_Y,
		src:"tachometer/tachometer_dial"
	},
	//@todo - digits
	{//TACHOMETER_arrow                     
		__proto__ : instrumentTypes.rotate,
		x: COL3_X,
		y: ROW3_Y,
		inputValueRange:[0, 3500 ],
		outputValueRange:[-36, 216 ],
		listen:"engine_rpm",
		src:"tachometer/tachometer_arrow"
	},
	{//COMPASS_rose
		//@todo - the scale ticks are blurred and tilted to the right
		//@done - cropped
		__proto__ : instrumentTypes.slideTexture,
		x: COL2_X,
		y: ROW3_Y,
		listen: "whiskey_compass",
		horizontal:true,
		repeat:true,
		inputValueRange:[0,2*AngleRef],
		//inputValueRange:[-AngleRef,AngleRef],
		peekSize:73,
		src:"compass/h5_compass_rose",
	},
	{//COMPASS_bg
		//@todo - why do we need compass_bg
		__proto__ : instrumentTypes.static,
		x: COL2_X,
		y: ROW3_Y,
		src:"compass/compass_bg"
	},
	{//VOR_gs_stripe         
		__proto__ : instrumentTypes.slideTexture,
		x: COL4_X+30,
		y: ROW1_Y-7,
		textureNotCentered: true,
		interpolateSpeed: 100,
		peekSize:6,      
		edgeRange: true,		
		inputValueRange: [1,0],
		listen:"nav1_gs_flag",
		src:"vor/vor_gs_stripe",//@todo - not antialiased
	},
	{//VOR_nav_stripe         
		__proto__ : instrumentTypes.slideTexture,
		x: COL4_X,
		y: ROW1_Y+35.5,
		//textureNotCentered: true,
		interpolateSpeed: 100,
		peekSize:12,//13,
		edgeRange: true,		
		listen:"nav1_to_from",
		src:"vor/h5_vor_nav_stripe",
		inputValueRange: [2,0],
		setValue : function(val){
			if( val>0 )
				val = val==1?2:1;
			instrumentTypes.slideTexture.setValue.call(this,val);				
		}
	},
	{//VOR_bg                     
		__proto__ : instrumentTypes.static,
		x: COL4_X,
		y: ROW1_Y,
		src:"vor/vor_bg"
	},
	{//VOR_cdi_arrow                     
		__proto__ : instrumentTypes.rotate,
		x: COL4_X,
		y: ROW1_Y-48,
		inputValueRange: [-127,127],
		outputValueRange: [127*0.005*360/(2*Math.PI),-127*0.005*360/(2*Math.PI)],//on the iOS version the constant is 0.004 - adjust to 0.005 to make 127 hit the end of the scale, I think the artwork is to blame
		listen:"nav1_cdi",
		src:"vor/vor_cdi_arrow"
	},
	{//VOR_gs_arrow                     
		__proto__ : instrumentTypes.rotate,
		x: COL4_X-57,
		y: ROW1_Y,
		inputValueRange: [-119,119],
		outputValueRange: [-119*0.004*360/(2*Math.PI),119*0.004*360/(2*Math.PI)],
		listen:"nav1_gsi",
		src:"vor/vor_gs_arrow"
	},
	{//VOR_dial_shadow                     
		__proto__ : instrumentTypes.static,
		x: COL4_X,
		y: ROW1_Y,
		src:"vor/vor_dial_shadow"
	},
	{//VOR_dial                     
		__proto__ : instrumentTypes.rotate,
		x: COL4_X,
		y: ROW1_Y,
		repeat:true,
		inputValueRange : [0,-2*AngleRef],
		listen:"nav1_obs",
		src:"vor/vor_dial"
	},
	{//VOR_glass
		//@todo - merge with fron panel?                     
		__proto__ : instrumentTypes.static,
		x: COL4_X,
		y: ROW1_Y,
		src:"vor/vor_glass"
	},
	
	{//VOR2_nav_stripe         
		__proto__ : instrumentTypes.slideTexture,
		x: COL4_X+40,
		y: ROW2_Y-2, 
		textureNotCentered: true,
		interpolateSpeed: 100,
		peekSize:12,//13  						
		listen:"nav2_to_from",
		src:"vor2/h5_vor2_nav_stripe",
		edgeRange: true,
		inputValueRange: [2,0],
		setValue : function(val){
			if( val>0 )
				val = val==1?2:1;
			instrumentTypes.slideTexture.setValue.call(this,val);				
		}
	},
	{//VOR2_bg                     
		__proto__ : instrumentTypes.static,
		x: COL4_X,
		y: ROW2_Y,
		src:"vor2/vor2_bg"
	},
	{//VOR2_cdi_arrow                     
		__proto__ : instrumentTypes.rotate,
		x: COL4_X,
		y: ROW2_Y-48,
		inputValueRange: [-127,127],
		outputValueRange: [127*0.004*360/(2*Math.PI),-127*0.004*360/(2*Math.PI)],
		listen:"nav2_cdi",
		src:"vor2/vor2_cdi_arrow"
	},
	{//VOR2_dial_shadow                     
		__proto__ : instrumentTypes.static,
		x: COL4_X,
		y: ROW2_Y,
		
		src:"vor2/vor2_dial_shadow"
	},
	{//VOR2_dial                     
		__proto__ : instrumentTypes.rotate,
		x: COL4_X,
		y: ROW2_Y,
		repeat:true,
		inputValueRange : [0,-2*AngleRef],
		listen:"nav2_obs",
		src:"vor2/vor2_dial"
	},
	{//VOR2_glass
		//@todo - merge with front panel?                     
		__proto__ : instrumentTypes.static,
		x: COL4_X,
		y: ROW2_Y,
		src:"vor2/vor2_glass"
	},
	{//the canvas that contains infrequently updated instruments
		__proto__ : instrumentTypes.canvas,
		x:0,y:0,
		instruments:[
		{//FUEL_QTY_bg
			__proto__ : instrumentTypes.static,
			x: AUX_COL1_X,
			y: AUX_ROW2_Y,
			src:"fuel_qty/fuel_qty_bg"		
		},
		{//FUEL_QTY_arrow_left
			__proto__ : instrumentTypes.rotate,
			x: AUX_COL1_X-38,
			y: AUX_ROW2_Y,
			debugDumpCount : 0, prev : 0, pcount : 0, stopnext:false,
			getRotate: function(){
				if( this.debugDumpCount<300 && (this.debugDumpCount || this.value) ){
					if( this.stopnext )
						this.stopnext = true;
					var dval = this.getDisplayedValue( this.getValue()/*valueObject.value*/, this.getValueProperty('repeat')?360:undefined );
					if( this.prev && Math.abs(this.prev-dval)<1 && dval < this.prev ){
						if( this.pcount==10 )
							this.stopnext = true;
						this.getDisplayedValue( this.getValue()/*valueObject.value*/, this.getValueProperty('repeat')?360:undefined );
						this.pcount++;
					} 
					//log( 'disp:fqtyleft '+ dval );
					this.debugDumpCount++;
					this.prev = dval;
				}
				return instrumentTypes.rotate.getRotate.call(this); 
			},
			inputValueRange:[0, 26 ],
			outputValueRange:[54,-54],
			listen:"fuel_qty_left_tank",
			src:"fuel_qty/fuel_qty_arrow_left",
		},
		{//FUEL_QTY_arrow_right
			__proto__ : instrumentTypes.rotate,
			x: AUX_COL1_X+38,
			y: AUX_ROW2_Y,		
			inputValueRange:[ 0, 26 ],
			outputValueRange:[-54,54],
			listen:"fuel_qty_right_tank",		
			src:"fuel_qty/fuel_qty_arrow_right"		
		},
		{//EGT_FUEL_FLOW_bg
			__proto__ : instrumentTypes.static,
			x: AUX_COL2_X,
			y: AUX_ROW2_Y,
			src:"egt_fuel_flow/egt_fuel_flow_bg"		
		},
		{//EGT_FUEL_FLOW_egt_arrow
			__proto__ : instrumentTypes.rotate,
			x: AUX_COL2_X-38,
			y: AUX_ROW2_Y,
			inputValueRange:[600, 1800 ], 
			outputValueRange:[54,-54],  
			listen:"engine_egt",
			src:"egt_fuel_flow/egt_arrow",
		},
		{//EGT_FUEL_FLOW_egt_reference_arrow
			__proto__ : instrumentTypes.rotate,
			x: AUX_COL2_X-38,
			y: AUX_ROW2_Y,
			inputValueRange:[600, 1800 ],
			outputValueRange:[54,-54],
			listen:"egt_ref",
			src:"egt_fuel_flow/egt_reference_arrow"
		},
		{//EGT_FUEL_FLOW_fuel_flow_arrow
			__proto__ : instrumentTypes.rotate,
			x: AUX_COL2_X+38,
			y: AUX_ROW2_Y,
			inputValueRange:[0,19],//calcSmallDialAngle:fuelFlow-2 min:C172_FUEL_FLOW_MIN(0 -> -2) max:C172_FUEL_FLOW_MAX(21 -> 19)]];//@check is that correct 
			outputValueRange:[-57,67],
			setValue: function(val){
				//map the input range 75->100 to 75->92 and 100->245 to 92->245
				//map the input range 0->5 to 0->1.5 and 5->19 to 1.5->19
				if( val<5 )
					val = 0 + (val-0)*(1.5-0)/(5-0);
				else{
					//5  -> 5
					//10 -> 7.9
					//15 -> 12.9
					//19 -> 19
					//see the 'polynominal' folder for a webapp that calculates the coefficients
					//the original is here, I've modified it cause it rounded the results too much: http://www.had2know.com/academics/cubic-through-4-points.html
					val = 0.001611111111111*val*val*val -0.010333333333333*val*val +  0.473055555555556*val + 2.691666666666667;
					val = 19 - (19-val)*(19-1.5)/(19-5);
				}
				this.value = this.convertValue(val); 
			},
			listen:"engine_fuel_flow",
			src:"egt_fuel_flow/fuel_flow_arrow"
		},
		{//OIL_bg
			__proto__ : instrumentTypes.static,
			x: AUX_COL1_X,
			y: AUX_ROW3_Y,
			src:"oil/oil_bg"		
		},
		{//OIL_temp_arrow
			__proto__ : instrumentTypes.rotate,
			x: AUX_COL1_X-38,
			y: AUX_ROW3_Y,
			inputValueRange:[75, 245 ],
			outputValueRange:[59,-59],
			setValue: function(val){
				//map the input range 75->100 to 75->92 and 100->245 to 92->245
				if( val<100 )
					val = 75 + (val-75)*(92-75)/(100-75);
				else
					val = 245 - (245-val)*(245-92)/(245-100);
				this.value = this.convertValue(val);
			},
			listen:"engine_oil_temp",
			src:"oil/oil_temp_arrow"
		},
		{///OIL_press_arrow
			__proto__ : instrumentTypes.rotate,
			x: AUX_COL1_X+38,
			y: AUX_ROW3_Y,
			inputValueRange:[115, 0 ],
			outputValueRange:[60,-65],
			listen:"engine_oil_pressure",
			src:"oil/oil_press_arrow"		
		},
		{//VAC_AMP_bg
			__proto__ : instrumentTypes.static,
			x: AUX_COL2_X,
			y: AUX_ROW3_Y,
			src:"vac_amp/vac_amp_bg"		
		},
		{//VAC_AMP_vac_arrow
			__proto__ : instrumentTypes.rotate,
			x: AUX_COL2_X-38,
			y: AUX_ROW3_Y,
			inputValueRange:[3, 7 ],
			outputValueRange:[64,-64],
			listen:"vacuum",
			src:"vac_amp/vac_amp_vac_arrow",
		},
		{///VAC_AMP_amp_arrow
			__proto__ : instrumentTypes.rotate,
			x: AUX_COL2_X+38,
			y: AUX_ROW3_Y,  
			inputValueRange:[60, -60 ],
			outputValueRange:[65,-65],
			listen:"amp",
			src:"vac_amp/vac_amp_amp_arrow"						
		},			
		//----------------------------------------------------------------------------------
		{//frontPanel
			//@todo - do not turn alternator off - a bit higher so it fits in the smallest height
			__proto__ : instrumentTypes.static,
			left:-128,
			top:-12,
			src:"h5_front_panel_l"
		},
		{//frontPanel
			//@todo - do not turn alternator off - a bit higher so it fits in the smallest height
			__proto__ : instrumentTypes.static,
			left:128,
			top:-12,
			src:"h5_front_panel_r"
		},
		//----------------------------------------------------------------------------------
		/*{//frontPanel
			__proto__ : instrumentTypes.static,
			top:0,
			left:0,
			src:"front_panel"
		},*/
		//----------------------------------------------------------------------------------
		{//show radio panel
			__proto__ : instrumentTypes.static,
			x: ICON1_X,
			y: ICON_Y+2,
			control:{
				__proto__ : controlTypes.base,
				x: 0, y: 0,
				w:50, h:50,
				onClick : function(){
					//console.log('radio onclick');
					radioPanel = !radioPanel;
					netSetLocalValue( radioPanel, "radio_panel" );
				}
			},
			isHidden: function(){
				return WebViewUI.isShown();
			},
			src:"radio_panel",                             
			noAltImage: true
		},
		{//show gps panel
			__proto__ : instrumentTypes.static,
			x: ICON1_X + ICON_OFFSET +6,
			y: ICON_Y,
			control:{
				__proto__ : controlTypes.base,
				x: 0, y: 0,
				w:50, h:50,
				onClick : function(){
					console.log('gps onclick');
					gpsPanel = !gpsPanel;
					netSetLocalValue( gpsPanel,  "gps_panel" );
				}
			},
			isHidden: function(){
				return WebViewUI.isShown();
			},
			src:"gps_panel",                             
			noAltImage: true
		},	
		{//show connect UI
			__proto__ : instrumentTypes.static,
			x: ICON1_X + ICON_OFFSET*2,
			y: ICON_Y,
			control:{
				__proto__ : controlTypes.base,
				x: 0, y: 0,
				w:50, h:50,
				onClick : function(){
					WebViewUI.show( true );
				}
			},
			isHidden: function(){
				return WebViewUI.isShown();
			},
			src:"server_panel",                             
			noAltImage: true
		},
		{//zoom_in/out
			__proto__ : instrumentTypes.switchImages,
			x: ICON1_X + ICON_OFFSET*3,
			y: ICON_Y,
			draw : function( ctx ){
				this.value = zoomIn;
			   	instrumentTypes.switchImages.draw.call(this,ctx);
			},
			control:{
				__proto__ : controlTypes.base,
				x: 0, y: 0,
				w:50, h:50,
				onClick : function(){
					if( !this.parentInstrument.isHidden() ){
						//this.parentInstrument.value = this.parentInstrument.value?0:1;
						localStorage.zoomIn = zoomIn = !zoomIn;//this.parentInstrument.value?true:false;
						log( zoomIn? "requested Zoomi-In" : "requested Zoomi-Out" );
						graphics.resize();
					}
				}
			},
			isHidden: function(){
				return WebViewUI.isShown() || !showZoomIcon || postponeResize;
			},
			value: (localStorage.getItem("zoomIn")!==null ) ? ( localStorage.zoomIn=="true" ) : false,
			src:["zoom_in","zoom_out"],
			noAltImage: true
		},
		{//annunciator
		    __proto__ : instrumentTypes.group,
		    x: 621,
		    y: 116,
		    testMode: false,
		    dimMode: false,
		    draw : function( ctx ){
		    	if( this.dimMode )
		    		graphics.setAlpha(0.7);//transparent => darker
		    	instrumentTypes.group.draw.call(this,ctx);
		    	if( this.dimMode )
		    		graphics.setAlpha(1);//as it was
			},
		    instruments:[
		    	{//SWITCH_annunciator
					__proto__ : instrumentTypes.switchImages,
					x: 178,
					y: 24,
					value: 1,
					control:{
						__proto__ : controlTypes.drag,
						x: 1, y: 0,
						w:40, h:60,
						onMouseDrag : function(x,y,dx,dy){
							var val = y-this.getAbsoluteY();
							if( Math.abs(val)<12 ){
								this.parentInstrument.value = 1;
								this.parentInstrument.parentInstrument.testMode = false;
								this.parentInstrument.parentInstrument.dimMode = false;
							}
							else
								if( val<0 ){
									this.parentInstrument.value = 2;
									this.parentInstrument.parentInstrument.testMode = true;
								}
								else{
									this.parentInstrument.parentInstrument.dimMode = true;
									this.parentInstrument.value = 0;
								}
						},
						onMouseUp : function(){
							if( this.parentInstrument.value==2 ){
								this.parentInstrument.value = 1;
								this.parentInstrument.parentInstrument.testMode = false;
							}	
						}
					},
					src:["switch/switch_down","switch/switch_mid","switch/switch_up"]
				},		    
		    	{//low_fuel_l
		    		__proto__ : instrumentTypes.static,
		    		left: 0, top:0,
		    		isHidden: function(){
						return !this.parentInstrument.testMode && this.value===0;
					},
		    		listen:"low_fuel_l",
		    		src:"annunciator/annunciator_l",
		    		noAltImage: true
				},
				{//low_fuel
		    		__proto__ : instrumentTypes.static,
		    		left: 11, top:0,
		    		low_fuel_l:0, low_fuel_r:0,
		    		isHidden: function(){return !this.parentInstrument.testMode && !this.low_fuel_l && !this.low_fuel_r;},
		    		listen:["low_fuel_l","low_fuel_r"],
		    		src:"annunciator/annunciator_low_fuel",
		    		noAltImage: true
				},
				{//low_fuel_r
		    		__proto__ : instrumentTypes.static,
		    		left: 67, top:0,
		    		isHidden: function(){return !this.parentInstrument.testMode && !this.value;},
		    		listen:"low_fuel_r",
		    		src:"annunciator/annunciator_r",
		    		noAltImage: true
				},
				{//line
		    		__proto__ : instrumentTypes.static,
		    		left: 77, top:6,
		    		isHidden: function(){return !this.parentInstrument.testMode && !this.parentInstrument.testMode;},
		    		src:"annunciator/annunciator_line",
		    		noAltImage: true
				},
				{//low_oil_pressure
		    		__proto__ : instrumentTypes.static,
		    		left: 0, top:18,
		    		isHidden: function(){return !this.parentInstrument.testMode && !this.value;},
		    		listen:"low_oil_pressure",
		    		src:"annunciator/annunciator_oil_pres",
		    		noAltImage: true
				},
				{//low_vac_l
		    		__proto__ : instrumentTypes.static,
		    		left: 60, top:18,
		    		isHidden: function(){return !this.parentInstrument.testMode && !this.value;},
		    		listen:"low_vac_l",
		    		src:"annunciator/annunciator_l",
		    		noAltImage: true
				},
				{//low_vac
		    		__proto__ : instrumentTypes.static,
		    		left: 68, top:18,
		    		low_vac_l:0, low_vac_r:0,
		    		isHidden: function(){return !this.parentInstrument.testMode && !this.low_vac_l && !this.low_vac_r;},
		    		listen:[ "low_vac_l", "low_vac_r" ],
		    		src:"annunciator/annunciator_vac",
		    		noAltImage: true
				},
				{//low_vac_r
		    		__proto__ : instrumentTypes.static,
		    		left: 91, top:18,
		    		isHidden: function(){return !this.parentInstrument.testMode && !this.value;},
		    		listen:"low_vac_r",
		    		src:"annunciator/annunciator_r",
		    		noAltImage: true
				},
				{//low_voltage
		    		__proto__ : instrumentTypes.static,
		    		left: 103, top:18,
		    		isHidden: function(){return !this.parentInstrument.testMode && !this.value;},
		    		listen:"low_voltage",
		    		src:"annunciator/annunciator_volts",
		    		noAltImage: true
				},
			]
		},
		{//CLOCK group  
			__proto__ : instrumentTypes.group,
			x: AUX_COL2_X,
			y: AUX_ROW1_Y,        
			clockMode:0, clockVoltageOATMode:0, flightTimeDelta:0, clockEllapsedStart:null, clockControlDownTimestamp:null,
			instruments:[
			{//CLOCK_oat_button
				__proto__ : controlTypes.base,
				x: 0, y: -53, w: 40,h: 40,  
				onClick : function(){
					this.parentInstrument.clockVoltageOATMode = this.parentInstrument.clockVoltageOATMode==2?0:this.parentInstrument.clockVoltageOATMode+1;
				}
			},
			{//CLOCK_select_button
				__proto__ : controlTypes.base,
				x: -22, y: 50, w: 40,h: 40,
				onClick : function(){
					this.parentInstrument.clockMode = this.parentInstrument.clockMode==3?0:this.parentInstrument.clockMode+1;
					this.parentInstrument.clockControlDownTimestamp = null;
				}
			},
			{//CLOCK_control_button
				__proto__ : controlTypes.base,
				x: 22, y: 50, w: 40,h: 40, 
				onClick : function(){
					if( this.parentInstrument.clockMode==CLOCK_ET ){
						//@todo - implement countdown
						if( this.parentInstrument.clockEllapsedStart===null )
							this.parentInstrument.clockEllapsedStart = time;
						else
							this.parentInstrument.clockEllapsedStart = null;
					}
				},
				onMouseDown : function(){
					if( this.parentInstrument.clockMode==CLOCK_FT )
						this.parentInstrument.clockControlDownTimestamp = time;
				},
				onMouseUp : function(){
					this.parentInstrument.clockControlDownTimestamp = null;
				},
			},
			{//CLOCK_arrow
				__proto__ : instrumentTypes.switchImages,
				x: -35, y: +12,
				src:["clock/clock_arrow_ut","clock/clock_arrow_lt","clock/clock_arrow_ft","clock/clock_arrow_et"],
				getFrame:function(){
					return this.parentInstrument.clockMode;	
				},
				noAltImage: true
			},
			{//CLOCK_text                 
				__proto__ : instrumentTypes.text,
				x: -34, y: -18,  
				listen:["zulu_time", "local_time", "outside_air_temp","bus_voltage"],
				zulu_time:0, local_time:0, outside_air_temp:0, bus_voltage:0,
				glyphCount:17,
				lineSpacing:28,
				glyphRemap:{ 'E':10,'C':11,'F':12,':':13,'.':14,'-':15},
				specialKerning:{
					':':{offset:-4,leading:5},
					'.':{offset:-4,leading:5},
					' ':{offset:-4,leading:5}//no glyph remap exists for ' ', but it needs not to be drawn anyway
				},
				getDisplayedValue:function(){
					var oatv, tempr, parentInstrument = this.parentInstrument;
					//would have been much simpler if I didn't have to put space between the digits, but would need two texts then
					//" " is as wide as "." and "_" as wide as a digit
					switch( parentInstrument.clockVoltageOATMode ){
						case 0://voltage, put a space between digits 
							oatv = (this.bus_voltage<10?"_":Math.floor(this.bus_voltage/10)) +' '+//first digit or space if none
								 Math.floor(this.bus_voltage%10) + '.'+//second digit
								 Math.floor( (this.bus_voltage*10)%10 ) + ' E';
							break;
						case 1://fahrenheit
							tempr = Math.floor(this.outside_air_temp);
							oatv = ' F'; 
							break;
						case 2://celsius
							tempr = Math.floor((this.outside_air_temp-32)*5/9);//convert fahrenheit to celsius
							oatv = ' C';
							break;
					}
					if( tempr!==undefined ){
						var neg = tempr<0;
						tempr = Math.abs(tempr);
						oatv = (tempr<10?(neg?"-":"_"):Math.floor(tempr/10)) +' '+ Math.floor(tempr%10) + oatv;
						if( !neg || tempr<10 )
							oatv = "_ "+oatv;
						else
							oatv = "- "+oatv;	
					}
					
					if( parentInstrument.clockControlDownTimestamp && (time-parentInstrument.clockControlDownTimestamp)>3000 ){
						parentInstrument.flightTimeDelta = time;//more than 3 seconds have ellapsed holding down the button
						parentInstrument.clockControlDownTimestamp = null;
					}
						
					var t, colon;			
					switch( parentInstrument.clockMode ){
						case CLOCK_UT://UT
							t = this.zulu_time;
							break;
						case CLOCK_LT://LT
							t = this.local_time;
							break;
						case CLOCK_FT://FT - flight time
							//@todo - reset when control button is pressed? according to the source it resets 3000 seconds after pressed
							t = Math.floor((time-this.parentInstrument.flightTimeDelta)/1000);//in seconds
							t = t%(99*60+60);//cycle every 99 minutes and 60 seconds
							t *= 60;
							break;
						case CLOCK_ET://ET
							if( parentInstrument.clockEllapsedStart!==null  ){
								t = Math.floor((time-parentInstrument.clockEllapsedStart)/1000);
								t = t%(99*60+60);
								t *= 60;
							}
							else{
								t = 0;
								colon = true;
							}
							break;
					}
					var	hours = Math.floor(t/(60*60)),
						minutes = Math.floor((t/60)%60);
					if( !colon )
						colon = (new Date().getMilliseconds()>500); 
					return " "+oatv+"\n__" + (hours<10?("0"+hours):hours) + (colon?":":" ") + (minutes<10?("0"+minutes):minutes);//"_" is a regular space, " " is shorter space
				},
				src:"clock/h5_clock_numbers",
				noAltImage: true
			}]
		},
		{//FLAPS
			//@todo - move sligtly up so we can cut more pixels	
			__proto__ : instrumentTypes.group,
			x: 385,
			y: 700,        
			value:0,
			interpolateSpeed: 0,//but not for the children
			listen: "flaps_handle_ratio",//@todo : 0-1 not 0-4
			setChildValues: true,//is it needed?
			control: {//FLAPS_lever_drag
				__proto__ : controlTypes.drag,
				x: 12, y: -10, w: 50,h: 50,
				getY: function(){
					return this.y+this.parentInstrument.value*FLAPS_LeverTravel*coordScale;
				},
				/*onMouseDown : function(x,y){
					//this.mouseDownTime = new Date().getTime();
					controlTypes.drag.onMouseDown.call(this,x,y);
					this.leverSpeed = lever.interpolateSpeed;
					lever.interpolateSpeed = 1000;//don't slow down when moving manualy
				},
				onMouseUp : function(x,y){
					//this.mouseDownTime = new Date().getTime();
					controlTypes.drag.onMouseUp.call(this,x,y);
					lever.interpolateSpeed = this.leverSpeed;
				},*/
				onMouseDown : function(x,y){
					this.mouseDownTime = new Date().getTime();
					controlTypes.drag.onMouseDown.call(this,x,y);
					this.changedValue = false;
				},
				onMouseUp : function(x,y, dx,dy){
					var dt = (new Date().getTime()-this.mouseDownTime);
					if( dt<330 && Math.abs(this.initialValue-this.parentInstrument.value )<0.01 ){
						//the value has not chaged, but if the lever has moved a bit in short time - give it some push
						if( dy<0 )
							this.onMouseDrag( x,y, dx,dy-FLAPS_LeverTravel/9 );//call the method from controlTypes.drag
						else
							this.onMouseDrag( x,y, dx,dy+FLAPS_LeverTravel/9 );
					}
					controlTypes.drag.onMouseUp.call(this,x,y);
				},
				onDrag : function( x,y, dx,dy ){
					var deltaVal = Math.abs(dy/(FLAPS_LeverTravel));
					if(deltaVal<1/6)
						deltaVal = 0;
					else if (deltaVal<3/6)
						deltaVal = 1/3.0;
					else if (deltaVal<5/6)
						deltaVal = 2/3.0;
					else
						deltaVal = 1;
					
					/*var dt = (new Date().getTime()-this.mouseDownTime);
					if( deltaVal>1/3.01 && deltaVal<1/2.99 && dt<250 )
						deltaVal = 1/3;//do not allow going 2 ticks up/down so fast (but 3 is ok)*/
					
					var val = this.initialValue + (dy<0?-deltaVal:deltaVal);
					if( val<0 )
						val = 0;
					if( val>1 )
						val = 1;
					
					this.changedValue |= Math.abs(this.initialValue-val)>0.01;
					//console.log( dy + ' , ' + val );//+ ' time:' + dt );
					return val;
				}
			},
			instruments:[
			{//FLAPS_bg
				//@todo - merge with frontPanel?
				__proto__ : instrumentTypes.static,
				x: 0, y: 0,
				src:"flaps/flaps_bg"
			},
			{//FLAPS_label
				//@todo - merge with frontPanel?
				__proto__ : instrumentTypes.static,
				x: 0, y: 57,
				src:"flaps/flaps_label"
			},
			{//FLAPS_pointer
				__proto__ : instrumentTypes.slide,
				x: -15, y: -35,
				interpolateSpeed: 100,
				intSnap: true,
				getValue: function(){
					var dist = 60;
					if( this.value==1/3 )//??
						return Math.floor(0.49*dist);
					else if( this.value==2/3 )
						return Math.floor(0.72*dist);
					else
						return Math.floor(this.value*dist);//0 and 1					
				},
				listen: "flaps_ratio",
				src:"flaps/flaps_pointer"
			},
			{//Flaps lever background control
				__proto__ : instrumentTypes.area,
				x:14,y:10,w:40,h:128,
				control:{
					__proto__ : controlTypes.base,
					x:0,y:0,w:40,h:128,
					onClick : function(x,y){
						//The flaps background is clicked, not the lever
						var flaps = this.parentInstrument.parentInstrument,
							flapsY = flaps.y+flaps.control.getY(),
							flapsValue = getInstrumentValue( flaps, 'flaps_handle_ratio' );
						if( flapsY>y )
							flapsValue-=1/3.0;
						else
							flapsValue+=1/3.0;
						if( flapsValue>-0.001 && flapsValue<1.001 )
							setInstrumentValue( flapsValue,'flaps_handle_ratio' );
					}
				}
			},
			{//FLAPS_lever
				__proto__ : instrumentTypes.slide,
				x: 21, y: 5,
				init: function(){
					lever = this;
					instrumentTypes.slide.init.call(this);
				},
				intSnap: true,
				inputValueRange:[0,1],
				outputValueRange:[0,FLAPS_LeverTravel],
				interpolateSpeed:150,			
				src:"flaps/flaps_lever"
			}]
		},
		{//ASI_Knob
			__proto__ : config.useDegrees?instrumentTypes.rotate:instrumentTypes.rotateRadians,
			x: COL1_X - KNOB_OFFSET_X,
			y: ROW1_Y + KNOB_OFFSET_Y,
			valueScale:4,
			//control is in ASI_temp
			listen: "airspeed_calibration_angle",
			src:"asi/asi_knob"
		},
		{//ASI_Knob_reflection
			__proto__ : instrumentTypes.static,
			x: COL1_X - KNOB_OFFSET_X,
			y: ROW1_Y + KNOB_OFFSET_Y,
			src:"asi/asi_knob_reflection"
		},
		{//AI_knob
			//@todo - no rotation is visible
			__proto__ : instrumentTypes.rotate,
			x: COL2_X,
			y: ROW1_Y + KNOB_OFFSET_Y,
			inputValueRange: [-1,1],
			//control is in AI_Bars
			listen:"ai_bars_position",
			src:"ai/ai_knob"
		},
		{//AI_knob
			__proto__ : instrumentTypes.static,
			x: COL2_X,
			y: ROW1_Y + KNOB_OFFSET_Y,
			src:"ai/ai_knob_reflection"
		},
		{//ALT_knob
			//@todo - no rotation is visible
			__proto__ : config.useDegrees?instrumentTypes.rotate:instrumentTypes.rotateRadians,
			x: COL3_X - KNOB_OFFSET_X,
			y: ROW1_Y + KNOB_OFFSET_Y,
			//control is in ALT_pressure
			listen: "kohlsman_setting",
			src:"altimeter/altimeter_knob"
		},
		{//ALT_knob_reflection
			__proto__ : instrumentTypes.static,
			x: COL3_X - KNOB_OFFSET_X,
			y: ROW1_Y + KNOB_OFFSET_Y,
			src:"altimeter/altimeter_knob_reflection"
		},
		{//DIR_GYRO_set_knob                     
			__proto__ : config.useDegrees?instrumentTypes.rotate:instrumentTypes.rotateRadians,
			x: COL2_X - KNOB_OFFSET_X,
			y: ROW2_Y + KNOB_OFFSET_Y,
			listen:["gyro_drift_angle"],
			control: {
				__proto__: knobControl,
				invert:true,
				sendDelta: false,//config.protocolVersion==3,//true,
				intValue: true,
				deltaScale: 360/(2*AngleRef)//send the delta in degrees
			},	
			src:"dir_gyro/dir_gyro_set_knob"
		},
		{//DIR_GYRO_set_knob_reflection                     
			__proto__ : instrumentTypes.static,
			x: COL2_X - KNOB_OFFSET_X,
			y: ROW2_Y + KNOB_OFFSET_Y,
			src:"dir_gyro/dir_gyro_set_knob_reflection"
		},
		{//dir_gyro_heading_knob                     
			__proto__ : config.useDegrees?instrumentTypes.rotate:instrumentTypes.rotateRadians,
			x: COL2_X + KNOB_OFFSET_X,
			y: ROW2_Y + KNOB_OFFSET_Y,
			listen:["plane_heading_bug"],
	        control: { __proto__: knobControl, invert:true },
			src:"dir_gyro/dir_gyro_heading_knob"
		},
		{//dir_gyro_heading_knob_reflection                     
			__proto__ : instrumentTypes.static,
			x: COL2_X + KNOB_OFFSET_X,
			y: ROW2_Y + KNOB_OFFSET_Y,
			src:"dir_gyro/dir_gyro_heading_knob_reflection"
		},
		{//ADF_knob                     
			__proto__ : config.useDegrees?instrumentTypes.rotate:instrumentTypes.rotateRadians,
			x: COL4_X - KNOB_OFFSET_X,
			y: ROW3_Y + KNOB_OFFSET_Y,
			listen:"adf1_card",
			control: { __proto__: knobControl, invert:true },
			src:"adf/adf_knob"
		},
		{//ADF_knob_reflection
			//@todo - no reflection overlay                     
			__proto__ : instrumentTypes.static,
			x: COL4_X - KNOB_OFFSET_X,
			y: ROW3_Y + KNOB_OFFSET_Y,
			src:"adf/adf_knob_reflection"
		},
		{//VOR_knob                     
			__proto__ : config.useDegrees?instrumentTypes.rotate:instrumentTypes.rotateRadians,
			x: COL4_X - KNOB_OFFSET_X,
			y: ROW1_Y + KNOB_OFFSET_Y,
			listen:"nav1_obs",
			control: { __proto__: knobControl, invert:true },
			src:"vor/vor_knob"
		},
		{//VOR_knob_reflection                     
			__proto__ : instrumentTypes.static,
			x: COL4_X - KNOB_OFFSET_X,
			y: ROW1_Y + KNOB_OFFSET_Y,
			src:"vor/vor_knob_reflection"
		},
		{//VOR2_knob                     
			__proto__ : config.useDegrees?instrumentTypes.rotate:instrumentTypes.rotateRadians,
			x: COL4_X - KNOB_OFFSET_X,
			y: ROW2_Y + KNOB_OFFSET_Y,
			listen:"nav2_obs",
			control: { __proto__: knobControl, invert:true },
			src:"vor2/vor2_knob"
		},
		{//VOR2_knob_reflection                     
			__proto__ : instrumentTypes.static,
			x: COL4_X - KNOB_OFFSET_X,
			y: ROW2_Y + KNOB_OFFSET_Y,
			src:"vor2/vor2_knob_reflection"
		},
		{//EGT_REF_knob
			__proto__ : instrumentTypes.rotate,
			x: AUX_COL2_X - 42,
			y: AUX_ROW2_Y,
			inputValueRange:[600, 1800 ],
			control: { __proto__: knobControl, invert:true, knobSpeed: 0.05 },
			listen:"egt_ref",
			src:"egt_fuel_flow/egt_knob",
		},
		{//EGT_REF_knob_refl
			__proto__ : instrumentTypes.static,
			x: AUX_COL2_X - 42,
			y: AUX_ROW2_Y,
			src:"egt_fuel_flow/egt_knob_reflection"
		}
		]
	},
	{//demo group
		__proto__ : instrumentTypes.group,
		x: 80,
		y: 60,
		doNotPause : true,
		isHidden: function(){
			return false;
		},
		screenPositioning: true,
		getEllapsedSeconds : function( startTime ){
			return Math.floor( (new Date().getTime() - startTime )/1000 );
		},  
		instruments:[
			{//demo box
				__proto__ : instrumentTypes.switchImages,
				x: 0,
				y: 0,
				src:["demo/demo_remaining","demo/demo_paused"],
				getDisplayedValue:function(){
					var timeSec = 600;
					var isPaused  = this.parentInstrument.isPaused;
					if( isPaused===undefined )
						isPaused =  this.parentInstrument.getEllapsedSeconds( localStorage.timeFlightStarted ) > timeSec;
					return isPaused?1:0;
				},
				noAltImage: true
			},
			{//time                 
				__proto__ : instrumentTypes.text,
				x: -30, y: 21,  
				listen:"indicated_airspeed",
				glyphCount:11,
				glyphRemap:{ ':':10 },
				leading: 4,
				specialKerning:{
					':':{offset:-4,leading:5}
				},
				indicated_airspeed : 0,
				getDisplayedValue:function(){
					var timeSec = 600, penaltySec = 600;
					if( this.timeFlightStarted===undefined ){
						if( localStorage.timeFlightStarted!==undefined  )  
							this.timeFlightStarted = parseInt(localStorage.timeFlightStarted);
						if( this.timeFlightStarted===undefined  && this.indicated_airspeed>70 )
							localStorage.timeFlightStarted = this.timeFlightStarted = new Date().getTime();                                 
					}
					var ellapsedSeconds;
					if( this.timeFlightStarted===undefined )
						ellapsedSeconds = 0;
					else
						ellapsedSeconds = this.parentInstrument.getEllapsedSeconds( this.timeFlightStarted );
						
					if( ellapsedSeconds>timeSec+penaltySec ){
						ellapsedSeconds = 0;//restart after the penalty time has passed
						localStorage.removeItem("timeFlightStarted");
						delete this.timeFlightStarted;
					}
					
					var t;
					if( ellapsedSeconds>timeSec ){
						paused = this.parentInstrument.isPaused = true;
						t =  timeSec+penaltySec - ellapsedSeconds;
					}  
					else{
						paused = this.parentInstrument.isPaused = false;
						t =  timeSec - ellapsedSeconds;
					}
					
					var minutes = Math.floor(t/60),
						seconds = Math.floor(t%60);  
	
					return (minutes<10?("0"+minutes):minutes) + ":" + (seconds<10?("0"+seconds):seconds);
				},
				src:"demo/demo_numbers",
				noAltImage: true
			}
			
		]
	}
];
