/* global altImagePrefix, log, lastProgress:true, drawProgressIteration:true, graphics, navigator, window, drawProgressBar, imagesFolder, imagePostfix, requestStopDrawLoop:true, controls, instruments, startDrawLoop, Image, initInstruments, onReady, postponeResize:true, TextManager */
/* exported showAltImages, reloadImages, onLoadedActiveImages */
/* jshint curly: false */
var ImageLoader = {
	init: function(){
		var loadingImages = false, loadingActiveImages = false,
			imageCount=0, altImageCount=0, sharedImageCount=0,
			showingAltImages = false,
			loaded = false,
			numLoadingImages = 0,
			loadingImagesLimit = 10,//limit the number of images that wait to be loaded to have more gradual progress in the progress bar
			pendingImages = [],
			loadedImageCount=0, callbackImageCount=0;

		this.initCalled = true;
		this.isLoadingImages = function(){return loadingImages;};
		this.isLoadingActiveImages = function(){return loadingActiveImages;};
		this.loadedImages = function(){return imageCount!==0 || loaded;};
		this.usingAltImages = function(){return showingAltImages;};
		
		this.loadImages = function(instruments,activeReadyCallback,allReadyCallback){
			if( !activeReadyCallback )
				activeReadyCallback = onLoadedActiveImages;
			if( !loadingImages ){
				loadingActiveImages = true;
				loadedImageCount = 0;
				if( !imageCount ){
					countImages( instruments );
					log( "total images:"+imageCount+", alt images:"+altImageCount+", shared images:"+sharedImageCount );
				}
				loaded = (imageCount===0);
				loadingImages = !loaded;
				//load the active images first
				callbackImageCount = showingAltImages?(altImageCount+sharedImageCount):(imageCount-altImageCount); 
				doLoadImages( instruments, showingAltImages, true, onLoadedImages );
				if( !imageCount )
					window.setTimeout(onLoadedImages,1);
			}
			function onLoadedImages(){	
				log("active images loaded, continuing...");
				loadingActiveImages = false;
				lastProgress=0;
				drawProgressIteration = 0;
				activeReadyCallback();
				//load the non-active images next
				callbackImageCount = imageCount;
				
				if( altImagePrefix )
					doLoadImages( instruments, !showingAltImages, false, onFinishedLoading );
				else
					onFinishedLoading();
				
				function onFinishedLoading(){
					log("... all images loaded, pending="+pendingImages.length);
					loadingImages = false;
					loaded = true;

					if( postponeResize ){
						log( "call postponed resize" );
						graphics.resize();
						postponeResize = false;
					}

					if( allReadyCallback )	
						allReadyCallback(); 
				}
			}
		};
		
		/*
		 * Reload the images, used when switching from hi-res to low-res textures 
		 */
		this.reloadImages = function(scale){
			requestStopDrawLoop = true;//drawLoopStopped = true;

			var activeImagesReloaded = function(){
				log('Active images re-loaded, pending='+pendingImages.length);
				adjustInstrumentCoordinates( instruments, scale );
				for( var i=0; i<controls.length; i++ )
					if( controls[i].adjustCoordinates )
						controls[i].adjustCoordinates( scale );
				log( 'Starting draw loop after reloading active images' );			
				startDrawLoop();
				
				function adjustInstrumentCoordinates( instruments, scale ){
					for( var i=0; i<instruments.length; i++ ){
						instruments[i].adjustCoordinates(scale);
						if( instruments[i].instruments ){
							adjustInstrumentCoordinates( instruments[i].instruments, scale );
						}
					}
				}
			};

			clearImages( instruments );
			this.loadImages( instruments, activeImagesReloaded  /*, allImagesLoaded*/ );
			
			
		};
		
		this.showAltImages = function( instruments, on ){
			if( loadingImages )
				return;//don't change this while images are loaded
			if( on!==undefined ){
				if( on==showingAltImages )
					return;
				showingAltImages = on;
			}

			for( var i=0; i<instruments.length; i++ ){
				var ti, ent = instruments[i];
				if( ent.src && !ent.noAltImage ){
					if( !(ent.src instanceof Array) ){
						ti = ent.image;
						ent.image = ent.altImage;
						ent.altImage = ti;
					}
					else{
						ti = ent.imageFrames;
						ent.imageFrames = ent.altImageFrames;
						ent.altImageFrames = ti;
						ent.image = ent.imageFrames[0];
					}
				}
				if( ent.instruments )
					this.showAltImages( ent.instruments );
			}
		};
		
		function countImages(instruments){
			//var count = 0;
			for( var i=0; i<instruments.length; i++ ){
				var ent = instruments[i];
				if( ent.src ){
					if( !(ent.src instanceof Array) ){
						imageCount++;
						if( altImagePrefix ){
							if( !ent.noAltImage ){
								imageCount++;
								altImageCount++;					
							}
							else
								sharedImageCount++;
						}
					}
					else{
						imageCount+=ent.src.length;
						if( altImagePrefix ){
							if( !ent.noAltImage ){
								imageCount+=ent.src.length;
								altImageCount+=ent.src.length;
							}
							else
								sharedImageCount+=ent.src.length;
						}
					}
				}
				if( ent.instruments )
					countImages( ent.instruments );
			}
		}
		
		/*
		 * altImg - load the alt images, othervise the regular images
		 * sharedImg - load the images that are used both as regular and alt image 
		 */
		function doLoadImages( instruments, altImg, sharedImg, allReadyCallback ){
			for( var i=0; i<instruments.length; i++ ){
				var ent = instruments[i];
				if( ent.src ){
					var doImage = (!ent.noAltImage && !altImg) || (sharedImg && ent.noAltImage),
						doAltImage = altImg && altImagePrefix && !ent.noAltImage;  
					if( !(ent.src instanceof Array) ){
						var image = null, altImage=null;
						if( doImage )
							image = loadImage(ent.src,allReadyCallback);
						if( doAltImage )
							altImage = loadAltImage(ent.src,allReadyCallback);
						if( showingAltImages && !ent.noAltImage ){			
							//image and alt image are swapped
							if( doAltImage )
								ent.image = altImage;
							if( doImage )
								ent.altImage = image;
						}
						else{
							if( doAltImage )
								ent.altImage = altImage;
							if( doImage )
								ent.image = image;
						}
					}
					else{
						var imageFrames=null, altImageFrames=null;
						if( doImage )
							imageFrames = new Array(ent.src.length);
						if( doAltImage )
							altImageFrames = new Array(ent.src.length);
						for( var n=0; n<ent.src.length; n++ ){
							if( doImage )
								imageFrames[n] = loadImage( ent.src[n],allReadyCallback );
							if( doAltImage )
								altImageFrames[n] = loadAltImage( ent.src[n],allReadyCallback );}

						if( showingAltImages && !ent.noAltImage ){			
							//image and alt image are swapped
							if( doImage )
								ent.altImageFrames = imageFrames;
							if( doAltImage ){
								ent.imageFrames = altImageFrames;
								ent.image = ent.imageFrames[0];
							}
						}
						else{
							if( doImage ){
								ent.imageFrames = imageFrames;
								ent.image = ent.imageFrames[0];
							}
							if( doAltImage )
								ent.altImageFrames = altImageFrames;
						}
					}
				}
				if( ent.instruments )
					doLoadImages( ent.instruments, altImg, sharedImg, allReadyCallback );
			}
		}
		
		function clearImages( instruments ){
			if( !navigator.isCocoonJS )
				return;
			for( var i=0; i<instruments.length; i++ ){
				var n, ent = instruments[i];
				if( ent.image )
				dispose( ent.image );
				dispose( ent.altImage );
				if( ent.imageFrames ){
					for( n=0; n<ent.imageFrames.length; n++ )
						dispose( ent.imageFrames[n] );
				}
				if( ent.altImageFrames ){
					for( n=0; n<ent.altImageFrames.length; n++ )
						dispose( ent.altImageFrames[n] );
				}
				if( ent.instruments )
					clearImages( ent.instruments );
			}
			function dispose( image ){
				if( image && image.dispose )
					image.dispose();
			}
		}
		
		function loadImage(src, allReadyCallback ){
			src = imagesFolder + '/' + src + imagePostfix;
			if( loadedImageCount===0 && numLoadingImages===0 && (typeof drawProgressBar!=='undefined') )
				drawProgressBar(0);
			var image = new Image();
				image.onload = function(){countLoadedImage(allReadyCallback,src);};
			if( numLoadingImages>=loadingImagesLimit )
				pendingImages.push({image:image,src:src});
			else{
				numLoadingImages++;
				image.src = src;
			}
			return image;
		}
		
		function loadAltImage(src, allReadyCallback){
			var n=src.lastIndexOf("/");
			if( n<0 )
				return loadImage( altImagePrefix + src, allReadyCallback );
			return loadImage( src.substring(0,n+1) + altImagePrefix + src.substring(n+1), allReadyCallback );
		}
		
		/*
		 * Decrement the image refcounter, when 0 - call onLoadedActiveImages 
		 * If the callback is the 'draw' method - calls it through requestAnimFrame 
		 */
		function countLoadedImage( callback ){
			loadedImageCount++;
			numLoadingImages--;
			if( numLoadingImages<loadingImagesLimit && pendingImages.length>0 ){
				numLoadingImages++;
				var imgdata = pendingImages.pop();
				imgdata.image.onload = function(){countLoadedImage(callback,imgdata.src);};
				imgdata.image.src = imgdata.src;
			}
			if( loadingActiveImages && (typeof drawProgressBar !== 'undefined') )
				drawProgressBar( loadedImageCount/callbackImageCount );
			//log( 'img:'+loadedImageCount+'/'+callbackImageCount + " - "+src );
			if( loadedImageCount==callbackImageCount )
				callback();//no more images left to load
		}
		
		/*
		 * Trim all images if possible
		 * Call draw() through requestAnimFrame when done 
		 * @todo - if ent has instruments - trim them too 
		 * @todo - trim only fixed, rotate and slide instruments 
		 */
		function onLoadedActiveImages(){
			if( postponeResize ){
				log( "call postponed resize" );
				graphics.resize();
				postponeResize = false;
			}

			log('Active images - initial loading, pending='+pendingImages.length );

			initInstruments( instruments );
			onReady();
		}	
	}
};

if( typeof TextManager === 'undefined' )
	ImageLoader.init();