/* global graphics, localValues, netGetValue, netSetLocalValue, coordScale, assert, widgetsZoom, drawXOrigin, drawYOrigin, time, netUpdate, config */
/* exported instrumentTypes, useOffscreenCanvas, getInstrumentValue, setInstrumentValue, defaultIntepolateSpeed*/
/* jslint browser: true*/
/* jshint proto: true */
/* jshint curly: false */

var instrumentTypes = {},
	instrumentRefs = {},
	useOffscreenCanvas = false,
	defaultIntepolateSpeed = 1000;//pixels or degrees per second

function getInstrumentValue(ent,key){
	if( key in localValues )
		return localValues[key];
	else
		return netGetValue(ent,key);
}

function setInstrumentValue( val, key, dontSend, isDelta ){
	if( key in localValues )
		localValues[key] = val;
	else
		return netSetLocalValue( val, key, dontSend, isDelta );
}

function initRefs( instruments ){
	for( var i=0; i<instruments.length; i++ ){
		var instrument = instruments[i];
		if( instrument.ref ){
			if( instrumentRefs[instrument.ref]===undefined )
				instrumentRefs[instrument.ref] = instrument;
			else{
				if( !(instrumentRefs[instrument.ref] instanceof Array) )
					instrumentRefs[instrument.ref] = [instrumentRefs[instrument.ref]];
				instrumentRefs[instrument.ref].push(instrument);
			}
		}
		if( instrument.instruments )//e.g. a group
			initRefs( instrument.instruments );
	}
}

instrumentTypes.base = {//value,[valueOffset],[valueScale],[interpolateSpeed],[intSnap],[x,y],[parentInstrument], draw()
	init:function(){
		//create all values, set them to 0, unless they already exist			
		if( this.value===undefined ){
			var hadProperty = false;
			if( this.listen ){ 
				//try to assign 0 to any properties corresponding to listened keys			
				if( !(this.listen instanceof Array) )
					hadProperty = this.initProperty(this.listen) || hadProperty;					
				else{				
					for( var n=0; n<this.listen.length; n++ )
						hadProperty = this.initProperty(this.listen[n]) || hadProperty;
				}
			}    
			if( !hadProperty )
				this.value = 0;//creates a this.value property										
		}
		if( coordScale!=1 )
			this.adjustCoordinates( coordScale, true );
		assert( !this.repeat || this.inputValueRange, "Instrument("+this.src+") has 'repeat', but no inputValueRange" );
	},
	adjustCoordinates: function( scale /*, atInit*/ ){
		if( this.x!==undefined ){
			this.x *= scale;
			this.y *= scale;
		}      
		else if( this.left!==undefined ){
			this.left *= scale;
			this.top *= scale;
		}
	},
	setInitialTransform: function( x, y ){
		//if( !this.parentInstrument )
		if( x===undefined )
			x = 0;
		if( y===undefined )
			y = 0;
		var parent = this.parentInstrument,
			screenPos = this.screenPositioning,
			zoomX = widgetsZoom,
			zoomY = widgetsZoom;
		while( parent && parent.instruments ){
			if( parent.fitContents ){
				//for now - only one level with zoomContents is allowed, because the position is not calculated properly
				zoomX *= parent.w/100;
				zoomY *= parent.h/100;
			}
			x += parent.getX()*widgetsZoom;
			y += parent.getY()*widgetsZoom;
			screenPos = screenPos || parent.screenPositioning;
			parent = parent.parentInstrument;
		}
		if( !screenPos ){
			x += drawXOrigin;
			y += drawYOrigin;
		}
		else
			x = x;  
		graphics.setPositionAndScale( x,y, zoomX, zoomY );
	},
	getX: function(){return this.x|0;},//An instrument may specify different behavior, e.g. getX(){return this.x+this.value*10;};
	getY: function(){return this.y|0;},
	/*
	 * Sets the value, this is the place to convert uints, limit min/max, etc
	 * There can be tree types of value:
	 * 1. this.value, the value parameters are in 'this.'
	 * 2. this[key], the value parameters are in 'this.' , allows multiple values per instrument, but only 1 value parameter set
	 * 3. this[key].value, the value parameters are in 'this.key.', allows multiple values and parameter sets per instrument
	 * Don't mix them in the same instrument!!!	 	 	  	 	 	 
	 */
	setValue: function(val,key){						
		var valueObject;
		if( this.value!==undefined || !isObject(this[key]) )
			valueObject = this;
		else
			valueObject = this[key];
			
		var newValue = convertObjectValue(valueObject,val);
			
		if( valueObject.value!==undefined || valueObject[key]===undefined )
			valueObject.value = newValue;//will create the 'value' property if not present
		else
			valueObject[key] = newValue;
	},
	/*
	 * get a property of a value(key) - like inputRange, outputRange, repeat...
	 * if key is not specified or no individual property for the key exists
	 * - return the property of the default value	  
	 */	 
	getValueProperty: function(prop,key){
		if( key===undefined ){
			if( this.defaultValue ){
				var val = this[this.defaultValue];
				if( isObject(val) ){
					return val[prop];
				}                                              
			}
		}
		else if( isObject(this[key]) )
			return this[key][prop];									
		return this[prop];
	},
	/*
	 * Get the value of the default property
	 */	 
	getValue:function(){
		if( this.defaultValue ){
			var val = this[this.defaultValue];
			if( isObject(val) ){
				var valueObject = this[this.defaultValue];
				return valueObject.value;
			}
			else
				return val;
		}
		else
			return this.value;			
	},
	/*
	 * Returns the currently displayed value, which may differ from getValue
	 * See timeAdjustValue
	 * @todo - make it work with any value, not just the default 	 	 	 
	 */	 
	getDisplayedValue: function(value,cycle){
		if( isObject(value) )
			return timeAdjustObjectValue(value,value.value,cycle);
		else if( !isNaN(value) ) 
			return timeAdjustObjectValue(this,value,cycle);
		else
			return value;
	},
	/* Convenience functions when you have a single value for this instrument*/
	convertValue: function(val){
		return convertObjectValue( this, val );
	}, 
	timeAdjustValue: function(val,cycle){
		return timeAdjustObjectValue( this, val, cycle );
	},
	//used by init, nothing to see here
	initProperty:function( key ){	
		if( this[key]!==undefined ){
			if( isObject(this[key]) ){
				if( this[key].value===undefined )
					this[key].value = 0;
			}
			return true;				
		}
		return false;//no such property exists		
	},
	getAbsoluteX : function(){
		var x = this.getX()/coordScale;
		var parent = this.parentInstrument;
		while( parent && parent.instruments ){
			x += parent.getX()/coordScale;
			parent = parent.parentInstrument;
		}
		return x|0;
	},
	getAbsoluteY : function(){
		var y = this.getY()/coordScale;
		var parent = this.parentInstrument;
		while( parent && parent.instruments ){
			y += parent.getY()/coordScale;
			parent = parent.parentInstrument;
		}
		return y|0;
	}	 	          
};

// Override setValue to set the values for the group instruments
instrumentTypes.group = {//instruments
	__proto__:instrumentTypes.base,
	setValue: function(value,key){//you may want to override this - the group may pass modified values to it's children
		instrumentTypes.base.setValue.call(this,value, key);
		if( this.setChildValues ){
			for( var i=0; i<this.instruments.length; i++ ){
				if( this.instruments[i].listen ){
					//netUpdate( this.instruments[i] );//this should not be needed as netUpdate calls netUpdate for the child instruments
				}
				else
					this.instruments[i].setValue(value,key);
			}
		}
	},
	draw: function(){
		for( var i=0; i<this.instruments.length; i++ )
			if( !this.instruments[i].hide && (!this.instruments[i].isHidden || !this.instruments[i].isHidden()) )
				this.instruments[i].draw();
	}
};

//============================ Helper functions ================================

/*
 * Converts the value to our scale/units
 * scales it(scaleValue)
 * adds an offset(offsetValue)
 * valueObject : contains the value properties: valueOffset, valueScale 
 */
function convertObjectValue(valueObject,val){
	if( isNaN(val) )
		return val;
	val = ( valueObject.valueOffset ? valueObject.valueOffset : 0 ) + ((valueObject.valueScale!==undefined ) ? valueObject.valueScale*val:val );
	if( config.trimValuesToRange && valueObject.outputValueRange && !valueObject.dontTrimValue ){
		if( val<valueObject.outputValueRange[0] )
			val = valueObject.outputValueRange[0];
		if( val>valueObject.outputValueRange[1] )
			val = valueObject.outputValueRange[1];
	}
	return val;
}

/* 
 * For instance it may be extrapolated (death reckoning)
 * or it's may converge gradualy to the actual value (interpolateSpeed) plus int snapping after the converging has ended
 * valueObject : contains the value properties: min/max, repeat, intSnap, extrapolate...  
 */	
//var dbgCounter = 0; 
function timeAdjustObjectValue(valueObject, val, cycle ){
	if( isNaN(val) )
		return val;
	var interpolateSpeed = (valueObject.interpolateSpeed? valueObject.interpolateSpeed : defaultIntepolateSpeed)/coordScale; 
	if( !config.dontConvergeValues && valueObject.interpolateSpeed!==0 ){//setInterpolateSpeed to 0 to disable interpolation (and extrapolation too)
		if( !valueObject.convergeData )
			valueObject.convergeData = { prevValue:val, timestamp:time };
		else{
			if( valueObject.convergeData.timestamp!=time ){
				var newval = convergeValue( val, time, valueObject.convergeData.prevValue, valueObject.convergeData.timestamp, /*valueObject.*/interpolateSpeed, valueObject.intSnap, cycle );	
				//if( val && valueObject.listen=="fuel_qty_left_tank" && dbgCounter<300){
					/*dbgCounter++;
					if( false )//prevData ){
					//log("netprev:"+convertObjectValue(valueObject,prevData.fuel_qty_left_tank) + 
					//	" netcurrent:"+convertObjectValue(valueObject,lastData.fuel_qty_left_tank) +
					//	" nextextr:"+convertObjectValue(valueObject,getInstrumentValue(valueObject,"fuel_qty_left_tank")),
					//	" extr:"+val+
					//	" prev:"+valueObject.convergeData.prevValue+
					//	" current:"+newval );
					if( dbgCounter==1 ){
						log("net-prev, net-current, net-extrapolated, final value, time");
					}
					else{
						var lo = convertObjectValue(valueObject,prevData.fuel_qty_left_tank) + 
						"\t"+convertObjectValue(valueObject,lastData.fuel_qty_left_tank) +
						"\t"+convertObjectValue(valueObject,getInstrumentValue(valueObject,"fuel_qty_left_tank"))+
						"\t"+newval+
						"\t"+time;
						lo=lo.replace(/\./g,','); 
						log( lo );
					}*/
				//}
				valueObject.convergeData.prevValue = val;
				valueObject.convergeData.timestamp = time;
				val = newval;
			}
			else
				val = valueObject.convergeData.prevValue; 
		}
	}
	return val;
}

//remove any cycles between value and prevValue
function removeCycles( prevValue, value, cycle ){
	cycle = Math.abs(cycle);             
	if( value-prevValue>cycle/2 )
		prevValue+=cycle*Math.max(1,Math.floor((value-prevValue)/cycle));
	else if( prevValue-value>cycle/2 )
		prevValue-=cycle*Math.max(1,Math.floor((prevValue-value)/cycle));
	return prevValue;
}	

/* 
 * Converges the prevValue towards value, limiting the converge speed to maxSpeed(in pixels/second)
 * This way when the value changes faster than maxSpeed - the change will happen gradualy
 * Also if intSnap is set - snaps the value to the nearest int once the animation is done       
 */  
var maxConvergeTime = 1;//the value should converge in one second max
function convergeValue( value, time, prevValue, prevTime, maxSpeed, intSnap, cycle ){//maxSpeed must be in pixels/second
	if( isNaN(value) )
		return value;
	/*if( cycle ){//remove any cycles between value and prevValue 
		cycle = Math.abs(cycle);             
		if( value-prevValue>cycle/2 )
			prevValue+=cycle*Math.max(1,Math.floor((value-prevValue)/cycle));
		else if( prevValue-value>cycle/2 )
			prevValue-=cycle*Math.max(1,Math.floor((prevValue-value)/cycle));		
	}*/
	if( cycle )//new
		prevValue = removeCycles( prevValue, value, cycle );
	var deltaTime = (time - prevTime )/1000,
		deltaValue = value - prevValue;
	if( (!intSnap || Math.abs(Math.round(value)-prevValue)>0.5) && Math.abs(deltaValue)/deltaTime>maxSpeed ){
		if( deltaValue/maxSpeed>maxConvergeTime )//if the time required for converging is greater than maxConvergeTime => converge immediately
			return value;
		//limit how fast the value converges, but if intSnap : limit coverge speed only if it's >0.5 pixels away from the destination's value rounding
		deltaValue = deltaValue>0 ? maxSpeed*deltaTime : -maxSpeed*deltaTime;
		return prevValue+deltaValue;
	}
	else{
		if( intSnap && (value==prevValue || Math.round(value)==prevValue) )
			return Math.round(value);//the value has settled - you can make it an int
		else
			return value;
	}
}

function isObject( obj ){
	return typeof obj == "object";
}