/* global config, graphics, instrumentTypes, coordScale:true, Cocoon, startPingLoop, onConnectionEstablished, ImageLoader, WebViewUI, drawXOrigin, drawYOrigin, widgetsZoom, TextManager, getControlAt, drawInstruments, drawControlBoxes, transformInputCoordinates, resizeCanvas, netOnFinshDraw, v2, InstrumentSettings, initRefs */
/* exported init, restart, ontrols, onDisconnect, onConnect, onPing, onFailPing, onReady, initInstruments, paused */
/* jslint browser: true*/
/* jshint curly: false */
/* jshint proto: true */

var connected = false,
	paused = false,
	drawLoopStopped = true,
	requestStopDrawLoop = false,
	startTime, time,
	controls = [];//Every instrument that is not a prototype of instrumentTypes.base is moved in controls
				  //Also some instruments may have a control property and these are also moved here.
				  //These controls can't listen for net values (used for knobs)
																										
function init(){
	log( "Init, platform=", navigator.platform );  
	graphics.init();
	
	if( !Math.log10 ){
		Math.log10 = Math.log10 || function(x) {
			return Math.log(x) / Math.LN10;
		};
	}
	
	if( typeof InstrumentSettings !== 'undefined' )
		InstrumentSettings.initSidebarInstruments();
	
	if( typeof TextManager !== 'undefined' ){
		TextManager.init( function(){//wait for the fonts to load
			WebViewUI.init();
			ImageLoader.init();
			resizeCanvas();
		});
	}
	else
		WebViewUI.init();
}

function restart(){
	if( connected )
		localStorage.reconnectOnStart = true;
	
	try{
		//window.top.location.reload(false);
		window.location.href = window.location.href;
		//window.top.location.href = window.location.href;
	}
	catch(e){
		error("---Error reloading location.href---");
	}
}

function onConnect( addr, serverName, flightSim ){
	if( addr!=="" ){	
		if( onConnectionEstablished )
			onConnectionEstablished(addr);
		
		//store a list of all succesfully connected addresses in history
		if( localStorage.connectionHistory ){
			var history = JSON.parse(localStorage.connectionHistory);
			for( var i=0; i<history.length;i++)
				if( history[i]==addr ){
					history.splice(i,1);//remove the connect addr if already in the list
					break;
				}
			history.reverse();
			history.push( addr );
			history.reverse();
			while( history.length>10 )
				history.pop();
			localStorage.connectionHistory = JSON.stringify(history);
		}
		else
			localStorage.connectionHistory = JSON.stringify([addr]);
	}
	graphics.resize();
	WebViewUI.sendMessage( "connect", {address:addr,serverName:serverName,flightSim:flightSim} );
	                                                                                                  
	//connected = true;
	uiLog( "onConnect hiding UI");
	WebViewUI.show( false );
	
	if( !config.constantDrawLoop && !ImageLoader.isLoadingActiveImages() )
		startDrawLoop();
}

/*
 * Called when the connection was disconnected  
 */ 
function onDisconnect( addr, reconnect ){
	uiLog( "onDisconnect show UI");
	WebViewUI.show( true );        	
	if( addr!=="" && addr!==undefined ){
		if( !reconnect )
			WebViewUI.sendMessage( "disconnect", addr );
		else
			WebViewUI.sendMessage( "reconnect", addr );
	}
}

function onPing( addr, serverName, flightSim ){
	WebViewUI.sendMessage( "ping-ok", {address:addr,serverName:serverName,flightSim:flightSim,connected:connected} );
}

function onFailPing( addr ){
	WebViewUI.sendMessage( "ping-fail", addr );
}

/*
 * Called after all active images have been loaded 
 */                
var isFSIReady = false;//true when eerything is running - events, images, etc
function onReady(){
	//@todo - stop the draw loop when loading images, start it from here
	 
	if( isFSIReady )
		return;
	isFSIReady = true;

	var canvas = graphics.getCanvas();
	canvas.addEventListener( "touchstart", onTouchStart );		
	canvas.addEventListener( "touchmove", function(event){onTouch(event,true);} );
	canvas.addEventListener( "touchend", function(event){onTouch(event,false);} );
	//canvas.addEventListener( "focusout", onFocusLost );//why this is not working

	log( 'Starting draw loop from onReady' );			
	startDrawLoop();
	
	if( !config.coldStartUI ){
		uiLog( "onReady showing UI");
		WebViewUI.show(true);
	}
	else{
		if( WebViewUI.isReady() ){
			startPingLoop();
			WebViewUI.sendMessage( "ready" );//FSI and UI are ready
		}
	}
	time = 0;
}

var touchControls = [],
	touchPoints = {};
	
/*
 * Simulate a mouse down event
 * Works with multiple instruments, but ignores multiple points on the same instrument (see ignoreControls in onMouseEvent) 
 */
function onTouchStart( evt ){
	var touches = evt.changedTouches;
	for (var i=0; i < touches.length; i++){ 
		if( touches.length && touchPoints[ touches[i].identifier ] ){
			//for some reason a mouseup event was not generated for this point, do some cleanup
			delete touchPoints[ touches[i].identifier ];
			touchControls.splice( touchControls.indexOf(mouseDownControl),1 );
		}
		var control = onMouseEvent( "onMouseDown", touches[i].clientX*window.devicePixelRatio, touches[i].clientY*window.devicePixelRatio, touchControls );
		if( control && touchControls.indexOf(control)==-1 ){       
			//log( "touchstart x:" + touches[i].clientX*window.devicePixelRatio + " y:" + touches[i].clientY*window.devicePixelRatio );		
			touchControls.push( control );//this control will be ignored by subsequent calls of onMouseEvent
			//touchPoints[ touches[i].identifier ] = {control:control, x:mouseDownX, y:mouseDownY};//x:touches[i].clientX*window.devicePixelRatio, y:touches[i].clientY*window.devicePixelRatio};
			touchPoints[ touches[i].identifier ] = {control:control, x:touches[i].clientX*window.devicePixelRatio, y:touches[i].clientY*window.devicePixelRatio};
		}
		//else - ignore, only the first touch for a control is used
    }
    evt.preventDefault();
}

/*
 * Simulate a mouse drag and mouse up events
 * Works with multiple instruments, but ignores multiple points on the same instrument 
 */
function onTouch( evt, isDrag ){
	if( !isDrag )
		isDrag = isDrag;
	var touches = evt.changedTouches;
	for (var i=0; i < touches.length; i++){
		var tp = touchPoints[ touches[i].identifier ];
		if( tp ){
			var mdc = mouseDownControl, mdx=mouseDownX, mdy =  mouseDownY;
			mouseDownControl = tp.control;
			mouseDownX = tp.x;
			mouseDownY = tp.y;  
			onMouseEvent( isDrag?"onMouseDrag":"onMouseUp", touches[i].clientX*window.devicePixelRatio, touches[i].clientY*window.devicePixelRatio );
			if( isDrag ){
				tp.moved = true;
			}
			else{//mouseup
				//if( !tp.moved )//@fixme if needed - on Chrome with touch emulation this is always true
					onMouseEvent( "onClick", touches[i].clientX*window.devicePixelRatio, touches[i].clientY*window.devicePixelRatio );
				delete touchPoints[ touches[i].identifier ];
				touchControls.splice( touchControls.indexOf(mouseDownControl),1 );
				i--; 
			}
			//return everything as it was - for the unlikely case where mouse and touch are used simultaneously
			mouseDownControl = mdc;
			mouseDownX = mdx; 
			mouseDownY = mdy; 
		}
	}
}

/*function onFocusLost(){
	console.log('defocus');
	for( var i in touchPoints ){
		var tp = touchPoints[i];
		mouseDownControl = tp.control;
		mouseDownX = tp.x;
		mouseDownY = tp.y;
		onMouseEvent( "onMouseUp", tp.x, tp.y );
	}
	touchControls = [];
	touchPoints = {};
}*/

//@todo - release on mouseleave
/*
 * Handles mouse events. Ignores the controls from mouseDownControl
 */ 
var mouseDownControl = null, mouseDownX, mouseDownY;
function onMouseEvent( name, x,y, ignoreControls ){   
	if( !connected || (name=="onMouseDrag" && !mouseDownControl ) )
		return null;
	var canvas = graphics.getCanvas();
	var func;
	//convert into canvas coordinates
	/*if( Cocoon ){
		x = x - drawXOrigin;
		y = y - drawYOrigin;
	}*/
	var mx = x,
		my = y;
	var mdx = mouseDownX,
		mdy = mouseDownY;
		
	if( !Cocoon ){		
		var brect = canvas.getBoundingClientRect();
		mx = mx-brect.left;
		my = my-brect.top;
		var scale = canvas.width/brect.width;
		mx = mx*scale;// - drawXOrigin;
		my = (my-brect.height+canvas.offsetHeight)*scale;// - drawYOrigin;

		//todo - mdx,mdy 		 
	}
	if( transformInputCoordinates ){
		mx = (mx-drawXOrigin)/(widgetsZoom*coordScale);
		my = (my-drawYOrigin)/(widgetsZoom*coordScale);
		mdx = (mdx-drawXOrigin)/(widgetsZoom*coordScale);
		mdy = (mdy-drawYOrigin)/(widgetsZoom*coordScale);
	}
	
	//log( name, x,y );
	if( name=="onMouseDrag" || name=="onMouseUp" ){
		if( mouseDownControl ){
			func = mouseDownControl[name];
			if( func ){
				if( mouseDownControl.transformCoord ){

					var p = v2(mx,my);
					mouseDownControl.transformCoord(p);
					mx = p[0];
					my = p[1];

					p = v2(mdx,mdy);
					mouseDownControl.transformCoord(p);
					mdx = p[0];
					mdy = p[1];
				}
				//drag and mouseup are sent to the control where the mouse was set down
				//before removing the below commented code - test it in c172
				//func.call( mouseDownControl, x/coordScale,y/coordScale, (x-mouseDownX)/coordScale, (y-mouseDownY)/coordScale );
				func.call( mouseDownControl, mx,my, (mx-mdx), (my-mdy) );
				//log( name, x,y );
			}
		}
		var mdc = mouseDownControl;
		if( name=="onMouseUp" )
			mouseDownControl = null;
		return mdc;
	}
	else{
		var currentControl = getControlAt(mx*coordScale,my*coordScale);
		if( currentControl ){
			if( ignoreControls && ignoreControls.indexOf(currentControl)!=-1 )
				return null;
			
			if( currentControl.transformCoord ){
				var p = v2(mx,my);
				currentControl.transformCoord(p);
				mx = p[0];
				my = p[1];
			}
				
			func = currentControl[name]; 
			if( func ){
				func.call( currentControl, mx,my );
			}
			if( name=="onMouseDown" ){
				mouseDownControl = currentControl;
				mouseDownX = x;
				mouseDownY = y;
			} 
		}
		return currentControl;
	} 
}

function initInstruments( instruments, parentInstrument ){
	if( !parentInstrument ){
		log( "initializing instruments" );
		if( typeof initRefs !== 'undefined' )
			initRefs(instruments);
	}
	//function a() {}
	//var protoSupported = ((new a()).__proto__ === a.prototype );//in case __proto__ is deprecated - ECMAScript 6 will support setPrototypeOf
	//better? - use Object.create( proto, {prop1:1,prop2:2} ) in data.js, a much uglier though
	
	var i=0;
	while( i<instruments.length ){
	
		if( parentInstrument )
			instruments[i].parentInstrument = parentInstrument;
			
		/*if( !protoSupported ){//@todo - not tested
			var e = instruments[i];
			while( e.__proto__ && e.__proto__ !== e.prototype ){
				var prot = e.__proto__;
				e.setPrototypeOf( prot );//@todo - the whole prototype chain
				delete e.__proto__;
				e = prot;
			}
		}*/
			
		if( instrumentTypes.base.isPrototypeOf(instruments[i]) ){
			if( instruments[i].delete )
				instruments.splice(i,1);//remove the i'th element
			else{
				instruments[i].init();
				if( instruments[i].control ){
					instruments[i].control.parentInstrument = instruments[i];//some controls are inside instruments
					controls.push(instruments[i].control);
					if( instruments[i].control.init )
						instruments[i].control.init();
				}
				if( instruments[i].instruments ){//e.g. a group
					initInstruments( instruments[i].instruments, instruments[i] );
				}
				i++;
			}
		}
		else{  
			//this is a control move it to 'controls'
			//it can be there by default, but it's more convenient
			//to keep it around the visual instruments
			var control = instruments[i];
			instruments.splice(i,1);//remove the i'th element
			controls.push(control); 
			if( control.init )
				control.init();
		}
	}
}

//timer stuff shim
window.performance = window.performance || {};
window.performance.now = (function() {
  return window.performance.now       ||
         window.performance.mozNow    ||
         window.performance.msNow     ||
         window.performance.oNow      ||
         window.performance.webkitNow ||
         function() { return new Date().getTime(); };
})();

function getTime(){
	if( startTime===undefined ){
		startTime = window.performance.now();
		return 0;
	}
	return window.performance.now() - startTime;
}

function startDrawLoop(){
	requestStopDrawLoop = false;
	if( drawLoopStopped ){
		drawLoopStopped = false;
		log('about to start draw loop');
		drawLoop();
		log('draw loop started');
	}
}
/*
 * The animation loop method
 */
function drawLoop(){	
	if( (!connected && !config.constantDrawLoop) || requestStopDrawLoop ){
		log('draw loop end');
		drawLoopStopped = true;
		requestStopDrawLoop = false;
		return;
	}
	time = getTime();//@todo - add the expected time to draw the frame - 1000/fps
	draw( time );
	netOnFinshDraw();
	window.requestAnimFrame(drawLoop);		
}

var FPSMeter = {
	//fps counter
	fps: 0,
	fpsCount: 0,
	fpsTime : null,
	draw: function(){
		if( config.drawFPS && time!==undefined ){
			this.fpsCount++;
			if( this.fpsTime!==null ){
				if( time-this.fpsTime>1000 ){
					//update the fps once a second
					this.fps = this.fpsCount*1000/(time-this.fpsTime);
					this.fpsTime = time;
					this.fpsCount = 0;
				}
				if( this.fps )
					graphics.drawFPS(this.fps);
			}
			else
				this.fpsTime = time;
		}
	}
};
	
/*
 * The animation loop method
 * The first time - called by onFinishImageLoad
 * see requestAnimFrame
 * do not call draw() directly - it needs the 'time' supplied by requestAnimFrame  
 */
function draw( time ){
	if( ImageLoader.isLoadingActiveImages() || !ImageLoader.loadedImages() )
		return;
	drawInstruments(time,connected,paused);
	if( config.drawInputControlBoxes )
		drawControlBoxes();
}

window.requestAnimFrame = (function() {
	if( !config.useVSync )
		return function(callback) { window.setTimeout(callback, 0); }; 
	return window.requestAnimationFrame || window.webkitRequestAnimationFrame || window.mozRequestAnimationFrame || window.oRequestAnimationFrame || window.msRequestAnimationFrame ||
	function(callback) {
		window.setTimeout(callback, 1000 / 60);//@todo - use proper shim
	};
})();

var log,error,assert,netLog,uiLog;
if( config.doLogging && !Cocoon){
	log = console.log.bind(console);
	error = console.error.bind(console);
	if( console.assert )
		assert = console.assert.bind(console);
	else{
		assert = function(condition, message) {
		if( config.doLogging && !condition) 
			console.log( "ASSERT: " + message );//@todo console.assert
		};
	} 
	netLog = console.log.bind(console);
	uiLog = console.log.bind(console);
}
else{
	assert = function(condition, message) {
		if( config.doLogging && !condition) 
			console.log( "ASSERT: " + message );//@todo console.assert
	};
	log = function(){
		if( config.doLogging )
			console.log.apply(console, arguments); 
	};
	error = function(){
		console.error.apply(console, arguments); 
	};
	netLog = function(){
		if( config.doLogging )
			console.log.apply(console, arguments); 
	};
	uiLog = function(){
		if( config.doLogging )
			console.log.apply(console, arguments);
	};
}