/* global config, dontExtrapolate, useOffscreenCanvas, stillInstruments, netLog, simulateStillInstruments:true, netSimulate, getTime, onDisconnect, connected:true, WebViewUI, netInterval:true, onConnect, onPing, onFailPing, startPingLoop, getInstrumentValue, time, extrapolateCustomThresholds, removeCycles, netEventNames, onReceiveData, onInvalidNetData */
/* exported netInit, netUpdate, endPingLoop, startPingLoop, netSetLocalValue, netGetValue, netGetRawValue, netExpireLocalValue, netLockLocalValue, netOnFinshDraw, netGetEvents */
/* jslint browser: true*/
/* jshint curly: false */

var //delay = 0,//@todo - server sends an ETA with each mesage (or the estimation can be done on the client side)
	extrapolateTimeThreshold = 1000,//don't extrapolate from very old data
	extrapolateValueThreshold = 0.25,// 1/4 of the value range, don't extrapolate if value changes too fast                                                                             
	lastData, prevData, prevExtrapolatedData = {}, lastExtrapolatedData = {}, /*serverStartTime,*/ serverTimeOffset,
	localData={}, localDataExpireTimes={}, valueTimes = {},
	isDeltaValue = {},
	netStopReconnect = false,
	printVersionMessage = true,
	//activeConnectionPingStopped = false,
	pings = [],//there can be only one connection, but many ping connections. The ping connections are stateless 
	pingSuccessAddr =[],
	//fetchedEventProperties = {},
	eventQueue = {};

var NETCOMMAND_Init = 1,
	NETLOCAL_Timeout = 1000;
	
//as per rfc6455
var NETCLOSE_OK = 1000,//1000 indicates a normal closure, meaning that the purpose for which the connection was established has been fulfilled.
	NETCLOSE_WrongType = 1003,//not text or not JSON
	NETCLOSE_ProtocolError = 1002;//right now - NETCOMMAND_Init has not been sent or protocol not supported
	//NETCLOSE_Abnormal = 3001,//connection was closed abnormally or the server cannot be reached
	//NETCLOSE_Timeout = 3002;//the connection attempt timed out (by our custom timing)
	//Status codes in the range 3000-3999 are reserved for use by libraries, frameworks, and applications.
	
var NETSTATE_Closed = 1,
	NETSTATE_Init = 2,
	NETSTATE_Opened = 3,
	NETSTATE_NoConnection = 4,
	NETSTATE_Error = 5,
	NETSTATE_Disconnected = 6;//unlike NETSTATE_Closed - will not try to reestablish the connection from onclose

var NETDATASET_All = 1,
	NETDATASET_C172 = 2,
	NETDATASET_Radio = 3,
	NETDATASET_Position = 4;
	
var NETTIMINGS_ReconnectInterval = 2000,//the time between connection failure and a new connect attempt
	NETTIMINGS_ConnectTimeout = 10000,//set to 0 for no timeout, cocoonjs has a bug where it's websocket connection attempts do not time out => use my own
	
	//the two timeouts below do not work because of a cocoonjs bug, disabled for now
	//NETTIMINGS_HTTPTimeout = 3000,//probably not needed at all
	NETTIMINGS_PingTimeout = 2000;//After this timeout onPingFail is issued, though the ping still has time to succeed before NETTIMINGS_ConnectTimeout
	

var netMessageInit = {//the first message sent to the server
	command: NETCOMMAND_Init,
	protocol_version: config.protocolVersion,
	dataset: NETDATASET_C172,
	user_agent: config.appName + " " +navigator.userAgent
	/*@todo: data_items=["adf_fail","engine_instruments_fail"]*/	
};

var netData = {
	//connection_state: NETSTATE_Closed
	//the rest of the properties are copied from the ACK message
};

if( useOffscreenCanvas ){
	//add all stillInstruments to dontExtrapolate
	for( var n=0; n<stillInstruments; n++ )
		dontExtrapolate[stillInstruments[n]] = true;
}


function netInit( addr, simStill ){
	netLog( "netinit: " + addr );  
	if( addr===null ){
		simulateStillInstruments = simStill;              
		netSimulate();
	}
	else
		netConnect( addr, 0 );
}

function checkVersion(responseText){
	var version = responseText.substring(3,responseText.length);
	if( version==="" )
		version = 3;
	version = parseInt(version);
	if( version==config.protocolVersion )
		return version;
	else{
		//output a message, but just once
		if( printVersionMessage ){
			printVersionMessage = false;//will be set to true the next time the user clicks connect
			if( version>config.protocolVersion )
				alert("Please upgrade " + config.appName + " to the latest version");
			else
				alert("Please upgrade the FSi Server to the latest version");
		}
		return 0;//error
	}
}

var activeConnectionData;
												
/*
 * Before connecting with websockets, try http to work around a bug in cocoonjs where websocket connections do not timeout
 */
function netConnect( addr, reconnectTryCount ){
	var url = (addr.indexOf(":")!=-1 ? addr : (addr +":"+ config.defaultPort));
	
	if( activeConnectionData ){
		if( activeConnectionData.address==addr ){
			netLog( "Already trying to establish connection to " + addr );
			return;
		}
		netWSDisconnect(addr);
	}
	
	if( !config.httpPing ){
		netWSConnect( addr, "ws://" + url, reconnectTryCount );
		return;
	}
	
	var time = getTime();
	
	var req = new XMLHttpRequest();
	req.onreadystatechange = function(){
		if( req.readyState == 4/* complete */ ){
			var fail = true;
			if( req.status==200 && req.responseText.substring(0,3).toLowerCase()=="fsi" ){
				var version = checkVersion(req.responseText);
				if( version ){
					fail = false;
					netLog( "Succesfull HTTP ping response from " + addr + ", version:" + version );
					netWSConnect( addr, "ws://" + url, reconnectTryCount );
				}
			}
			if( fail ){
				//failed - retry
				netLog( "HTTP request to " + addr + " failed, status:" + req.status );
				onConnectFail();
				var waitTime = NETTIMINGS_ReconnectInterval - (getTime() - time); 
				if( waitTime<1 )
					waitTime = 1;
				setTimeout( function(){
					if( !WebViewUI.autoConnectOn() )
						netConnect( addr, reconnectTryCount+1 );
				}, waitTime );
			}
		}
	};
	req.ontimeout = function () {
		//timed out - retry
		netLog( "Http connect timeout" );
		onConnectFail();
		setTimeout( function(){
			if( !WebViewUI.autoConnectOn() )
				netConnect( addr, reconnectTryCount+1 );
		}, 0 );
	};
	
	req.open("GET", "http://" + url, true);
	//req.timeout = NETTIMINGS_HTTPTimeout;
	req.send(null);
	
	function onConnectFail(){
		var pid = pingSuccessAddr.indexOf(addr);
		if( pid!=-1 )
			pingSuccessAddr.splice(pid,1);
		if( !activeConnectionData || activeConnectionData.address==addr )//try to avoid a bug probably caused by dangling old connection after another connection has been established
			onDisconnect( addr, true );
	}
}

function netWSDisconnect( addr ){
	if( activeConnectionData ){
		//we are connected to something
		/* if in the future I decide the get this back - take notice that the timeout calls netConnect again to close & try to reconnect
		if( addr==activeConnectionData.address )
			return;//already connecting or connected to the same address*/
		/*if( pingLooping && addr!=activeConnectionData.address && activeConnectionPingStopped ){
			pingAddress( activeConnectionData.address );//we do not ping the active connection, now that it's not active anymore - start pinging it
		}
		activeConnectionPingStopped = false;*/
		netInitiateDisconnect();//disconnect the currently active connection
		
	}
	connected = false;
	/*if( disconnectPings ){	
		//if addr is being pinged now - stop it
		for( var i=0; i<pings.length; i++ ){
			if( pings[i]!==null && pings[i].address==addr ){
				//don't close a ping which connection is not open - cocoonjs hates it
				if( pings[i].connection.readyState==1 ){//1==OPEN
					netLog( "Ping to"+addr +" interrupted by new connection to the same address" );
					pings[i].connection.close( NETCLOSE_OK, "Ping interrupted by connection" );
				}
				break;
			}
		}
	}*/
}

function netWSConnect( addr, cURL, reconnectTryCount ){
	addr = addr.trim();
	netStopReconnect = false;
	
	/*if( activeConnectionData ){
		//we are connected to something
		// if in the future I decide the get this back - take notice that the timeout calls netConnect again to close & try to reconnect
		//if( addr==activeConnectionData.address )
		//	return;//already connecting or connected to the same address
		netInitiateDisconnect();//disconnect the currently active connection
		if( pingLooping && addr!=activeConnectionData.address && activeConnectionPingStopped ){
			pingAddress( activeConnectionData.address );//we do not ping the active connection, now that it's not active anymore - start pinging it
		}
		activeConnectionPingStopped = false;
	}
	else*/{		
		//reset some values in case we were running local simulation	
		//connected = false;
		lastData = prevData = null;
		prevExtrapolatedData = {};
		lastExtrapolatedData = {};
		//messageNo = undefined;
	}
	
	if( !reconnectTryCount )
		netLog( "Connecting to "+addr);
	else
		netLog( "Reconnecting(#" + reconnectTryCount +") to "+addr);  
		
	var connection,
		error = false,
		//time = getTime(),
		connectionState = NETSTATE_Init;
  
	try{
		connection = new WebSocket( cURL );//, ['fsi'] );
	}
	catch( err ){
		netLog( "Invalid connect address : " + cURL + " (#" + reconnectTryCount +")" );
		error = true;
		onDisconnect( addr, false );
		return;
	}
	
	var connectionData = {
		connection : connection,
		address : addr,
		time: getTime(),
		disconnect : false,
		connectionState : connectionState
	};
	
	activeConnectionData = connectionData;

	connection.onopen = function( /*event*/ ){
		if( activeConnectionData!=connectionData ){
			netLog("A non-active connection was opened");
			return;//just in case
		}
		//console.log( connection.readyState );
		//setTimeout( function(){//use setTimeout to avoid a bug - the connection is still not open while in onopen
			netSend( netMessageInit, true );
		//},0);
	};
	
	connection.onerror = function (/*err*/) {
		if( connectionData.disconnect )
			connectionState = NETSTATE_Disconnected;
			
		netLog('WebSocket Error ');// + err); - apparently err contains no useful data, print the error from onclose
		if( connectionState==NETSTATE_Init )
			connectionState = NETSTATE_NoConnection;
		else
			connectionState = NETSTATE_Error;
		if( activeConnectionData==connectionData )
			activeConnectionData.connectionState = connectionState;
		error = true;                   
		netOnFail( addr, reconnectTryCount, true );
	};
	
	connection.onmessage = function(event){
		if( connectionData.disconnect )
			connectionState = NETSTATE_Disconnected;
			
		if( error || (connectionState!=NETSTATE_Opened && connectionState!=NETSTATE_Init) || connectionData!=activeConnectionData )
			return;
		
		var data, badJSON = false;	
		try{		
			data = JSON.parse(event.data);
		}
		catch( err ){
			netLog( event.data );
			if( connectionState==NETSTATE_Opened ){
				netLog( "Mallformed JSON received on active connection:" + err + "." );//@todo break connection if too many consecutive wrong messages are received
				if (typeof onInvalidNetData !== 'undefined') {
					onInvalidNetData(event.data);
				}
				badJSON = true;
				//we have an established connection, no error but ignore wrong messages 
			}
			else
				error = true;
		}
		if( !badJSON && !error ){
			//if( data && prevData && data.autopilot_altitude!=prevData.autopilot_altitude && data.autopilot_altitude!==undefined ){
			//	console.log("New autopilot_altitude data:" + data.autopilot_altitude );
			//}
			/*if( data.instrument_settings ){
				log( "New instrument data:" );
				log( JSON.stringify(data.instrument_settings) );
			}*/
			if( netProcessMessage( connection, connectionState, addr, data) ){
				connectionState = NETSTATE_Opened;
				if( activeConnectionData==connectionData )
					activeConnectionData.connectionState = connectionState;
			}
			else
				error = true;
		}
			
		if( error ){
			connectionState = NETSTATE_Closed;
			if( activeConnectionData==connectionData )
				activeConnectionData.connectionState = connectionState;
			netInitiateClose( connection, addr, NETCLOSE_WrongType, connectionState/*state*/==NETSTATE_Init?"JSON expected for init message.":"JSON expected."  );//close the connection
			netOnFail( addr, reconnectTryCount, true );
		}
		else{
			if( activeConnectionData==connectionData )
				activeConnectionData.connectionState = connectionState;
		}
	}; 
	 
	connection.onclose = function(event){//http://msdn.microsoft.com/en-us/library/ie/hh772748(v=vs.85).aspx
		if( connectionData.disconnect )
			connectionState = NETSTATE_Disconnected;
		
		if( activeConnectionData==connectionData )
			activeConnectionData.connectionState = connectionState;
			
		if( event.wasClean ){
			if( !error ){
				if( event.code==NETCLOSE_OK )
					netLog( "Connection to " + addr + "(#" + reconnectTryCount +") closed normally." );//well, not always?
				else
					netLog( "Connection " + addr + "(#" + reconnectTryCount +") closed abnormaly by the server with error#"+event.code+" :",event.reason );
			}
			else
				netLog( "Connection " + addr + "(#" + reconnectTryCount +") closed due to an error" );
		}
		else{
			if( connectionState==NETSTATE_NoConnection )
				netLog( "Could not connect to server " + addr + "(#" + reconnectTryCount +")." );
			else if( connectionState==NETSTATE_Error )
				netLog( "Connection with the server " + addr + "(#" + reconnectTryCount +") was lost." );
			else
				netLog( "Connection to " + addr +  "(#" + reconnectTryCount +") closed abnormaly with error#"+event.code+" :",event.reason  );
		}				
		if( !error )//if error - netOnClose has already been called from netOnFail
			netOnClose( addr, reconnectTryCount, connectionState!=NETSTATE_Disconnected );//if NETSTATE_Disconnected state - don't reconnect
		connectionState = NETSTATE_Closed;
		
		if( activeConnectionData==connectionData )
			activeConnectionData.connectionState = connectionState;
	};
	
	if( NETTIMINGS_ConnectTimeout ){
		setTimeout( function(){ 
			//use our own timeout because the default one takes ages on cocoonjs
			if( activeConnectionData==connectionData && activeConnectionData.connectionState==NETSTATE_Init ){
				//still the same connection and still no reply
				if( !WebViewUI.autoConnectOn() ){
					netLog( "Connection to " + addr + ", timed out - reconnecting");
					netConnect( addr, reconnectTryCount+1 );//this will disconnect before connecting
				}
				else
					netLog( "Connection to " + addr + ", timed out, not reconnecting because in auto mode");
			} 
		}, NETTIMINGS_ConnectTimeout );
	}
}

/*
 * Unlike netInitiateClose, netInitiateDisconnect
 * will not try to reestablish the connection
 * will not call onDisconnect
 * works only on the active connection  
 */  
function netInitiateDisconnect(){ 
	if( activeConnectionData ){ 
		netLog( "Initiating disconnect ");
		var connection = activeConnectionData.connection; 
		activeConnectionData.disconnect = true;		
		netResetConnection( activeConnectionData.address );//activeConnectionData will become null, but we have another pointer in netConnect's connectionData
		connection.close( NETCLOSE_OK, "Closing normally" );
	}
}

function netInitiateClose( connection, addr, code, message ){	
	netResetConnection( addr );
	if( code==NETCLOSE_OK || code===undefined )
		netLog( "Connection to " + addr + " closing by client normally" + ((message===undefined)?".":(" : "+message+".")) );
	else
		netLog( "Closing connection to " + addr + " with error #" +code+ " : " + message );
	connection.close( code, message );
}

function netOnFail( addr, reconnectTryCount, reconnect ){
	var pid = pingSuccessAddr.indexOf(addr);
	if( pid!=-1 )
		pingSuccessAddr.splice(pid,1);
	netOnClose( addr, reconnectTryCount, reconnect && !netStopReconnect);
	netStopReconnect = false;
}

function netOnClose( addr, reconnectTryCount, reconnect ){
	connected = false;
	var time;
	if( activeConnectionData )
		time = activeConnectionData.time; 
	var closingCurrentConnection = netResetConnection( addr );	
				
	onDisconnect( addr, reconnect );
	
	var reconnectAddr;
	if( closingCurrentConnection && reconnect ){
		reconnectTryCount = (reconnectTryCount===undefined)? 1:(reconnectTryCount+1);
		if( pingSuccessAddr.length>0 && WebViewUI.autoConnectOn() && reconnectTryCount>1 )
			reconnectAddr = pingSuccessAddr[0];//@todo - round robin
		else
			reconnectAddr = addr;
		
		//measure time passed since initiating the connect and substract from the interval
		var waitTime = NETTIMINGS_ReconnectInterval - (getTime() - time); 
		if( waitTime<1 )
			waitTime = 1;
		setTimeout( function(){
			//check if a reconnect is still needed
			if( !activeConnectionData || (activeConnectionData.address==reconnectAddr && activeConnectionData.connectionState!=NETSTATE_Init && activeConnectionData.connectionState!=NETSTATE_Opened) )
				//reconnect if no active connection or the current connection is to the same address and it's not open or initializing
				if( !WebViewUI.autoConnectOn() )
					netConnect( reconnectAddr, reconnectTryCount, true );
		}, waitTime );
	}
	
	if( pingLooping && reconnectAddr!=addr ){
		//the currently connecting address is not pingelooped,
		//as addr is not being connected anymore - ping it
		pingAddress( addr );
		//@todo - maybe try to ping all addresses if you have problems
	}
}

function netResetConnection( addr ){
	if( activeConnectionData && activeConnectionData.address==addr ){
		//reset the active connection
		//connected = false;
		//activeConnectionPingStopped = false;
		activeConnectionData = null;
		lastData = prevData = null;
		prevExtrapolatedData = {};
		lastExtrapolatedData = {};
		//messageNo = undefined;                                                                                               
		return true;                                                                                                         
	}
	return false;
}

function netSend( data, isInit ){
	
	if( activeConnectionData && ( (isInit && activeConnectionData.connectionState==NETSTATE_Init ) || (!isInit && activeConnectionData.connectionState==NETSTATE_Opened) ) ){
		activeConnectionData.connection.send( JSON.stringify(data) );
		netLog('sending packet:\n'+JSON.stringify(data));
	}
	else
		netLog('Not sending packet, connection not established');
}

//var debugCount = 0, prevServerTime=0, prevTime=0, prevWC=0;
function netProcessMessage( connection, connectionState, addr, data ){  
	switch( data.command ){			
		case undefined://no command, a regular message
			if( connectionState==NETSTATE_Init ){
				netInitiateClose( connection, addr, NETCLOSE_ProtocolError, "Connection has not been properly initialized, Init command expected." );
				return false;
			}
			else if( connectionState==NETSTATE_Opened ){
				processLocalValues(data);
				
				var prePrevData = prevData;
				prevData = lastData;
				lastData = data; 
				
				/*if( serverStartTime==undefined || !prevData ){
					serverStartTime = data.server_time;
				}
				data.server_time -= serverStartTime;//not necessary, but more convenient for debugging*/

				if( !prevData ){//this is the first message with sim data
					serverTimeOffset = getTime()-data.server_time;
					console.log( "Initial time, local: " + getTime()+ " server:" +  data.server_time + " calculated:" + serverToLocalTime(data.server_time) );
					if( data.netInterval )
						netInterval = data.netInterval;
					connected = true;
					onConnect( addr, "", "" );//netData.server_name, netData.flight_sim );//everything starts when the message AFTER the init command is received			
					prevData = lastData;
			    }
				if( 
					(prePrevData && prevData &&
					Math.abs(serverToLocalTime(data.server_time)-getTime())>1000 &&
					Math.abs(serverToLocalTime(prevData.server_time)-getTime())>1000 &&
					Math.abs(serverToLocalTime(prePrevData.server_time)-getTime())>1000 )
				){
					//the time is out of sync, reset the offset
					//serverStartTime = data.server_time + serverStartTime;
					//data.server_time = 0;//start from 0, see serverStartTime
					serverTimeOffset = getTime()-data.server_time;
				}
				//console.log(data.server_time);
				//console.log( "time, local: " + getTime()+ " server:" +  data.server_time + " calculated:" + serverToLocalTime(data.server_time) );
			    /*else*/{ 
					/*if( debugCount<200 ){
			    		//console.log( "tSrv:" + ((data.server_time-prevServerTime)|0) + ", tRcv:" + ((t-prevTime)|0) + ', compass:'+(data.whiskey_compass-prevWC)*1000 );
						//console.log( 'compass:'+(data.whiskey_compass-prevWC)*1000 );
						prevWC = data.whiskey_compass;
						prevTime = t;
						prevServerTime = data.server_time;
			    		//log( JSON.stringify(data) );			    		
			    		debugCount++;
			    	}*/
			    	var sto = getTime()-data.server_time;
			    	if( sto<serverTimeOffset ){
			    		//pick the smallest serverTimeOffset possible, e.g. use the most recent packet ever received as a reference
			    		serverTimeOffset = sto;
			    	}
			    	//take a snapshot of the lastExtrapolatedData in prevExtrapolatedData	
					for( var prop in lastExtrapolatedData )
						prevExtrapolatedData[prop] = lastExtrapolatedData[prop];
					
					//collect all events
					for( prop in netEventNames ){
						var id = data[prop+'_id'];
						if( id ){
							var value = data[prop+'_value'];
							eventQueue[id] = {
								value:value,
								name: prop
							};
							delete data[prop+'_id'];
							delete data[prop+'_value'];
						}
					}
					
					//copy all missing properties from prevData             
					for( prop in prevData ){
						if( lastData[prop]===undefined && !(prop in occasionalProperties) )
							lastData[prop] = prevData[prop];
					}
					
					//copy all missing properties from prevData             
					/*for( prop in prevData ){
						//( data[prop]===undefined && !netEventNames[prop] )//copy only missing properties that are not an event
						//	data[prop] = prevData[prop];
						if( data[prop]===undefined ){
						   if( netEventNames[prop] ){
							   //this is an event property
								if( !fetchedEventProperties[prop] ){
									//the event has not been fetched yet, copy its properties it to the new packet so it can be fetched later
									data[prop] = prevData[prop];
								}
							   else
								   fetchedEventProperties[prop] = false;//the event has been fetched and it's properties will not be copied to the new packet
							}
							else
								data[prop] = prevData[prop];
						}
					}*/
					
					if( typeof onReceiveData !== 'undefined' )
						onReceiveData(data);
				}
			}			
			//else - receiveing last data sent after initiating connection close - ignore it
			break;
		case NETCOMMAND_Init:
			connectionState = activeConnectionData.connectionState = NETSTATE_Opened;
			//copy the message data to netData except for command
			for( var p in data )
				if( p!="command")
		        	netData[p] = data[p];
		    netLog( "Handshake succesfull");
			break;
		default:
			if( connectionState==NETSTATE_Init ){
				netInitiateClose( connection, addr, NETCLOSE_ProtocolError, "Unknown command on connection initialization, Init command expected." );
				return false;
			}
			else
				netLog("Unknown command on active connection, continuing.");
			break;
	}
	return true;
}

function netOnFinshDraw(){
	for( var i in eventQueue ){
		if( eventQueue[i].fetched )
			delete eventQueue[i];
	}
	/*for( var e in fetchedEventProperties ){
		if( fetchedEventProperties[e] ){
			//the event has been fetched,
			//make sure the event will not be fetched in the next draw frame
			if( prevData )
				delete prevData[e];
			if( lastData )
				delete lastData[e];
			delete fetchedEventProperties[e];
		}
	}*/
}

var ignoredDeltaReplies = {};
function processLocalValues(data){
	//delete all expired local data
	var t = getTime();
	for( var key in localDataExpireTimes ){
		if( localDataExpireTimes[key]<t ){
			netLog('The local value of ' + key + ' has expired');
			if( !(key in data) && ignoredDeltaReplies[key] ){
				//to make 100% sure we're in sync with the server - use the last ignored value after done sending deltas
				netLog('Use last received value for '+ key +' : '+ignoredDeltaReplies[key]);//@debug
				lastExtrapolatedData[key] = lastData[key] = prevData[key] = ignoredDeltaReplies[key];
			}
			delete localData[key];
			delete localDataExpireTimes[key];
		}
		else{
			if( (key in data) && isDeltaValue[key+'_delta'] )
				ignoredDeltaReplies[key] = data[key];//when sending deltas - store the last reply and set it as final value when done to make sure we're synced
			//use the locally set value, ignore the one sent by the server
			data[key] = localData[key];
			if( prevData )
				prevData[key] = localData[key];
			if( lastData )
				lastData[key] = localData[key];
			if( lastExtrapolatedData )
				lastExtrapolatedData[key] = localData[key];
		}
	}
	/*if( "gyro_drift_angle" in data ){
		if( !localDataExpireTimes.gyro_drift_angle  )
			console.log('-> received gyro_drift_angle:'+data.gyro_drift_angle);
		else
			console.log('ignoring gyro_drift_angle:'+data.gyro_drift_angle);
	}*/
}

//@todo pingLog
var pingId = 1;
function netPing( addr ){
	if( config.httpPing ){
		netHTTPPing( addr );
		return;
	}
	/*if( activeConnectionData && addr==activeConnectionData.address ){//&& (connectionState==NETSTATE_Opened || connectionState==NETSTATE_Init) ){
		if( config.logPingAttempts ) 
			netLog( "Don't try to ping an active connection:" + addr+" ("+id+")" );
		return;
	}*/
	
	var success = false,
		interrupted = false,
		id = pingId,
		time = getTime();
	pingId++;
	
	for( var i=0; i<pings.length; i++ ){
		if( pings[i]!==null && pings[i].address==addr ){
			if( config.logPingAttempts ) 
				netLog( addr+" ("+id+") is already ongoind ping" );
			return;//currently this addr is ongoing ping
		}
	}
		
	if( config.logPingAttempts )	
		netLog( "Pinging "+addr+" ("+id+")" );
	
	var pingURL = "ws://" + (addr.indexOf(":")!=-1 ? addr : (addr +":"+ config.defaultPort)),
		pingConnection;
	try{
		pingConnection = new WebSocket( pingURL );//, ['fsi'] );
	}
	catch( err ){
		if( config.logPingAttempts ) 
			netLog( "Invalid ping url : " + pingURL+" ("+id+")" );
		return;
	}
	
	var index;
    for( index=0; index<pings.length; index++ ){
    	if( pings[index]===null ){
    		pings[index] = {connection:pingConnection, address:addr};
    		break;			    	    	
		}
	}	 
	if( index==pings.length )
		pings.push( {connection:pingConnection, address:addr} );
		
	pingConnection.onopen = function( /*event*/ ){
		if( config.logPingAttempts ) 
			netLog( 'Ping connection to ' + addr + ' opened.'+" ("+id+")");
		/*if( disconnectPings && activeConnectionData && addr==activeConnectionData.address  ){
			interrupted = true;
			if( config.logPingAttempts ) 
				netLog( "Ping to"+addr +"."+" ("+id+")"+" interrupted by connection to the same address when ping connection was opened" );
			pingConnection.close( NETCLOSE_OK, "Ping interrupted by connection" );
		}*/
		//don't declare a success yet, waiting for the data packet
	};
		
	var error = false;
	pingConnection.onmessage = function(event){
		if( error || interrupted )
			return;
			
		if( success ){
			netLog( "Unnecesary packet received as ping responce from "  + addr+" ("+id+")" );
			return;
		}
			
		var data;
		try{
			data = JSON.parse(event.data);
		}
		catch( err ){
			netLog( "Mallformed ping responce received from " + addr+" ("+id+")" );
			error = true;
		}
		
		if( !error ){
			if( data.command===undefined || data.command!=NETCOMMAND_Init ){
				if( config.logPingAttempts ) 
					netLog("Non-init reply received when pinging " + addr +"."+" ("+id+")");
				pingConnection.close( NETCLOSE_ProtocolError, "First message must be an Init" );
				error = true;  
			}
			else{
				if( config.logPingAttempts ) 
					netLog("Received ping responce from " + addr +"."+" ("+id+")");
				if( pingSuccessAddr.indexOf(addr)==-1 )
					pingSuccessAddr.push(addr);
				
				success = true;
				
				if( activeConnectionData && addr==activeConnectionData.address ){
					/*if( disconnectPings ){
						interrupted = true;
						if( config.logPingAttempts ) 
							netLog( "Ping to"+addr +"."+" ("+id+")"+" interrupted by connection to the same address when receiving init data" );
						pingConnection.close( NETCLOSE_OK, "Ping interrupted by connection" );
					}*/
				} 
				else if( pingLooping )//it may be possible that by the time the ping is confirmed - onPing is not needed anymore
					onPing( addr, "", "" );//data.server_name, data.flight_sim );
				
				//pingConnection.close( NETCLOSE_OK, "Pinged succesfully" );
			}
		}
		else
			pingConnection.close( NETCLOSE_WrongType, "Mallformed ping responce"+" ("+id+")" );
		
		if( error || success ){
			if( pings[index] ){//just in case - make sure this ping has not been alrady repeated
				pings[index] = null;
				repeatPing( addr, time, id );
			}
		}
			 
		if( error ){
			pingFail( addr );
			//repeatPing( addr, time, id ); - will be called by onClose!!!
		}
	};
	
	pingConnection.onclose = function(event){
		if( error || interrupted )
			return;                                                                                          	
			
		if( !event.wasClean || !success ){
			error = true;
			if( config.logPingAttempts ) 
				netLog( "Ping to " + addr + " failed" + (event.wasClean?"":": not a clean close")+" ("+id+")" );
			pingFail( addr );
		}
		else{
			if( config.logPingAttempts )
				netLog( "Ping to " + addr + " ended normaly"+" ("+id+")" );
		}
		
		//fire a new ping
		if( pings[index] ){
			pings[index] = null;
			repeatPing( addr, time, id );//will wait at least NETTIMINGS_ReconnectInterval between repings
		}
	};
	
	setTimeout( function(){ 
		//use our own timeout because the default one takes ages,
		//this timeout does not close the connection yet, but onFailPing is issued
		//the ping is still ongoing and has time to succeed before onclose (which often takes a lot of time)	
		if( !success && !error && !interrupted ){
			if( config.logPingAttempts ) 
				netLog( "Ping to " + addr + ", initial time out"+" ("+id+")");
			pingFail( addr );
		}
	}, NETTIMINGS_PingTimeout );
	
	if( NETTIMINGS_ConnectTimeout ){
		setTimeout( function(){ 
			//use our own timeout because the default one takes ages on cocoonjs
			if( !success && !error && !interrupted ){
				if( config.logPingAttempts ) 
					netLog( "Ping to " + addr + ", final time out"+" ("+id+"), disconnecting");
				pingConnection.close( NETCLOSE_OK, "Ping interrupted by timeout" );
				
				//repeatPing not needed - will be done in onclose
				//repeatPing( addr, time, id );//will wait at least NETTIMINGS_ReconnectInterval between repings
			}
		}, NETTIMINGS_ConnectTimeout );
	}
}

function netHTTPPing( addr ){
	/*if( activeConnectionData && addr==activeConnectionData.address ){//&& (connectionState==NETSTATE_Opened || connectionState==NETSTATE_Init) ){
		if( config.logPingAttempts ) 
			netLog( "Don't try to ping an active connection:" + addr+" ("+id+")" );
		return;
	}*/
	
	var id = pingId,
		time = getTime();
	pingId++;
	
	for( var i=0; i<pings.length; i++ ){
		if( pings[i]!==null && pings[i].address==addr ){
			if( config.logPingAttempts ) 
				netLog( addr+" ("+id+") is already ongoind ping" );
			return;//currently this addr is ongoing ping
		}
	}
		
	if( config.logPingAttempts )	
		netLog( "Pinging "+addr+" ("+id+")" );
	
	var req = new XMLHttpRequest(),
		pingURL = (addr.indexOf(":")!=-1 ? addr : (addr +":"+ config.defaultPort));
	
	//find an empty slot...
	var index;
    for( index=0; index<pings.length; index++ ){
    	if( pings[index]===null ){
    		pings[index] = {request:req, address:addr};
    		break;			    	    	
		}
	}
	//... or add a new one
	if( index==pings.length )
		pings.push( {request:req, address:addr} );
	
	req.onreadystatechange = function(){
		if( req.readyState == 4/* complete */ ){
			var fail = true;
			if( req.status==200 && req.responseText.substring(0,3).toLowerCase()=="fsi" ){ 
				var version = checkVersion(req.responseText);
				if( version ){
					fail = false;
					//success
					if( config.logPingAttempts ) 
						netLog( 'Ping request to ' + addr + ' is succesfull.'+" ("+id+")");

					if( pingSuccessAddr.indexOf(addr)==-1 )
						pingSuccessAddr.push(addr);

					/*if( disconnectPings && activeConnectionData && addr==activeConnectionData.address  ){
						interrupted = true;
						if( config.logPingAttempts ) 
							netLog( "Ping to"+addr +"."+" ("+id+")"+" interrupted by connection to the same address while waiting for ping request" );
					}
					else*/
					if( pingLooping )//it may be possible that by the time the ping is confirmed - onPing is not needed anymore
						onPing( addr, "", "" );//data.server_name, data.flight_sim );
				}
			}
			if( fail ){
				//failed - retry
				if( config.logPingAttempts ) 
					netLog( "HTTP ping to " + addr + " failed, status:" + req.status );
				pingFail( addr );
			}
			if( pings[index] ){//just in case - make sure this ping has not been alrady repeated
				pings[index] = null;
				repeatPing( addr, time, id );//always repeat, no mater if success or fail
			}
		}
		else
			netLog( 'HTTP request to ' + addr + ' , readyState=' + req.readyState );
	};
	
	req.ontimeout = function () {
		//timed out
		if( config.logPingAttempts )
			netLog( "Http ping timeout" );
		if( pings[index] ){//just in case - make sure this ping has not been alrady repeated
			pings[index] = null;
			pingFail( addr );
		}
		repeatPing( addr, time, id );
	};
	
	req.open("GET", "http://" + pingURL, true);
	if( NETTIMINGS_PingTimeout )
		req.timeout = NETTIMINGS_PingTimeout;
	req.send(null);
}

function repeatPing( addr, time, id ){
	if( !pingLooping ){
		if( config.logPingAttempts ) 
			netLog( "Cancel ping to " + addr +" ("+id+")" );
		return;
	}
	/*if( activeConnectionData && addr==activeConnectionData.address ){
		activeConnectionPingStopped = true;
		return;
	}*/
	var waitTime = NETTIMINGS_ReconnectInterval - (getTime() - time); 
	if( waitTime<1 )
		waitTime = 1;
	if( config.logPingAttempts ) 
		netLog( "rePing of " + addr +" ("+id+") in " + Math.round(waitTime*100)/100000 + " sec"  );
	setTimeout( function(){ netPing(addr);}, waitTime );
}
			
function pingFail( addr ){
	var pid = pingSuccessAddr.indexOf(addr);
	if( pid!=-1 )
		pingSuccessAddr.splice(pid,1);
	if( pingLooping )
		onFailPing( addr );
}

var pingLooping = false;
function startPingLoop(){ 
	if( pingLooping || !config.doPingLoops )
		return;
	pingLooping = true;//each ping will be repeated after onclose
	
	if( config.logPingAttempts )
		netLog( "Starting ping loop" );

	//Ping all addresses from the history, that are not currently being pinged or connected	
	if( localStorage.connectionHistory ){
		//fire pings to all addresses except addr
		var history = JSON.parse(localStorage.connectionHistory);
		if( history.length>0 ){
			for( var i=0; i<history.length;i++)
				pingAddress(history[i]);
		}
	}
}

function endPingLoop(){
	pingLooping = false;
}

function pingAddress( addr ){
	if( !pingLooping )
		return;
	//if( !activeConnectionData || addr!=activeConnectionData.address ){//don't ping the currently connecting address
		for( var i=0; i<pings.length; i++ )
			if( pings[i]!==null && pings[i].address==addr )
				return;//currently this addr is ongoing ping, don't ping it
		netPing( addr );
	//}
	//else
	//	activeConnectionPingStopped = true;
}

/*@todo - move to base instrument*/
function netUpdate( ent ){
	var value,n,p;
	if( ent.listen ){
		if( !(ent.listen instanceof Array) ){
			value = getInstrumentValue(ent, ent.listen);
			if( !isNaN(value) && value!==undefined )
				ent.setValue( value, ent.listen );
		}
		else
			for( n=0; n<ent.listen.length; n++ ){
				value = getInstrumentValue(ent, ent.listen[n]);
				if( !isNaN(value) && value!==undefined )
					ent.setValue( value, ent.listen[n] );
			}
	}
	if( ent.instruments )
		for( n=0; n<ent.instruments.length; n++ )
			netUpdate( ent.instruments[n] );
	
	if( ent.parameterProperties )
		for( p in ent.parameterProperties )
			netUpdate( ent.parameterProperties[p] );
	
	if( ent.clipBoxAbsolute ){
		for( var i=0; i<ent.clipBoxAbsolute.length; i++ ){
			var cb = ent.clipBoxAbsolute[i];
			if( cb.parameterProperties ){
				for( var p in cb.parameterProperties )
					netUpdate( cb.parameterProperties[p] );
			}
		}
	}
}

var netSendMessage = {}, netLastSendTime = 0, netSendWaiting = false;
function netSetLocalValue( value, key, dontSend, isDelta ){
	if( value===undefined || key===undefined )
		return;
	console.log( 'localval ' + key + '='+value);
	function sendValue(t){
		//netLog( "time since last:" + (t-netLastSendTime) );
		netSendWaiting = false;
		netSend( netSendMessage );
		netSendMessage = {};
		netLastSendTime = t;
	}
	var now = getTime();
	if( isDelta ){
		isDeltaValue[key] = true;
		if( localData[key] )
			localData[key] += value;
		else
			localData[key] = value;
		//console.log( 'delta:' + localData[key] );
	}
	else{
		localData[key] = value;
		if( lastExtrapolatedData )
			lastExtrapolatedData[key] = value;
		if( lastData )
			lastData[key] = value;
		if( prevData )
			prevData[key] = value;
	}
	
	if( !dontSend && activeConnectionData.connectionState==NETSTATE_Opened ){
		netSendMessage[key] = localData[key];
		if( !config.netSendInterval || now-netLastSendTime>config.netSendInterval ){
			sendValue(now);
			if( isDelta )
				localData[key]=0;
		}
		else{
			if( !netSendWaiting ){
				netSendWaiting = true;
				setTimeout( function(){
					if( netSendWaiting ){
						sendValue(getTime());
						if( isDelta )
							localData[key]=0;
					}
				}, config.netSendInterval-(now-netLastSendTime)  );
			}	
		}
	}
}

function netLockLocalValue( key ){
	netExpireLocalValue( key, getTime() + 1000*60*60*24 );//24hours is plenty
}

function netExpireLocalValue( key, expireTime ){
	if( localData[key]!==undefined ){
		log( "Expire local value " + key );
		if( expireTime===undefined )
			expireTime = NETLOCAL_Timeout; 
		localDataExpireTimes[key] = getTime()+expireTime;
	}
}

function serverToLocalTime( stime ){
	return stime + serverTimeOffset;
}

/*function netGetEventValue(key){
	if( lastData ){
		fetchedEventProperties[key] = true;
		return lastData[key];
	}
}*/

function netGetEvents(name){
	var ret;
	for( var id in eventQueue ){
		if( eventQueue[id].name==name ){
			if( !ret )
				ret = {};
			ret[id] = eventQueue[id].value;
			eventQueue[id].fetched = true;
		}
	}
	return ret;
}

/* @todo - netGetValue with just one parameter? */
function netGetRawValue(key){
	if( key in localData )
		return localData[key];
	if( lastData )
		return lastData[key];
}

//var prev=0; 
function netGetValue( ent, key ){
	if( key in localData ){
		lastExtrapolatedData[key] = localData[key];
		lastExtrapolatedData.time = time;
		return localData[key];
	}
	if( !lastData )
		return;
	var value = lastData[key];
	if( netEventNames[key] && lastData )
		return lastData[key];
	if( !lastData )
		return 0;
	if( valueTimes[key]==time )
		return lastExtrapolatedData[key];//do not extrapolate the more than once
	valueTimes[key]=time;
		
	var	inputRange = ent.getValueProperty("inputValueRange",key),
		repeat = inputRange && ent.getValueProperty("repeat",key),
		nocycle = ent.getValueProperty("nocycle",key); 
	if( !config.dontExtrapolateValues && value!==undefined && prevData && !dontExtrapolate[key] &&
	   	(time-serverToLocalTime( prevData.server_time ))<extrapolateTimeThreshold && //don't extrapolate if too much time has passed
	   	prevData[key]!=lastData[key] &&//do not extrapolated if data stays the same
	   	(prevExtrapolatedData[key]!=prevData[key] || prevExtrapolatedData[key]!=lastExtrapolatedData[key] ) //extrapolate only values that change all the time, do not extrapolate if static value has just changed for the first time to avoid overshooting. Happens with autopilot_heading_magnetic
	  ){
		//extrapolate the value
		var prevValue = prevData[key],
			prevDataTime = serverToLocalTime( prevData.server_time ),
			lastDataTime = serverToLocalTime( lastData.server_time ),
			//prevDataTime = lastDataTime - netInterval,
			nextDataTime = lastDataTime + netInterval;//the expected arrival of the next packet  
			
		if( repeat && !nocycle ){
				//remove any cycles between value and prevValue				
				prevValue = removeCycles( prevValue, value, inputRange[1]-inputRange[0] );              	
		}
		var abruptChange = false;
		if( inputRange || extrapolateCustomThresholds[key] ){
			var maxDelta = extrapolateCustomThresholds[key]?
								extrapolateCustomThresholds[key] :
								Math.abs(extrapolateValueThreshold*(inputRange[1]-inputRange[0]));
			abruptChange = Math.abs(value-prevValue)>maxDelta;
		}
		if( !inputRange || !abruptChange  ){ 
			var extrapolateDirection = (value-prevValue)/(lastDataTime-prevDataTime),
				extrapolateStart = (repeat && !nocycle) ? removeCycles( prevExtrapolatedData[key], value, inputRange[1]-inputRange[0] ) : prevExtrapolatedData[key]; 
			if( (extrapolateStart===undefined || time>nextDataTime) ){//when time==nextDataTime - the extrapolate target has been reached continue to extrapolate with the regular algho
				value += extrapolateDirection*(time-lastDataTime);
			}
			else{
				//when changing from one value/prevValue pair to a new one, the extrapolated value changes abruptly
				//to avoid this extrapolate starting from the last known position towards the estimated target after one network tick
				var extrapolateTarget = value+extrapolateDirection*(netInterval),
					extrapolateStartTime = prevExtrapolatedData.time; 
				value = extrapolateStart + (extrapolateTarget-extrapolateStart)*(time-extrapolateStartTime)/(nextDataTime-extrapolateStartTime);    
			}
		}
		else if( abruptChange && prevExtrapolatedData[key] )
			prevExtrapolatedData[key] = value;//just for some border cases, we probably won't do extrapolation with this prevExtrapolatedData anyway  		
	}				
	if( inputRange && !repeat ){
		if( inputRange[1]>inputRange[0] ){
			value = Math.min( value, inputRange[1] );
			value = Math.max( value, inputRange[0] );
		}
		else{
			value = Math.min( value, inputRange[0] );
			value = Math.max( value, inputRange[1] );
		}
	} 
	lastExtrapolatedData[key] = value;
	lastExtrapolatedData.time = time;
	
	return value;
}