/* global Cocoon, uiLog, log, connected, netStopReconnect:true, netInit, useOffscreenCanvas:true, config, isFSIReady, startPingLoop, getTime, endPingLoop, connectDialogSettings */
/* exported WebViewUI */
/* jslint browser: true*/
/* jshint curly: false */

var WebViewUI = {
	init: function(){
		var suspended = false, reloadUI = false, showAfterReload = false,
			showingUI, showingNews, postponeHide, isWebViewReady, isUIReady, webViewInitReceived;
		var autoConnect = (localStorage.getItem("autoConnect")!==null ) ? ( localStorage.autoConnect=="true" ) : false,
			waitAutoConnect, autoConnecting, connectedOrAttemptingConnect;


		function initVars(){
			showingUI = undefined;
			showingNews = false;
			postponeHide = false;
			isWebViewReady = false;
			isUIReady=false;
			webViewInitReceived = false;
			waitAutoConnect = false;
			autoConnecting = false;
			connectedOrAttemptingConnect = false;
		}

		initVars();
		
		connectDialogSettings.reconnectOnStart = (localStorage.getItem("reconnectOnStart")!==null ) ? ( localStorage.reconnectOnStart=="true" ) : false;
		delete localStorage.reconnectOnStart;
		
		if( !Cocoon )		
			if( (localStorage.getItem("autoConnect")===null ) )
				localStorage.autoConnect = autoConnect;
			window.addEventListener("message", 	function(e){
				/*this*/WebViewUI.onMessage( e.data.command, e.data.value );
			}, false);//the html ui is in an iframe
		//else - ui.js will use the cocoonJS API to call onMessage
		

		Cocoon.App.onSuspended.addEventListener(function(){
			console.log("On suspend");
			suspended = true;
			//if( showingUI && showingNews ){
				Cocoon.App.hideTheWebView();//we must hide the webview so it does not hang when we come back (A cocoonjs bug)
			//}
		});

		Cocoon.App.onActivated.addEventListener(function(){
			if( suspended ){
				suspended = false;//avoid a cocoonjs bug where onActivated is called twice
				console.log("On activated");
				//if( showingUI && showingNews ){
					WebViewUI.reloadWebView();
				//}
			}
		});

		this.reloadWebView = function(){
			showAfterReload = showingUI;
			log("Reload web view, showingUI:"+showingUI);
			initVars();
			reloadUI = true;
			Cocoon.App.reloadWebView();//we must reload the webview so it does not hang when we show it (A cocoonjs bug)
			if( showAfterReload )
				Cocoon.App.showTheWebView();
			else
				Cocoon.App.hideTheWebView();
			WebViewUI.sendMessage( "reload-ui", autoConnect );
		}
		
		this.isShown = function(){return showingUI;};
		this.isReady = function(){return isUIReady;};
		this.autoConnectOn = function(){return autoConnect;};

		/*
		 * Receive a message from the UI window
		 */   
		this.onMessage = function(command, value){
			uiLog( "(" + new Date().getTime() +") Main: Received message, command=",command,", value=",JSON.stringify(value) );
			switch( command ){
				case "init":
					webViewInitReceived = true;
					startWebView();
				break;
				case "connect":
					var addr = value.address;
					if( waitAutoConnect ){
						if( connected )
							break;
						waitAutoConnect = false;
						if( addr==="" )
							addr = getLastConnectAddr();
					}
					if( addr!=="" ){
						connectedOrAttemptingConnect = true;	
						netStopReconnect = false;
						netInit( addr );
						autoConnecting = value.autoConnecting;//set whether the connection was initiated by the Auto connect mode
					}
					else{
						connectedOrAttemptingConnect = false;				
						if( !connected && autoConnecting )
							netStopReconnect = true;//the last connection was initiated by the Auto connect mode - stop trying to connect
						autoConnecting = false;
					}

				break;
				case "hide-ui":
					uiLog( "Hiding UI on hide-ui message");
					this.show( false );
				break;
				case "show-news":
					uiLog("Show news:"+value + " postponeHide=" + postponeHide );
					showingNews = value;
					if( showingNews ){
						if( reloadUI ){
							reloadUI = false;
							uiLog( "Hiding UI on news reload after suspend/activate");
							this.show( false );//will get hidden after news are hidden
						}
						else{
							if( !showingUI ){
								uiLog( "Showing UI for the news");
								this.show(true);
								postponeHide = true;//hide when closing the news
							}
						}
					}
					else{
						if( postponeHide ){
							postponeHide = false;
							if( connected ){
								uiLog( "Hiding UI, postponed");
								this.show( false );
							}
						}
					}
				break;
				case "auto":
					log("auto:",autoConnect);
					localStorage.autoConnect = autoConnect = value;
					if( autoConnect ){
						//endPingLoop();//we will try to connect (which also pings the addresses, but the ping loop is not needed)
						waitAutoConnect = true;//connect message is sent right after the 'auto' message
					}
				break;
				case "open-url":
					Cocoon.App.openURL( value );
				break;
			}
		};
		
		/*
		 * Send a message to the UI window
		 */           
		this.sendMessage = function( comm, val ){
			if( typeof val=="undefined" )
				val = null;
			if( !Cocoon )
				document.getElementsByTagName('iframe')[0].contentWindow.postMessage( {command:comm, value:val}, '*' );
			else{

				var fwd_message = "onMessage({data:{command:'"+comm+"',value:" + JSON.stringify(val) +"}});";
				uiLog("Forwarding message: " + fwd_message );
				var t = getTime();
				Cocoon.App.forwardAsync(fwd_message);
				uiLog("WebViewUI.SendMessage is done, t=" + (getTime()-t) );
			}
		};
		
		this.show = function( on ){
			uiLog("Call webview.show("+on+") reload=" +reloadUI +" connected="+connected + " news="+showingNews );
			if( !on && showingNews ){
				uiLog("Postponing hide till news are closed");
				postponeHide = true;//will hide it later when news are closed
				return;
			}
			postponeHide = false;
			var url = "common//ui.html";//?" + serializeToURL(connectDialogSettings);
			if( showingUI===undefined ){
				if( !Cocoon )
					document.getElementsByTagName('iframe')[0].src = url+"#iframe";//don't set the url in the source because ot will start loading immediately
				else{
					Cocoon.App.WebView.on("load",{//http://support.ludei.com/hc/en-us/articles/200807797-Canvas-Webview
						 success: function() {
							uiLog( "load WebView success" );
							isWebViewReady = true;
							startWebView();			 
							if( reloadUI && !showAfterReload ){//connected && !showingNews ){
								uiLog("onload: not showing the WebView UI because reloading and UI is not needed");
								//Cocoon.App.hideTheWebView();//just in case
								showingUI = false;
							}
							else{
								uiLog("onload: show the WebView UI");
								showingUI = false;//temporary, the next line will show it
								WebViewUI.show(true);// Show the webview. By default, the webview is hidden.
							}
							reloadUI = false;
						 },
						 error: function() {
							 console.error("Cannot show the Webview for some reason :/");
							 uiLog(JSON.strigify(arguments));
						 }
					});
					Cocoon.App.loadInTheWebView(url);
				} 
			}
			if( showingUI==on )
				return;
			showingUI = on;	
			if( on ){
				if( !Cocoon )
					document.getElementsByTagName('iframe')[0].style.display = 'block';
				else
					Cocoon.App.showTheWebView();//Cocoon.App.show(0,0,window.innerWidth,window.innerHeight);
				if( isUIReady && isFSIReady )				
					startPingLoop();
			}
			else{
				if( !Cocoon )
					document.getElementsByTagName('iframe')[0].style.display = 'none';
				else
					Cocoon.App.hideTheWebView();
				endPingLoop();
			}
			/*function serializeToURL(obj) {
				var str = [];
				for(var p in obj)
					if (obj.hasOwnProperty(p)) {
						str.push(encodeURIComponent(p) + "=" + encodeURIComponent(obj[p]));
					}
				return str.join("&");
			}*/
		};
		
		function startWebView(){
			if( !webViewInitReceived || !isWebViewReady )
				return;
			isUIReady = true;		
			log("auto:",autoConnect);	      
			WebViewUI.sendMessage( "auto", autoConnect );
			WebViewUI.sendMessage( "setaddr", getLastConnectAddr() );

			if( reloadUI )
				connectDialogSettings.noAutoconnect = connected;
			if( isFSIReady )
				WebViewUI.sendMessage( "ready", connectDialogSettings );//FSI and UI are ready  
			if( !reloadUI ){
				if( !config.coldStartUI || isFSIReady )
					startPingLoop();
			}
			else{
				//signal if we're connected
				if( connected ){
					WebViewUI.sendMessage( "connect", {address:getLastConnectAddr(),serverName:"",flightSim:""} );
				}
			}
		}
		
		function getLastConnectAddr(){
			var addr = "";
			if( localStorage.connectionHistory ){
				var history = JSON.parse(localStorage.connectionHistory);
				if( history.length>0 )
					addr = history[0];
			}
			return addr;
		}
		
		if( config.coldStartUI	){
			log( "Show UI on cold-start");
			this.show(true);
		}
	}
};
