/* global Cocoon, resizeCanvas, clearTop, widgetsZoom, config, graphics:true */
/* jslint browser: true*/
/* jshint curly: false */

var CanvasGraphics = {
	init : function () {
		var canvas, context,
			onScreenContext, fontInitialized =false;
		this.beginDraw = function(){
			context = canvas.getContext('2d');
			if( context && clearTop )
				context.clearRect ( 0 , 0 , canvas.width , clearTop );
			return context;
		};
		this.setPositionAndScale = function( x,y, scaleX, scaleY ){
			context.setTransform( scaleX, 0, 0, scaleY, x , y );
		};
		this.setAlpha = function( alpha ){
			context.globalAlpha = alpha;
		};
		this.drawImage = function(){
			context.drawImage.apply(context,arguments);
		};
		/* Use to draw the tiled image once */
		this.drawTiledImage = function( image, clipX, clipY, x,y, w,h ){
			if( !image.pattern )
				image.pattern = context.createPattern(image,'repeat');
			context.save();
			context.fillStyle = image.pattern;
			context.translate( x-clipX, y-clipY);
			context.fillRect( clipX,clipY, w,h );
			context.restore();
			//or instead of save-restore:
			//context.fillStyle = "#000000";
			//context.translate( clipX-x, clipY-y);
		};
		/* Use to draw a tiled image many times */
		this.activateTiledImage= function( image ){
			if( !image.pattern )
				image.pattern = context.createPattern(image,'repeat');
			context.fillStyle = image.pattern;
		};
		this.drawActiveTiledImage = function( clipX, clipY, x,y, w,h ){
			context.translate( x-clipX, y-clipY);
			context.fillRect( clipX,clipY, w,h );
			context.translate( clipX-x, clipY-y);
		};
		this.deactivateTiledImage= function(){
			context.fillStyle = "#000000";
		};
		this.rotate = function(){
			context.rotate.apply(context,arguments);
		};
		this.translate = function(){
			context.translate.apply(context,arguments);
		};
		this.drawRect = function(x,y,w,h,color){
			context.save();
			context.fillStyle = color;
			this.setInitialTransform();
			context.rect( x,y,w,h );
			context.fill();
			context.restore();
		};
		this.drawFPS = function(fps){
			if( !fontInitialized ){
				fontInitialized = true;
				canvas.getContext('2d').font = "bold " + 16 * coordScale + "px Arial"; //for the fps meter
			}
			context.fillStyle = "#FFFF00";
			context.setTransform(widgetsZoom, 0, 0, widgetsZoom, 0, 0);
			context.fillText( fps.toFixed() + "fps", 100, 32);
		};
		this.createOffscreenCanvas = function(){
			return document.createElement('canvas');
		};
		this.invalidateOffscreenCanvas = function( cnv ){
			onScreenContext = context;//wil restore it later
			context = cnv.getContext('2d');
			context.setTransform(widgetsZoom, 0, 0, widgetsZoom, 0 , 0 );
			context.clearRect ( 0,0, cnv.width, cnv.height );//@todo - clear partiallyonly where instruments show
		};
		this.drawOffscreenCanvas = function(cnv){
			context.setTransform(widgetsZoom, 0, 0, widgetsZoom, 0/*x*/ , 0/*y*/ );//@todo - x,y
			context.drawImage( cnv,0,0 );//@todo - position
		};
		this.endOffscreenCanvas = function(){
			if( onScreenContext )
				context = onScreenContext;//as it was
		};
		//context = onScreenContext;//in case it was changed
		this.resize = resizeCanvas?resizeCanvas:function(){};
		this.getCanvas = function () {
			return canvas;
		};
		this.width = function(){return canvas.width;};
		this.height = function(){return canvas.height;};
		initCanvas();
		function initCanvas(){
			//canvas setup
			canvas = document.createElement(navigator.isCocoonJS ? 'screencanvas' : 'canvas');
			canvas.screencanvas = true;
			//canvas.setAttribute("id", "canvas");
			document.body.appendChild(canvas);
			
			//canvas = document.getElementById('canvas');
			
			if (Cocoon)
				canvas.style.cssText = "idtkscale:ScaleAspectFit"; //fit the canvas while preserving the aspect, usefull for when there are black borders at both sides
			
			resizeCanvas();
			window.addEventListener('resize', resizeCanvas);
		}
	}
};

if( !config.useWebGL )
	graphics = CanvasGraphics;