/* global Shaders, mat3, vec2, vec3, vec4, resizeCanvas, config, log,error, graphics:true, Cocoon */
/* jslint browser: true*/
/* exported ALIGN, restart */
/* jshint curly: false */
/* jslint multistr: true */

var ALIGN =  {
	Left:1,
	Center:2,
	Right:3,
	Top:4,
	Middle:8,
	Bottom:12,
	MaskHorizontal:3,
	MaskVertical:12
};

var tempV2 = [0,0];
//use this to avoid creating unncessary work for the garbage collector
function v2(x,y){
	tempV2[0] = x;
	tempV2[1] = y;
	return tempV2;
}

var WebGLGraphics = {
	init: function(){
		var RULER_MaxTicks = 100, LINE_MaxTicks = 100, resizeId = 0,
			canvas, gl, rectVertsBuffer, rulerVertsBuffer, lineVertsBuffer, rectVertsFloatArray, 
			useAlphaInFramebuffer=false,
			useOffscreenFramebuffer = true,
			rttFramebuffer, rttTexture, rttWidth, rttHeight, renderToTexture=false,
			programs,
			usedProgram, boundTexture, boundBuffer, boundIndexBuffer, blending,
			trMatrixBase = mat3.create(),
			trMatrixStack = [], trMatrixStackDepth = 0,
			pMatrix = mat3.create(),
			whiteColor = vec4.clone([1,1,1,1]),
			currentImage,
			polarMode = false,
			clipping = false,
			clipToRing = false,
			clippingBox = vec4.create(),
			epsilon = 0.00001,
			vertexAttribArrayCount,
			lastDrawElementNumIndexes = 0,
			lastDrawArraySize = 0,
			shaderVars = {
				trMatrix : mat3.create(),
				textureMatrix : mat3.create(),
				polarCenter : vec2.create(),
				polarXToAngle : 0,
				screenAspect : 1,
				ticks: 0,
				tickUnit:0.1,
				currentColor : vec4.create(),
				color1 : vec4.create(),
				color2 : vec4.create(),
				textureTintColor : vec4.clone([1,1,1,1]),
				textureTintEnabled : false,
				clipCircle : vec3.create(),
				clipConcentric: vec4.create(),
				pieRange: vec2.create(),
				barRange: vec2.create(),
				//pie2Ranges: vec3.create(),
				pieMultiRanges: vec4.create(),//@todo rename pie3Ranges
				pieColor3Ranges: vec3.create(),//@todo - rename pie4Ranges
				pixelSize: 1.0,
				tickWidth: 1.0
			},
			recordDrawImage = {
				isRecording: function(){
					return this.recording;
				},
				start: function(buffer, itemSize){
					this.index = 0;
					this.recording = true;
					this.array = null;
					if( buffer ){
						this.buffer = buffer;
						this.floatArray = buffer.floatArray;
					}
					else{
						this.buffer = null;
						this.floatArray = null;
						this.itemSize = itemSize;
					}
				},
				stop: function(){
					this.recording = false;
				},
				resume: function(){
					this.recording = true;
				},
				updateBuffer: function(dynamicBuffer){
					if( this.array ){
						this.floatArray = new Float32Array(this.array);
						this.array = null;
					}
					boundBuffer = null;
					if( !this.floatArray  )
						this.floatArray = [];
					if( !this.buffer || this.buffer.numItems<this.index/this.buffer.itemSize )
						this.buffer = WebGLGraphics.createShapeBuffer( this.floatArray, this.itemSize, gl.dynamicBuffer?gl.DYNAMIC_DRAW:gl.STATIC_DRAW );//no buffer, or need a bigger buffer
					else{
						gl.bindBuffer(gl.ARRAY_BUFFER, this.buffer);
						if( this.index )
							gl.bufferData(gl.ARRAY_BUFFER, this.floatArray, dynamicBuffer?gl.DYNAMIC_DRAW:gl.STATIC_DRAW);
						//boundBuffer = this.buffer;
					}
					this.buffer.floatArray = this.floatArray;
					this.buffer.activeSize = this.index;
					this.buffer.resizeId = resizeId;
					return this.buffer;
				},
				addValues: function(){
					for( var i=0; i<arguments.length; i++ ){
						if( !this.array ){
							if( !this.floatArray ){
								this.array = [];
								this.array.push(arguments[i]);
							}
							else{
								if( this.floatArray.length==this.index ){
									//not enough space in floatArray, copy everything to array (temporarily)
									this.array = new Array(this.floatArray.length);
									for( var n=0; n<this.floatArray.length; n++ )//can I use the floatArray in array constructor?
										this.array[n] = this.floatArray[n];
									this.array.push(arguments[i]);
								}
								else
									this.floatArray[this.index] = arguments[i];
							}
						}
						else
							this.array.push(arguments[i]);
						this.index++;
					}
				}
			};

		mat3.ortho2d = function(out,w,h){
			out[0] = 2.0/w;
			out[1] = 0.0;
			out[2] = 0.0;
			out[3]= 0.0;
			out[4] = -2.0/h;
			out[5] = 0.0;
			out[6] = -1;
			out[7] = 1;
			out[8] = 1.0;
		};
		/*mat3.subOrtho2d = function(out,w,h,cx,cy,cw,ch){
			out[0] = 2.0/cw;
			out[1] = 0.0;
			out[2] = 0.0;
			out[3]= 0.0;
			out[4] = -2.0/ch;
			out[5] = 0.0;
			out[6] = -1+2.0*cx/w;
			out[7] = 1-2.0*cy/h;
			out[8] = 1.0;
		};*/
		mat3.translateAndScale = function( out, a, x,y,w,h ){
			out[6] = a[6] + x*a[0] + y*a[3];
			out[7] = a[7] + x*a[1] + y*a[4];
			out[0] = a[0]*w;
			out[1] = a[1]*w;
			out[3] = a[3]*h;
			out[4] = a[4]*h;
		};
		this.checkGLErrors = function(){
			var err = gl.getError();
			if( err!=gl.NO_ERROR ){
				error( "WebGL error: " + err );
				debugger;
				return false;
			}
			return true;
		};
		this.startRecordDrawImage = function(buffer,itemSize){
			recordDrawImage.start(buffer,itemSize);
		};
		this.stopRecordDrawImage = function(dynamicBuffer){
			recordDrawImage.stop();
			return recordDrawImage.updateBuffer(dynamicBuffer);
		};
		this.pauseRecordDrawImage = function(){
			recordDrawImage.stop();
		};
		this.resumeRecordDrawImage = function(){
			recordDrawImage.resume();
		};
		this.bufferNeedsUpdate = function(buffer){
			return !buffer || buffer.resizeId!=resizeId;
		};
		this.getResizeId = function(){
			return resizeId;
		};
		this.getCanvas = function () {
			return canvas;
		};
		this.resetTransformMatrix = function(){
			mat3.copy(trMatrixBase,pMatrix);
		};
		this.resetLocalTransformation = function(){
			mat3.copy(shaderVars.trMatrix,trMatrixBase);
		};
		this.getOrthoMatrix = function(){
			return pMatrix;
		};
		this.getTransformMatrix = function(){
			return trMatrixBase;
		};
		this.setTransformMatrix = function(m){
			mat3.copy(trMatrixBase,m);
		};
		this.pushTransformMatrix = function(){
			if( trMatrixStack.length<=trMatrixStackDepth )
				trMatrixStack.push( mat3.create() );
			mat3.copy( trMatrixStack[trMatrixStackDepth], trMatrixBase );
			trMatrixStackDepth++;
		};
		this.popTransformMatrix = function(){
			trMatrixStackDepth--;
			mat3.copy( trMatrixBase, trMatrixStack[trMatrixStackDepth] );
		};
		this.width = function(){return renderToTexture?rttWidth:canvas.width;};
		this.height = function(){return renderToTexture?rttHeight:canvas.height;};
		this.worldToScreenCoords = function(out,a,trmatrix){
			vec2.transformMat3(out, a, trmatrix?trmatrix:trMatrixBase);
			out[0] = (out[0]+1)*0.5*this.width();
			out[1] = (1-out[1])*0.5*this.height();
		};
		this.worldToScreenVector = function(out,a){
			var x = a[0],
				y = a[1];
			out[0] = (trMatrixBase[0] * x + trMatrixBase[3] * y)*0.5*this.width();
    		out[1] = (trMatrixBase[1] * x + trMatrixBase[4] * y)*0.5*this.height();
		};
		this.worldToClipspaceCoords = function( out,a ){
			vec2.transformMat3(out, a, trMatrixBase);
		};
		this.clipSpaceToScreenCoords = function( out,a ){
			out[0] = (a[0]+1)*0.5*this.width();
			out[1] = (1-a[1])*0.5*this.height();
		};
		this.beginDraw = function(){
			mat3.ortho2d(pMatrix, canvas.width, canvas.height);
			if( useAlphaInFramebuffer )
				gl.colorMask( 1,1,1,1 );
			gl.clearColor(0.0, 0.0, 0.0, 1.0);
			gl.clear(gl.COLOR_BUFFER_BIT);
			if( useAlphaInFramebuffer )
				gl.colorMask( 1,1,1,0 );
			return true;
		};
		this.renderingToTexture = function(){
			return renderToTexture;
		};
		this.beginRenderToTexture = function(width,height,linearFiltering){
			renderToTexture = true;
			if( useOffscreenFramebuffer ){
				//@todo - renderbufferStorage with ARM_rgba8
				rttFramebuffer = gl.createFramebuffer();
				gl.bindFramebuffer(gl.FRAMEBUFFER, rttFramebuffer);
				rttFramebuffer.width = width;
				rttFramebuffer.height = height;
				
				//@todo - AA setting have no effect here, why??
				//@todo - mipmap & linear filter support
				rttTexture = gl.createTexture();
				gl.bindTexture(gl.TEXTURE_2D, rttTexture);
				gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, linearFiltering?gl.LINEAR:gl.NEAREST);
				gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, linearFiltering?gl.LINEAR:gl.NEAREST);
				gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
				gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
				gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, rttFramebuffer.width, rttFramebuffer.height, 0, gl.RGBA, gl.UNSIGNED_BYTE, null);
				gl.framebufferTexture2D(gl.FRAMEBUFFER, gl.COLOR_ATTACHMENT0, gl.TEXTURE_2D, rttTexture, 0);
				gl.bindTexture(gl.TEXTURE_2D, null);
				rttTexture.width = rttFramebuffer.width;
				rttTexture.height = rttFramebuffer.height;
				boundTexture = null;
				
				//gl.disable( gl.BLEND );
				//blending = false;//@bug - why does this blur the compass???
				//gl.blendEquationSeparate( gl.FUNC_ADD, gl.FUNC_ADD );
				//gl.blendFuncSeparate( gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA, gl.ONE, gl.ONE );
				
				if (gl.checkFramebufferStatus(gl.FRAMEBUFFER) != gl.FRAMEBUFFER_COMPLETE) {
				   error("Could not allocate framebuffer");//@todo - fail gracefuly
				   return;
				}
				
				gl.viewport(0, 0, rttFramebuffer.width, rttFramebuffer.height);
				mat3.ortho2d(pMatrix, width,height);
			}
			else{
				
				//render to the regular framebuffer, this does not work in cocoonjs
				//gl.colorMask( 1,1,1,1 );
			}
			rttWidth = width;
			rttHeight = height;
			shaderVars.screenAspect = rttHeight/rttWidth;
			if( useAlphaInFramebuffer )
				gl.colorMask( 1,1,1,1 );
			/*if( rttWidth!=canvas.width ){
				gl.clearColor(1.0, 1.0, 0.0, 0.5);//use white color for clearing - much better white anti-aliased lines, some border around dark objects, etc
				gl.clear(gl.COLOR_BUFFER_BIT);
				//graphics.drawRect(10,10, width-20, height-20, [1,0,0,1] );
			}
			else{
				gl.clearColor(1.0, 1.0, 0.0, 0.0);//use white color for clearing - much better white anti-aliased lines, some border around dark objects, etc
				gl.clear(gl.COLOR_BUFFER_BIT);
			}*/
			gl.clearColor(1.0, 1.0, 1.0, 0.0);//use white color for clearing - much better white anti-aliased lines, some border around dark objects, etc
			gl.clear(gl.COLOR_BUFFER_BIT);
			
			if( config.debugWebGL )
				graphics.checkGLErrors();
			
			
		};
		this.endRenderToTexture = function(texture){
			renderToTexture = false;
			shaderVars.screenAspect = canvas.height/canvas.width;
			if( useAlphaInFramebuffer )
				gl.colorMask( 1,1,1,0 );
			if( useOffscreenFramebuffer ){
				gl.framebufferTexture2D(gl.FRAMEBUFFER, gl.COLOR_ATTACHMENT0, gl.TEXTURE_2D, null, 0);
				gl.bindFramebuffer(gl.FRAMEBUFFER, null);
				gl.deleteFramebuffer(rttFramebuffer);
				gl.viewport(0, 0, canvas.width, canvas.height);
				mat3.ortho2d(pMatrix, canvas.width, canvas.height);
				rttTexture.resizeId = resizeId;
				return rttTexture;
			}
			else{
				//render to the regular framebuffer, this does not work in cocoonjs
				if( !texture ){
					texture = gl.createTexture();
					gl.bindTexture(gl.TEXTURE_2D, texture);
					gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
					gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
					gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
					gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
				}
				else
					gl.bindTexture(gl.TEXTURE_2D, texture);
				boundTexture = texture;
				gl.copyTexImage2D(gl.TEXTURE_2D,0,gl.RGBA,0,0,rttWidth,rttHeight,0);
				texture.width = rttWidth;
				texture.height = rttHeight;
				texture.resizeId = resizeId;
				return texture;
			}
		};
		this.renderTextureNeedsUpdate = function(texture){
			return texture.resizeId != resizeId;
		};
		/* Can use WebGLTexture object instead of an image, but the width/height properties have to be set (the default object does not have them)*/
		this.drawImage = function( image, cx,cy, cw,ch, x, y, w,h, sprog, disableBlend, rotate ){
			var cropped, sx, sy, tx, ty;
			if( x===undefined ){
				x=cx;
				y=cy;
				if( cw===undefined ){
					w = image.width;
					h = image.height;
				}
				else{
					w=cw;
					h=ch;
				}
			}
			else{
				//a cropped texture
				cropped = true;
				sx = cw/image.width;
				sy = ch/image.height;
				tx = cx/image.width;
				ty = cy/image.height;				
			}
			
			if( recordDrawImage.isRecording() ){
				mat3.copy( shaderVars.trMatrix, trMatrixBase );//@todo - not always needed?
				recordDrawImage.addValues(	x,y,tx,ty,
											x+w,y,tx+sx,ty,
											x+w,y+h,tx+sx,ty+sy,
											x,y+h,tx,ty+sy,
											x,y,tx,ty,
											x+w,y+h,tx+sx,ty+sy );
			}
			else{
				if( cropped ){
					if( !sprog )
						sprog = polarMode?programs.shaderProgramCroppedTexturePolar:programs.shaderProgramCroppedTexture;
					//mat3.identity( textureMatrix ); 
					//mat3.translate(textureMatrix, textureMatrix,  v2(tx, ty));
					//mat3.scale(textureMatrix, textureMatrix, v2(sx, sy));
					shaderVars.textureMatrix[6] = tx;
					shaderVars.textureMatrix[7] = ty;
					shaderVars.textureMatrix[0] = sx;
					shaderVars.textureMatrix[1] = 0;
					shaderVars.textureMatrix[3] = 0;
					shaderVars.textureMatrix[4] = sy;
				}
				else if( !sprog ){
					if( !rotate )
						sprog = polarMode?programs.shaderProgramTexturePolar:programs.shaderProgramTexture;
					else
						sprog = polarMode?programs.shaderProgramCroppedTexturePolar:programs.shaderProgramCroppedTexture;
				}
				if( rotate ){
					shaderVars.textureMatrix[0] = 0;
					shaderVars.textureMatrix[1] = 1;
					shaderVars.textureMatrix[3] = 1;
					shaderVars.textureMatrix[4] = 0;
					shaderVars.textureMatrix[6] = 0;//tx(y)
					shaderVars.textureMatrix[7] = 0;//ty(x)
				}
				this.translateAndScale(x,y,w,h);
				glActivateDraw( rectVertsBuffer, null, null, sprog, image, !disableBlend );
				gl.drawArrays(gl.TRIANGLE_FAN, 0, rectVertsBuffer.numItems);
			}
			
		};
		this.drawTickedImage = function(image,x,y,w,h,unit,tickWidth,unit2,tickWidth2,range1,range2,rotate){
			shaderVars.tickUnit = unit;
			var v = v2(rotate?0:w,rotate?h:0);
			graphics.worldToScreenVector(v,v);
			shaderVars.pixelSize = 1.0/vec2.length(v);//as a fraction of the image
			shaderVars.tickWidth = tickWidth;//as a fraction of the image
			if( unit2 ){
				shaderVars.tickWidth2 = tickWidth2;//as a fraction of the image
				shaderVars.tickUnit2 = unit2;
			}
			var program;
			if( range1!==undefined ){
				if( range2===undefined )
					range2 = 100000000;//after the end
				vec2.set(shaderVars.barRange,range1,range2);
				if( unit2 )
					program = rotate?programs.shaderProgramTexTextransformedRangesTick2:programs.shaderProgramTextureRangesTick2;//use tex transform matrix when rotating
				else
					program = rotate?programs.shaderProgramTexTextransformedRangesTick:programs.shaderProgramTextureRangesTick;//use tex transform matrix when rotating
			}
			else{
				program = rotate?programs.shaderProgramTexTextransformedTick:programs.shaderProgramTextureTick;//use tex transform matrix when rotating
			}
			this.drawImage( image, x,y,w,h, undefined, undefined, undefined, undefined, program, true, rotate );
		};
		this.drawTexturedQuad = function(image,x,y,w,h,tintColor){
			if( x!==undefined ){
				if( w!==undefined )
					this.translateAndScale(x,y,w,h);
				else
					this.translate(x,y);
			}
			if( tintColor ){
				shaderVars.textureTintEnabled = true;
				vec4.copy(shaderVars.textureTintColor,tintColor);
			}
			var sprog = polarMode?programs.shaderProgramTexturePolar:programs.shaderProgramTexture;
			glActivateDraw( rectVertsBuffer, null, null, sprog, image, true );
			gl.drawArrays(gl.TRIANGLE_FAN, 0, rectVertsBuffer.numItems);
			
			shaderVars.textureTintEnabled = false;
		};
		/*
		 * Todo - rename to localTranslateAndScale and call resetLocal before using
		 */
		this.translateAndScale = function( x,y,w,h ){
			/* //Identical code, but much slower, we win some fps by replacing it
			//mat3.copy( trMatrix, trMatrixBase );
			//mat3.translate(trMatrix, trMatrix,  v2(x, y));
			//mat3.scale(trMatrix, trMatrix, v2(w, h));
			*/
			mat3.translateAndScale(shaderVars.trMatrix,trMatrixBase,x,y,w,h);
		};
		this.localTranslate = function( x,y ){
			mat3.translate(shaderVars.trMatrix, shaderVars.trMatrix,  v2(x, y));
		};
		this.setPositionAndScale = function( x,y, w, h ){
			/*mat3.copy(trMatrixBase,pMatrix);
			mat3.translate(trMatrixBase, trMatrixBase,  v2(x, y));
			if( scaleX )
				mat3.scale(trMatrixBase, trMatrixBase, v2(scaleX, scaleY));*/
			mat3.translateAndScale(trMatrixBase,pMatrix,x,y,w,h);
		};
		this.rotate = function(angle){
			mat3.rotate(trMatrixBase, trMatrixBase, angle);
			//context.rotate.apply(context,arguments);
		};
		this.translate = function(x,y){
			mat3.translate(trMatrixBase, trMatrixBase,  v2(x, y));
			//context.translate.apply(context,arguments);
		};
		this.scale = function(x,y){
			mat3.scale(trMatrixBase, trMatrixBase,  v2(x, y));
		};
		this.drawRect = function(x,y,w,h,fillColor,outlineColor, disableBlend ){
			this.drawShape( rectVertsBuffer, null, x,y,w,h, fillColor, outlineColor, disableBlend );
		};
		this.drawPie = function(x,y,radius1,radius2,start,end,image,color1,color2,range1,range2,range3, tickAngle, disableBlend, repeatVectors/*these are disposable and will be changed*/ ){
			//convert to pixel space:
			var i, v = v2(radius2,0);//the size is in width %
			graphics.worldToScreenVector(v,v);
			var r2 = vec2.length(v),
				scale = r2/radius2;
			v = v2(x,y);//position
			graphics.worldToScreenCoords(v,v);
			var posx = v[0],
				posy = v[1];
			vec4.set( shaderVars.clipConcentric, posx,this.height()-posy, radius1*scale, r2 );
			
			if( repeatVectors ){
				for(i=0; i<repeatVectors.length; i++ )
					graphics.worldToScreenVector(repeatVectors[i],repeatVectors[i]);
			}
			
			var program;
			if( image ){
				if( range1!==undefined ){
					if( range3!==undefined ){
						program = tickAngle?programs.shaderProgramPie4RangesTicks:programs.shaderProgramPie4Ranges;
						vec3.set(shaderVars.pieColor3Ranges,range1*Math.PI/180.0,range2*Math.PI/180.0,range3*Math.PI/180.0);
						vec2.set(shaderVars.pieRange,start*Math.PI/180.0,end*Math.PI/180.0);
					}
					else{
						program = tickAngle?programs.shaderProgramPieRangesTicks:programs.shaderProgramPieRanges;
						if( range2===undefined )
							range2 = 100000000;//after the end
						vec4.set(shaderVars.pieMultiRanges,start*Math.PI/180.0,range1*Math.PI/180.0,range2*Math.PI/180.0,end*Math.PI/180.0);
					}
				}
				else
					program = tickAngle?programs.shaderProgramPieTexturedTicks:programs.shaderProgramPieTextured;
				/*else if( range1!==undefined ){
					//program = tickAngle?programs.shaderProgramPie2RangesTicks:programs.shaderProgramPie2Ranges;
					program = tickAngle?programs.shaderProgramPie2RangesTicks:programs.shaderProgramPie2Ranges;
					vec3.set(shaderVars.pie2Ranges,start*Math.PI/180.0,range1*Math.PI/180.0,end*Math.PI/180.0);
				}
				else
					program = tickAngle?programs.shaderProgramPieTexturedTicks:programs.shaderProgramPieTextured;*/
			}
			else{
				if( color1 && !color2 ){
					program = tickAngle?programs.shaderProgramPieTicks:programs.shaderProgramPie;
					vec4.copy(shaderVars.currentColor, color1);
				}
				else if( color2 ){
					program = programs.shaderProgramPieGradient;//no ticks for gradient pie
					vec4.copy(shaderVars.color1, color1);
					vec4.copy(shaderVars.color2, color2);
				}
			}
			
			if( range1===undefined /*|| range2===undefined*/ )
				vec2.set(shaderVars.pieRange,start*Math.PI/180.0,end*Math.PI/180.0);
			
			if( tickAngle )
				shaderVars.tickAngle = tickAngle*Math.PI/180.0;
			
			this.pushTransformMatrix();
			graphics.resetTransformMatrix();
			this.translateAndScale(posx-r2,posy-r2,r2*2,r2*2);
			
			glActivateDraw( rectVertsBuffer, null, null, program, image, !disableBlend/*blend*/ );
			gl.drawArrays( gl.TRIANGLE_FAN, 0,4 );
			
			if( repeatVectors ){
				for(i=0; i<repeatVectors.length; i++ ){
					graphics.resetTransformMatrix();
					shaderVars.clipConcentric[0] = posx+repeatVectors[i][0];//in pixels
					shaderVars.clipConcentric[1] =  this.height()-posy+repeatVectors[i][1];//in pixels
					this.translateAndScale(posx-r2+repeatVectors[i][0],posy-r2-repeatVectors[i][1],r2*2,r2*2);
					
					usedProgram.uniforms.trMatrixUniform.load();//update the position
					usedProgram.uniforms.clipConcentricUniform.load();//update the circle screen space center
					
					gl.drawArrays( gl.TRIANGLE_FAN, 0,4 );
				}
			}
			
			this.popTransformMatrix();
		};
		/*Wll draw a circle in polar mode*/
		var srange =[ 0,1 ];
		this.drawSegmentedLine = function( x,y,w,numSegments,fillColor,outlineColor, disableBlend ){
			if( numSegments>=lineVertsBuffer.length )
				numSegments=lineVertsBuffer.length-1;
			srange[1] = numSegments+1;
			this.drawShape( lineVertsBuffer, null, x,y,w/(numSegments),1, fillColor, outlineColor, disableBlend, false, true, srange );
		};
		this.createShapeBuffer = function( vertices, itemSize, usage ){
			var buffer = gl.createBuffer();
			gl.bindBuffer(gl.ARRAY_BUFFER, buffer);
			gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(vertices), usage?usage:gl.STATIC_DRAW);
			buffer.itemSize = itemSize;
			buffer.numItems = vertices.length/itemSize;
			boundBuffer = null;//buffer;
			return buffer;
		};
		this.createIndexBuffer = function( indexes, usage ){
			var buffer = gl.createBuffer();
			gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, buffer);
			gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, new Uint16Array(indexes), usage?usage:gl.STATIC_DRAW );
			buffer.itemSize = 1;//not needed
    		buffer.numItems = indexes.length;
			boundIndexBuffer = null;//buffer;
			return buffer;
		};
		this.deleteBuffer = function( buffer ){
			gl.deleteBuffer(buffer);
		};
		this.drawTexturedTriangles = function(image,buffer,rangeEnd,disableBlend){
			var program = polarMode?programs.shaderProgramTextureCoordPolar:programs.shaderProgramTextureCoord;
			glActivateDraw( buffer, null, null, program, image, !disableBlend );
			gl.drawArrays( gl.TRIANGLES, 0, rangeEnd?rangeEnd:buffer.numItems);
		};
		this.repeatTexturedTriangles = function(buffer,rangeEnd){
			usedProgram.uniforms.trMatrixUniform.load();//load the new transform matrix
			gl.drawArrays( gl.TRIANGLES, 0, rangeEnd?rangeEnd:buffer.numItems);
		};
		/*
		 * This will also update the program uniforms, except for images. Will update blend, Will not update attributes.
		 */
		this.changeProgram = function( hasColorAttrib, hasImage, blend ){
			updateBlend(blend);
			var program = getShapeProgram( hasColorAttrib, hasImage );
			if( usedProgram!=program ){
				usedProgram = program;
				gl.useProgram(program.glProgram);
				for( var u in program.uniforms )
					program.uniforms[u].load();
			}
		};
		function getShapeProgram( hasColorAttrib, image, interpolateVertexColor ){
			if( hasColorAttrib ){
				if( !interpolateVertexColor )
					return polarMode?programs.shaderProgramVertexColorPolar:(shaderVars.clipCircle[2]? programs.shaderProgramCircleClipVertexColor : programs.shaderProgramVertexColor);
				else
					return polarMode?programs.shaderProgramInterpolatedVertexColorPolar:(shaderVars.clipCircle[2]? programs.shaderProgramCircleClipVertexColor : programs.shaderProgramInterpolatedVertexColor);
			}
			else{
				if( image ){
					return clipToRing?programs.shaderProgramTexCoordRingClip:programs.shaderProgramTextureCoord;
				}
				else
					return polarMode?programs.shaderProgramColorFillPolar:(shaderVars.clipCircle[2]? programs.shaderProgramCircleClip : programs.shaderProgramColorFill);
			}
		}
		this.drawShape = function(buffer,colorBuffer,x,y,w,h,fillColor,outlineColor, disableBlend, strip, open, range, indexBuffer, image, interpolateColor ){
			//var cc = shaderVars.currentColor;
			
			if( config.debugWebGL )
				graphics.checkGLErrors();

			if( x!==undefined ){
				if( w!==undefined )
					this.translateAndScale(x,y,w,h);
				else
					this.translate(x,y);
			}
			
			if( image )
				mat3.identity( shaderVars.textureMatrix ); 
			
			var program = getShapeProgram( colorBuffer, image, interpolateColor );
			var n;
			if( fillColor || colorBuffer ){				
				//shaderVars.currentColor = fillColor;
				if( fillColor )
					vec4.copy(shaderVars.currentColor, fillColor);
				glActivateDraw( buffer, indexBuffer, colorBuffer, program, image, 
								   ((fillColor?(fillColor[3]!=1.0):false) && !disableBlend) ||
								   (config.antialiasShapes && colorBuffer) ||//why does this cause a white blink in the compass
								   disableBlend===false/*not undefined - explicitly set to false*/ );
				n=0;
				if( indexBuffer ) {
				   	lastDrawElementNumIndexes = indexBuffer.numItems;
					gl.drawElements(gl.TRIANGLES, lastDrawElementNumIndexes, gl.UNSIGNED_SHORT, 0);
				}
				else{
					lastDrawArraySize = range?range[1]:buffer.numItems;
					gl.drawArrays( strip?gl.TRIANGLE_STRIP:gl.TRIANGLE_FAN, range?range[0]:0, lastDrawArraySize );
				}
				
			}
			
			if( outlineColor ){
				//shaderVars.currentColor = outlineColor;
				vec4.copy(shaderVars.currentColor, outlineColor);
				if( colorBuffer || !fillColor ){
					//use a new shader for the outline
					program = polarMode?programs.shaderProgramColorFillPolar:programs.shaderProgramColorFill;
					glActivateDraw( buffer, indexBuffer, null, program, image, false );
				}
				else// if( fillColor )
					changeActiveColor(program,outlineColor);//use the same shader but with different color
				
				n = 0;
				//@todo - gl.LINES index list will be needed for multisegment shapes with outlines
				/*if( indexBuffer ){
					gl.drawElements(gl.TRIANGLES, indexBuffer.numItems, gl.UNSIGNED_SHORT, 0);
				}
				else*/
				lastDrawArraySize = range?range[1]:buffer.numItems;
				gl.drawArrays( open?gl.LINE_STRIP:gl.LINE_LOOP, range?range[0]:0, lastDrawArraySize );
			}
			if( config.debugWebGL )
				graphics.checkGLErrors();
			//shaderVars.currentColor = cc;//as it was
		};
		this.drawRuler = function( rd, blend, vertsBuffer,indexBuffer,numIndexes,colorBuffer, image, interpolateVertexColor ){
			if( rd.ticks>0 ){
				mat3.translate(shaderVars.trMatrix,rd.matrix,v2(rd.origin,0));
				updateBlend(blend);
				var program = getShapeProgram( colorBuffer, image, interpolateVertexColor );
				if( vertsBuffer ){
					glActivateDraw( vertsBuffer, indexBuffer, colorBuffer, program, image, blend || shaderVars.currentColor[3]!=1.0 || colorBuffer );
					lastDrawElementNumIndexes = numIndexes;
					gl.drawElements(gl.TRIANGLES, numIndexes, gl.UNSIGNED_SHORT, 0);
				}
				else{
					glActivateDraw( rulerVertsBuffer, null, null, program, image, blend || shaderVars.currentColor[3]!=1.0 );	
					lastDrawArraySize = rd.ticks*2;
					gl.drawArrays(gl.LINES, 0, lastDrawArraySize );
				}
				if( config.debugWebGL )
					graphics.checkGLErrors();
			}
		};
		this.repeatDraw = function( drawArray, updateUniforms ){
			if( updateUniforms ){
				for( var u in usedProgram.uniforms )
					usedProgram.uniforms[u].load();
			}
			if( drawArray )
				gl.drawArrays( gl.TRIANGLE_FAN, 0, lastDrawArraySize );
			else
				gl.drawElements(gl.TRIANGLES, lastDrawElementNumIndexes, gl.UNSIGNED_SHORT, 0);
			
			if( config.debugWebGL )
				graphics.checkGLErrors();
		};
		/* @todo - move this to instruments */
		this.getRulerData = function( rd, x,y,w,h, start, end, delta, min, max, horizontal, flip, recalculateMatrix ){
			if( !rd ){
				rd = {
					//matrix:mat3.create(),
					ticks:0,
					start:0,
					resizeId:resizeId
				};
			}
			if( horizontal )
				flip = !flip;
			var //origin = 0,
				ori = Math.ceil(start)-start+delta,
				scale = 1/(end-start);
			if( min!==undefined ){
				if( Math.ceil(start)+0.001<min ){
					//var deltaMin = Math.floor(min-Math.ceil(start)-0.001)+1;
					var deltaMin = Math.floor(min-Math.ceil(start+0.1)-0.001)+1;
					start+=deltaMin;
					ori+=deltaMin;
				}
				/* Same as above, keep for refernece
				while( Math.ceil(start)+0.001<min && start<end ){
					start++;
					ori++;
				}*/
			}
			if( max!==undefined ){
				if( Math.floor(end)-0.001>max ){
					var deltaMax = Math.floor(end)-max;
					end-=deltaMax;
				}
			}
			var ticks = Math.floor(end)-Math.ceil(start);
			if( Math.ceil(start)<=Math.floor(end) ) 
				ticks++;
			if( ticks<=0 ){
				if( !rd.matrix )
					rd.matrix = mat3.create();
				rd.ticks = rd.start = rd.end = rd.origin = 0;
				return rd;
			}
			if( ticks>rulerVertsBuffer.numItems/2 )
				ticks = rulerVertsBuffer.numItems/2;
			rd.resized = (rd.resizeId!=resizeId);
			if( !rd.matrix || recalculateMatrix || rd.resized ){
				if( !rd.matrix )
					rd.matrix = mat3.create();
				rd.resizeId=resizeId;
				var matrix = rd.matrix;
				mat3.copy( matrix, trMatrixBase );
				if( !flip )
					scale = -scale;
				if( horizontal ){
					if( !flip )
						mat3.translateAndScale(matrix,matrix, x/*+origin*scale*w*/+w, y, w*scale, h );
					else
						mat3.translateAndScale(matrix,matrix, x/*+origin*scale*w*/, y, w*scale, h );
				}
				else{
					//rotate 90 degrees
					mat3.translate(matrix, matrix,  v2(w, 0));
					//matrix[6] = matrix[6] + w*matrix[0];
					//@todo - optimize:
					mat3.rotate(matrix,matrix,Math.PI/2);
					if( !flip )
						mat3.translateAndScale(matrix,matrix,y/*+origin*scale*h*/+h, 0, h*scale, w);
					else
						mat3.translateAndScale(matrix,matrix,y/*+origin*scale*h*/, x, h*scale, w);
				}
				rd.horizontal = horizontal;
			}
			rd.origin = ori;
			rd.start = Math.ceil(start);
			rd.ticks = ticks;
			return rd;
		};
		this.drawFPS = function(fps){
			//this.disableClip();
			//this.resetTransformMatrix();
			//TextManager.printNumber(fps,10,10,0.025,0);
		};
		/* The clip box is specified in workld coordinates*/
		this.setClip = function(x,y,w,h, skipScissor ){
			if( !clipping ){
				if( !skipScissor )
					gl.enable(gl.SCISSOR_TEST);
				clipping = true;
			}
			if( x!==undefined ){
				var x2 = x+w,
					y2 = y+h;
				var v = v2(x,y);
				this.worldToScreenCoords(v,v);
				x = v[0]|0;
				y = v[1]|0;
				v = v2(x2,y2);
				this.worldToScreenCoords(v,v);
				x2 = Math.ceil(v[0]);
				y2 = Math.ceil(v[1]);
				if( skipScissor || Math.abs(clippingBox[0]-x)>epsilon || Math.abs(clippingBox[1]-(canvas.height-y2))>epsilon ||
				    Math.abs(clippingBox[2]-(x2-x))>epsilon || Math.abs(clippingBox[3]-(y2-y))>epsilon ){
					vec4.set( clippingBox, x, canvas.height-y2, x2-x, y2-y );
					if( !skipScissor )
						gl.scissor( clippingBox[0], clippingBox[1], clippingBox[2], clippingBox[3] );
				}
			}
		};
		this.disableClip = function( skipScissor ){
			if( clipping ){
				clipping = false;
				if( !skipScissor )
					gl.disable(gl.SCISSOR_TEST);
			}
		};
		this.isClipEnabled=function(){
			return clipping;
		};
		this.getClipScreenBox = function(){
			return clippingBox;
			//return gl.getParameter(gl.SCISSOR_BOX);
		};
		this.setClipScreenBox = function(box){
			if( box ){
				if( !clipping ){
					clipping = true;
					gl.enable(gl.SCISSOR_TEST);
				}
				if( clippingBox[0] != box[0]|| clippingBox[1] != box[1] || clippingBox[2] != box[2] || clippingBox[3] != box[3] ){
					gl.scissor( box[0], box[1], box[2], box[3] );
					vec4.copy( clippingBox, box );
				}
			}
			else
				this.disableClip();
		};
		this.setClipCircle = function(x,y,radius){
			var v = v2(0,radius);
			graphics.worldToScreenVector(v,v);
			var r = vec2.length(v);
			v = v2(x,y);
			graphics.worldToScreenCoords(v,v);
			vec3.set( shaderVars.clipCircle, v[0],this.height()-v[1], r );
		};
		this.disableClipCircle = function(){
			shaderVars.clipCircle[2] = 0.0;
		};
		this.setRingClip = function(x,y,r1,r2 ){
			clipToRing = true;
			var v = v2(x,y);
			this.worldToScreenCoords(v,v);
			x = v[0];
			y = v[1];
			
			v = v2(0,r1);
			this.worldToScreenVector(v,v);
			r1 = vec2.length(v);
			
			v = v2(0,r2);
			this.worldToScreenVector(v,v);
			r2 = vec2.length(v);
			
			vec4.set( shaderVars.clipConcentric, x,this.height()-y, r1, r2 );
		};
		this.disableRingClip = function(){
			clipToRing = false;
		};
		this.resize = function(){
			resizeId++;
			log( 'window resize: w=' + window.innerWidth +' ,h=' +  window.innerHeight + ' ,pixel ratio=' + window.devicePixelRatio );
			if( resizeCanvas )	
				resizeCanvas();
			else{
				if( canvas ){
					canvas.width = window.innerWidth*window.devicePixelRatio;
					canvas.height = window.innerHeight*window.devicePixelRatio;
				}
			}
			if( gl )
				gl.viewport(0, 0, canvas.width, canvas.height);
			if( useAlphaInFramebuffer )
				gl.colorMask( 1,1,1,1 );
			gl.clear(gl.COLOR_BUFFER_BIT);//set alpha to 1 and don't touch it
			if( useAlphaInFramebuffer )
				gl.colorMask( 1,1,1,0 );
			shaderVars.screenAspect = canvas.height/canvas.width;
		};
		this.setColor = function( color ){
			shaderVars.currentColor = vec4.clone(color);
		};
		this.setLineWidth = function( width ){
			gl.lineWidth(width);
		};
		this.activateTextureTint = function(){
			shaderVars.textureTintEnabled = true;
		};
		this.deactivateTextureTint = function(){
			shaderVars.textureTintEnabled = false;
		};
		this.activatePolar = function( centerX, centerY, width ){
			polarMode = true;
			//width = 1004;
			//scale the clipspace X to get the polaar angle, don't put the scaling in the matrix because the text rendering works in screen space
			var clipSpaceWidth = 2*width/this.width();//in clispace the screen width is 2
			shaderVars.polarXToAngle = 2*Math.PI/clipSpaceWidth;
			
			//this.translate( -centerX,-centerY );
			vec2.set( shaderVars.polarCenter,centerX,centerY );
			this.worldToClipspaceCoords( shaderVars.polarCenter,shaderVars.polarCenter );
		};
		this.deactivatePolar = function(){
			polarMode = false;
		};
		this.reActivatePolar = function(){
			polarMode = true;
		};
		this.isPolarActive = function(){
			return polarMode;
		};
		this.getPolarCenter = function( v ){
			v[0] = shaderVars.polarCenter[0];
			v[1] = shaderVars.polarCenter[1];
			this.clipSpaceToScreenCoords(v,v);
		};
		this.getPolarWidth = function(){
			var clipSpaceWidth = 2*Math.PI/shaderVars.polarXToAngle;
			return clipSpaceWidth*this.width()/2;//in clispace the screen width is 2
		};
		this.polarToViewspace = function(v,mat,xScale){
			if( !mat )
				mat = shaderVars.trMatrix;
			vec2.transformMat3(v,v,mat);
			vec2.subtract(v,v,shaderVars.polarCenter);
			var angle = v[0]*shaderVars.polarXToAngle;
			v[0] = v[1]*Math.sin(angle)*shaderVars.screenAspect*(xScale?xScale:1)+shaderVars.polarCenter[0];
			v[1] = v[1]*Math.cos(angle)+shaderVars.polarCenter[1];
			/*vec2 pos = (uTransformMatrix * vec3(aVertexPosition, 1)).xy - uPolarCenter;\
			float angle = pos.x*uPolarXToAngle;\
			gl_Position = vec4(pos.y*sin(angle)*uScreenAspect+uPolarCenter.x,pos.y*cos(angle)+uPolarCenter.y,0,1);\*/
		};
		this.getPolarData = function(obj){
			if( polarMode ){
				if( !obj.polarCenter )
					obj.polarCenter = vec2.create();
				vec2.copy( obj.polarCenter, shaderVars.polarCenter );
				obj.polarXToAngle = shaderVars.polarXToAngle;
			}
		};
		this.setPolarData = function(obj){
			if( obj.polarCenter ){
				polarMode = true;
				vec2.copy( shaderVars.polarCenter, obj.polarCenter );
				shaderVars.polarXToAngle = obj.polarXToAngle;
			}
			else
				polarMode = false;
		};
		this.setTextureTintColor = function( tColor ){
			if( tColor )
				vec4.copy( shaderVars.textureTintColor, tColor );
			else
				vec4.copy( shaderVars.textureTintColor, whiteColor );
		};
		this.createTexture = function( image, mipmap ){
			var texture = gl.createTexture();
			
			gl.bindTexture(gl.TEXTURE_2D, texture);
			gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, image);
			gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
			gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
			gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
			if( mipmap ){
				gl.generateMipmap(gl.TEXTURE_2D);
				gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR_MIPMAP_LINEAR);//hmm, we don't care about mipmap transitions, but trilinear looks better
			}
			else{
				gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
			}
			
			//gl.bindTexture(gl.TEXTURE_2D, null);
			boundTexture=texture;
			return texture;
		};
		this.deleteTexture = function( texture ){
			gl.deleteTexture(texture);
		};
		
		initGL();
		window.addEventListener('resize', function(){graphics.resize();});
		
		function initGL() {
			if( Cocoon )
				Cocoon.Utils.setAntialias(!config.disableWebGLAA);
			
			//webgl setup
			canvas = document.createElement(navigator.isCocoonJS ? 'screencanvas' : 'canvas');
			canvas.screencanvas = true;
			canvas.width = window.innerWidth*window.devicePixelRatio;
			canvas.height = window.innerHeight*window.devicePixelRatio;
			//canvas.width = canvas.height = 500;
			canvas.setAttribute("id", "canvas");
			//canvas.setAttribute("tabindex", "1");//so that we can catch blur events
			document.body.appendChild(canvas);
			
			boundTexture = null;
			boundBuffer = null;
			boundIndexBuffer = null;
			usedProgram = null;
			vertexAttribArrayCount = 0;
			polarMode = false;
			clipping = false;
			blending = false;
			vec4.set( clippingBox, 0,0,0,0 );
			currentImage = null;
			
			//canvas = document.getElementById('canvas');
			
			//setup handling of context lost
			canvas.addEventListener("webglcontextlost", function(event) {
				event.preventDefault();
				log('WebGL context lost, restarting');
				restart();
			}, false);
			canvas.addEventListener( "webglcontextrestored", 
									function(){
										log('WebGL context lost, reinitializing');
										//recordDrawImage.reInit();
										/*notifyInstruments(instruments);
										TextManager.onContextLost();
										initGL();
										function notifyInstruments(instr){
											for( var i=0; i<instr.length; i++ ){
												if( instr[i].onContextLost )
													instr[i].onContextLost();
												if( instr[i].instruments )
													notifyInstruments(instr[i].instruments);
											}
										}*/
									},
									false);
			var exc = false,
				ie = window.navigator.userAgent.indexOf("MSIE ")>=0 || window.navigator.userAgent.indexOf("Trident")>=0;
			try {
				gl = canvas.getContext(ie?"experimental-webgl":"webgl",{alpha: useAlphaInFramebuffer, premultipliedAlpha: false , depth:false, antialias : !config.disableWebGLAA, stencil:false});
			} catch (e) {
				error("Exception error initializeing WebGL");
				exc = true;
			}
			if( !gl ){
				if( !exc )
					error("Error initializing WebGL: can not get context");
				return false;
			}
			
			/*var contextAttribs = gl.getContextAttributes();
			if( contextAttribs && !contextAttribs.antialias) {
				//@todo - do your own anti-aliasing, not supported by cocoonjs yet
			}*/

			gl.disable(gl.DEPTH_TEST);
			gl.clearColor(0.0, 0.0, 0.0, 1.0);
			
			//gl.enable( gl.BLEND );
			gl.blendFunc( gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA );
			
			//use a separate blending function for the alpha - it should add-up. Will have to remove disableBlend for many objects for this to work properly
			//disable for now as it blurs the compass text (and sharpens the vspeed one - so there's a solution, maybe shadow problems...)
			gl.blendEquationSeparate( gl.FUNC_ADD, gl.FUNC_ADD );
			gl.blendFuncSeparate( gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA, gl.ONE, gl.ONE );
			
			gl.activeTexture(gl.TEXTURE0);
			
			//initShaders();
			programs = Shaders.init( gl, shaderVars );
			initBuffers();

			WebGLGraphics.resize();
			
			
			/*function initShaders(){
				
			};*/

			function initBuffers(){
				var vertices = [
					0.0,  0.0,
					1.0,  0.0,
					1.0,  1.0,
					0.0,  1.0,
				];
				rectVertsFloatArray = new Float32Array(vertices);
				rectVertsBuffer = WebGLGraphics.createShapeBuffer(vertices,2);
				
				vertices = new Array(RULER_MaxTicks*4);
				for( var i=0; i<RULER_MaxTicks; i++ ){
					vertices[4*i] = i;
					vertices[4*i+1] = 0;
					vertices[4*i+2] = i;
					vertices[4*i+3] = 1;
				}
				rulerVertsBuffer = WebGLGraphics.createShapeBuffer(vertices,2);
				
				vertices = new Array(LINE_MaxTicks*2);
				for( i=0; i<LINE_MaxTicks; i++ ){
					vertices[2*i] = i;
					vertices[2*i+1] = 0;
				}
				lineVertsBuffer = WebGLGraphics.createShapeBuffer(vertices,2);
			}
			
			log("WebGL init...");
			if( graphics.checkGLErrors() )
				log("...success");
		}
		function glActivateDraw( buffer, indexBuffer, colorBuffer, program, image, blend ){
			if( config.debugWebGL )
				graphics.checkGLErrors();
			
			if( program.vertexAttribCount!=vertexAttribArrayCount ){
				var i;
				//disable the extra attrib arrays
				for( i=program.vertexAttribCount; i<vertexAttribArrayCount; i++ )
					gl.disableVertexAttribArray(i);
				//enable the new attrib arrays
				for( i=vertexAttribArrayCount; i<program.vertexAttribCount; i++ )
					gl.enableVertexAttribArray(i);
				
				vertexAttribArrayCount = program.vertexAttribCount;
			}
		
			updateBlend(blend);
			
			if( usedProgram!=program ){
				usedProgram = program;
				gl.useProgram(program.glProgram);
				boundBuffer = null;//is this necessary?
			}
			if( boundBuffer!=buffer ){
				gl.bindBuffer(gl.ARRAY_BUFFER, buffer);
				gl.vertexAttribPointer(program.vertexAttributes.vertexPositionAttribute.location, buffer.itemSize, gl.FLOAT, false, 0, 0);
				boundBuffer = buffer;
			}
			
			if( boundIndexBuffer!=indexBuffer && (boundIndexBuffer || indexBuffer ) ){/*null!==undefined*/
				if( indexBuffer ){
					gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, indexBuffer);
					boundIndexBuffer = indexBuffer;
				}
				else{
					gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, null);
					boundIndexBuffer = null;
				}
			}
			
			if( colorBuffer ){
				gl.bindBuffer(gl.ARRAY_BUFFER, colorBuffer);
				gl.vertexAttribPointer(program.vertexAttributes.vertexColorAttribute.location, colorBuffer.itemSize, gl.FLOAT, false, 0, 0);
				boundBuffer = null;
				currentImage = null;
				//gl.bindTexture(gl.TEXTURE_2D, null);
			}
			
			var texture;
			if( program.useTexture ){
				if( currentImage!=image )
					currentImage = image;
				if( image ){ 
					if( !image.src )
						texture = image;
					else{
						if( !currentImage.texture )
							currentImage.texture = graphics.createTexture(currentImage);
						texture = currentImage.texture;
					}
					if( boundTexture!=texture ){
						//gl.activeTexture(gl.TEXTURE0);//@todo in init
						gl.bindTexture(gl.TEXTURE_2D, texture);
						boundTexture=texture;
					}
				}
				else{
					currentImage = null;
					if( boundTexture ){
						gl.bindTexture(gl.TEXTURE_2D, null);
						boundTexture=null;
					}
				}
			}
			else{
				currentImage = null;
				if( boundTexture ){
					gl.bindTexture(gl.TEXTURE_2D, null);
					boundTexture=null;
				}
			}
			
			for( var u in program.uniforms )
				program.uniforms[u].load();
			
			if( config.debugWebGL )
				graphics.checkGLErrors();
		}
		function updateBlend(blend){
			/*var blendEnabled = gl.isEnabled(gl.BLEND);
			if( (blending && !blendEnabled) ||  (!blending && blendEnabled) )
				debugger;*/
			if( ( blending && !blend ) || (!blending && blend) ){
				blending = blend;
				if( blend )
					gl.enable( gl.BLEND );
				else
					gl.disable( gl.BLEND );
			}
		}
		function changeActiveColor( program, color ){
			//shaderVars.currentColor = color;
			vec4.copy(shaderVars.currentColor, color);
			if( program.uniforms.fillColorUniform )
				program.uniforms.fillColorUniform.load();
		}
	}
};

if( config.useWebGL )
	graphics = WebGLGraphics;