/* global vec2, vec3, vec4, log, error */
/* exported Shaders */
/* jshint curly: false */
/* jslint multistr: true */

var SHADER_DISCARD = function(){
	return config.useShaderDiscard?"if( alpha<0.01 ) discard;":"";
};

var Shaders = {
	init: function( gl, shaderVars ){
		var whiteColor = vec4.clone([1,1,1,1]);
		Shaders.AA_BLUR = config.shaderAAStrength.toString();//used in the shaders with smoothstep
		//Shaders.AA_EXTRUDE = 1.2;//different than AA_BLUR because unlike smoothstep it uses linear interpolation
		var tmp = vec4.create();
		var shaders = {
			vShader:{
				type: gl.VERTEX_SHADER,
				source:"attribute vec2 aVertexPosition;\
						uniform mat3 uTransformMatrix;\
						void main(void) {\
							gl_Position = vec4(uTransformMatrix * vec3(aVertexPosition, 1), 1);\
						}",
				uniforms:{
					trMatrixUniform:{
						name: "uTransformMatrix",
						load: function(){loadUniformMatrix3fv(this, shaderVars.trMatrix);}
					}
				}
				//transform: transformVertex
			},
			vShaderPolar:{
				type: gl.VERTEX_SHADER,
				source: 
					"attribute vec2 aVertexPosition;\
					uniform mat3 uTransformMatrix;\
					uniform vec2 uPolarCenter;\
					uniform float uPolarXToAngle;\
					uniform float uScreenAspect;\
					void main(void) {\
						vec2 pos = (uTransformMatrix * vec3(aVertexPosition, 1)).xy - uPolarCenter;\
						float angle = pos.x*uPolarXToAngle;\
						gl_Position = vec4(pos.y*sin(angle)*uScreenAspect+uPolarCenter.x,pos.y*cos(angle)+uPolarCenter.y,0,1);\
					}",
				uniforms:{
					trMatrixUniform:{
						name: "uTransformMatrix",
						load: function(){loadUniformMatrix3fv(this, shaderVars.trMatrix);}
					},
					polarCenterUniform : {
						name: "uPolarCenter",
						load: function(){loadUniform2fv(this, shaderVars.polarCenter );}
					},
					polarXToAngleUniform : {
						name: "uPolarXToAngle",
						load: function(){loadUniform1f(this, shaderVars.polarXToAngle );}
					},
					screenAspectUniform : {
						name: "uScreenAspect",
						load: function(){loadUniform1f(this, shaderVars.screenAspect );}
					}
				}
				//transform: transformVertexPolar
			},
			vShaderTex:{
				//a vertex shader that outputs the input coordinate as a texture coordinate
				type: gl.VERTEX_SHADER,
				source:
					"attribute vec2 aVertexPosition;\
					uniform mat3 uTransformMatrix;\
					varying vec2 vTexCoord;\
					void main(void) {\
						gl_Position = vec4(uTransformMatrix * vec3(aVertexPosition, 1), 1);\
						vTexCoord = aVertexPosition;\
					}",
				uniforms:{
					trMatrixUniform:{
						name: "uTransformMatrix",
						load: function(){loadUniformMatrix3fv(this, shaderVars.trMatrix);}
					}
				},
				//transform: transformVertex
			},
			vShaderTexPolar:{
				type: gl.VERTEX_SHADER,
				source:
					"attribute vec2 aVertexPosition;\n\
					uniform mat3 uTransformMatrix;\
					uniform vec2 uPolarCenter;\
					uniform float uPolarXToAngle;\
					uniform float uScreenAspect;\
					varying vec2 vTexCoord;\
					void main(void) {\
						vec2 pos = (uTransformMatrix * vec3(aVertexPosition, 1)).xy - uPolarCenter;\
						float angle = pos.x*uPolarXToAngle;\
						gl_Position = vec4(pos.y*sin(angle)*uScreenAspect+uPolarCenter.x,pos.y*cos(angle)+uPolarCenter.y,0,1);\
						vTexCoord = aVertexPosition;\
					}",
				uniforms:{
					trMatrixUniform:{
						name: "uTransformMatrix",
						load: function(){loadUniformMatrix3fv(this, shaderVars.trMatrix);}
					},
					polarCenterUniform : {
						name: "uPolarCenter",
						load: function(){loadUniform2fv(this, shaderVars.polarCenter );}
					},
					polarXToAngleUniform : {
						name: "uPolarXToAngle",
						load: function(){loadUniform1f(this, shaderVars.polarXToAngle );}
					},
					screenAspectUniform : {
						name: "uScreenAspect",
						load: function(){loadUniform1f(this, shaderVars.screenAspect );}
					}
				}
				//transform: transformVertexPolar
			},
			vShaderTexCoord:{
				//a vertex shader that outputs the input coordinate as a texture coordinate
				type: gl.VERTEX_SHADER,
				source:
					"attribute vec4 aVertexPosition;\
					uniform mat3 uTransformMatrix;\
					varying vec2 vTexCoord;\
					void main(void) {\
						gl_Position = vec4(uTransformMatrix * vec3(aVertexPosition.xy, 1), 1);\
						vTexCoord = aVertexPosition.zw;\
					}",
				uniforms:{
					trMatrixUniform:{
						name: "uTransformMatrix",
						load: function(){loadUniformMatrix3fv(this, shaderVars.trMatrix);}
					}
				},
				positionItemSize: 4,//using 4 floats per vertex instead of the default 2
				//transform: transformVertex
			},
			vShaderTexCoordPolar:{
				//a vertex shader that outputs the input coordinate as a texture coordinate
				type: gl.VERTEX_SHADER,
				source:
					"attribute vec4 aVertexPosition;\
					uniform mat3 uTransformMatrix;\
					uniform vec2 uPolarCenter;\
					uniform float uPolarXToAngle;\
					uniform float uScreenAspect;\
					varying vec2 vTexCoord;\
					void main(void) {\
						vec2 pos = (uTransformMatrix * vec3(aVertexPosition.xy, 1)).xy - uPolarCenter;\
						float angle = pos.x*uPolarXToAngle;\
						gl_Position = vec4(pos.y*sin(angle)*uScreenAspect+uPolarCenter.x,pos.y*cos(angle)+uPolarCenter.y,0,1);\
						vTexCoord = aVertexPosition.zw;\
					}",
				positionItemSize: 4,//using 4 floats per vertex instead of the default 2
				uniforms:{
					trMatrixUniform:{
						name: "uTransformMatrix",
						load: function(){loadUniformMatrix3fv(this, shaderVars.trMatrix);}
					},
					polarCenterUniform : {
						name: "uPolarCenter",
						load: function(){loadUniform2fv(this, shaderVars.polarCenter );}
					},
					polarXToAngleUniform : {
						name: "uPolarXToAngle",
						load: function(){loadUniform1f(this, shaderVars.polarXToAngle );}
					},
					screenAspectUniform : {
						name: "uScreenAspect",
						load: function(){loadUniform1f(this, shaderVars.screenAspect );}
					}
				}
				//transform: transformVertex
			},
			vShaderVertexColor:{
				//Shader program : vertex color, has no polar version
				type: gl.VERTEX_SHADER,
				source:
					"attribute vec2 aVertexPosition;\
					attribute vec4 aVertexColor;\
					uniform mat3 uTransformMatrix;\
					varying vec4 vColor;\
					void main(void) {\
						gl_Position = vec4(uTransformMatrix * vec3(aVertexPosition, 1), 1);\
						vColor = aVertexColor;\
					}",
				uniforms:{
					trMatrixUniform:{
						name: "uTransformMatrix",
						load: function(){loadUniformMatrix3fv(this, shaderVars.trMatrix);}
					}
				},
				vertexAttributes:{
					vertexColorAttribute:{
						name: "aVertexColor"
					}
				},
				//transform: transformVertex
			},
			vShaderInterpolatedVertexColor:{
				type: gl.VERTEX_SHADER,
				source:
					"attribute vec2 aVertexPosition;\
					attribute vec4 aVertexColor;\
					uniform mat3 uTransformMatrix;\
					varying vec4 vColor;\
					uniform vec4 uInterpolateColor;\
					void main(void) {\
						gl_Position = vec4(uTransformMatrix * vec3(aVertexPosition, 1), 1);\
						vColor = vec4( aVertexColor.rgb*uInterpolateColor.a+uInterpolateColor.rgb*(1.0-uInterpolateColor.a), aVertexColor.a );\
					}",
				uniforms:{
					trMatrixUniform:{
						name: "uTransformMatrix",
						load: function(){loadUniformMatrix3fv(this, shaderVars.trMatrix);}
					},
					interpolateColorUniform:{
						name: "uInterpolateColor",
						load: function(){loadUniform4fv(this, shaderVars.currentColor );}
					}
				},
				vertexAttributes:{
					vertexColorAttribute:{
						name: "aVertexColor"
					}
				},
			},
			vShaderVertexColorPolar:{
				//Shader program : vertex color, has no polar version
				type: gl.VERTEX_SHADER,
				source:
					"attribute vec2 aVertexPosition;\
					attribute vec4 aVertexColor;\
					uniform mat3 uTransformMatrix;\
					uniform vec2 uPolarCenter;\
					uniform float uPolarXToAngle;\
					uniform float uScreenAspect;\
					varying vec4 vColor;\
					void main(void) {\
						vec2 pos = (uTransformMatrix * vec3(aVertexPosition, 1)).xy - uPolarCenter;\
						float angle = pos.x*uPolarXToAngle;\
						gl_Position = vec4(pos.y*sin(angle)*uScreenAspect+uPolarCenter.x,pos.y*cos(angle)+uPolarCenter.y,0,1);\
						vColor = aVertexColor;\
					}",
				uniforms:{
					trMatrixUniform:{
						name: "uTransformMatrix",
						load: function(){loadUniformMatrix3fv(this, shaderVars.trMatrix);}
					},
					polarCenterUniform : {
						name: "uPolarCenter",
						load: function(){loadUniform2fv(this, shaderVars.polarCenter );}
					},
					polarXToAngleUniform : {
						name: "uPolarXToAngle",
						load: function(){loadUniform1f(this, shaderVars.polarXToAngle );}
					},
					screenAspectUniform : {
						name: "uScreenAspect",
						load: function(){loadUniform1f(this, shaderVars.screenAspect );}
					}
				},
				vertexAttributes:{
					vertexColorAttribute:{
						name: "aVertexColor"
					}
				},
			},
			vShaderInterpolatedVertexColorPolar:{
				//Shader program : vertex color, has no polar version
				type: gl.VERTEX_SHADER,
				source:
					"attribute vec2 aVertexPosition;\
					attribute vec4 aVertexColor;\
					uniform mat3 uTransformMatrix;\
					uniform vec2 uPolarCenter;\
					uniform float uPolarXToAngle;\
					uniform float uScreenAspect;\
					varying vec4 vColor;\
					uniform vec4 uInterpolateColor;\
					void main(void) {\
						vec2 pos = (uTransformMatrix * vec3(aVertexPosition, 1)).xy - uPolarCenter;\
						float angle = pos.x*uPolarXToAngle;\
						gl_Position = vec4(pos.y*sin(angle)*uScreenAspect+uPolarCenter.x,pos.y*cos(angle)+uPolarCenter.y,0,1);\
						vColor = vec4( aVertexColor.rgb*(1.0-uInterpolateColor.a)+uInterpolateColor.rgb*uInterpolateColor.a, aVertexColor.a );\
					}",
				uniforms:{
					trMatrixUniform:{
						name: "uTransformMatrix",
						load: function(){loadUniformMatrix3fv(this, shaderVars.trMatrix);}
					},
					polarCenterUniform : {
						name: "uPolarCenter",
						load: function(){loadUniform2fv(this, shaderVars.polarCenter );}
					},
					polarXToAngleUniform : {
						name: "uPolarXToAngle",
						load: function(){loadUniform1f(this, shaderVars.polarXToAngle );}
					},
					screenAspectUniform : {
						name: "uScreenAspect",
						load: function(){loadUniform1f(this, shaderVars.screenAspect );}
					},
					interpolateColorUniform:{
						name: "uInterpolateColor",
						load: function(){loadUniform4fv(this, shaderVars.currentColor );}
					}
				},
				vertexAttributes:{
					vertexColorAttribute:{
						name: "aVertexColor"
					}
				},
			},
			vShaderTexTextransformed:{
				//Shader program : cropped texture
				type: gl.VERTEX_SHADER,
				source:
					"attribute vec2 aVertexPosition;\
					uniform mat3 uTransformMatrix;\
					uniform mat3 uTextureMatrix;\
					varying vec2 vTexCoord;\
					void main(void) {\
						vec3 pos = vec3(aVertexPosition, 1);\
						gl_Position = vec4(uTransformMatrix * pos, 1);\
						vTexCoord = (uTextureMatrix * pos).xy;\
					}",
				uniforms:{
					trMatrixUniform:{
						name: "uTransformMatrix",
						load: function(){loadUniformMatrix3fv(this, shaderVars.trMatrix);}
					},
					textureMatrixUniform: {
						name: "uTextureMatrix",
						load: function(){loadUniformMatrix3fv(this, shaderVars.textureMatrix);}
					}
				}
				//transform: transformVertex
			},
			vShaderTexTextransformedPolar:{
				//Shader program : cropped texture
				type: gl.VERTEX_SHADER,
				source:
					"attribute vec2 aVertexPosition;\
					uniform mat3 uTransformMatrix;\
					uniform mat3 uTextureMatrix;\
					uniform vec2 uPolarCenter;\
					uniform float uPolarXToAngle;\
					uniform float uScreenAspect;\
					varying vec2 vTexCoord;\
					void main(void) {\
						vec2 pos = (uTransformMatrix * vec3(aVertexPosition, 1)).xy - uPolarCenter;\
						float angle = pos.x*uPolarXToAngle;\
						gl_Position = vec4(pos.y*sin(angle)*uScreenAspect+uPolarCenter.x,pos.y*cos(angle)+uPolarCenter.y,0,1);\
						vTexCoord = (uTextureMatrix * vec3(aVertexPosition, 1)).xy;\
					}",
				uniforms:{
					trMatrixUniform:{
						name: "uTransformMatrix",
						load: function(){loadUniformMatrix3fv(this, shaderVars.trMatrix);}
					},
					polarCenterUniform : {
						name: "uPolarCenter",
						load: function(){loadUniform2fv(this, shaderVars.polarCenter );}
					},
					polarXToAngleUniform : {
						name: "uPolarXToAngle",
						load: function(){loadUniform1f(this, shaderVars.polarXToAngle );}
					},
					screenAspectUniform : {
						name: "uScreenAspect",
						load: function(){loadUniform1f(this, shaderVars.screenAspect );}
					},
					textureMatrixUniform: {
						name: "uTextureMatrix",
						load: function(){loadUniformMatrix3fv(this, shaderVars.textureMatrix);}
					}
				}
				//transform: transformVertexPolar
			},
			fShaderColorFill:{
				type: gl.FRAGMENT_SHADER,
				source:
					"precision mediump float;\
					uniform vec4 uColor;\
					void main(void){\
						gl_FragColor = uColor;\
					}",
				uniforms:{
					fillColorUniform:{
						name: "uColor",
						load: function(){loadUniform4fv(this, shaderVars.currentColor );}
					}
				}
			},
			fShaderVertexColor: {
				type: gl.FRAGMENT_SHADER,
				source:
					"precision mediump float;\
					varying vec4 vColor;\
					void main(void){\
						gl_FragColor = vColor;\
					}",
				/*uniforms:{
					textureTintColorUniform:{
						name: "uColor",
						load: function(){loadUniform4fv(this, shaderVars.textureTintEnabled?shaderVars.textureTintColor:whiteColor );}
					}
				}*/
			},
			fShaderTexture: {
				type: gl.FRAGMENT_SHADER,
				source:
					"precision mediump float;\
					uniform sampler2D uImage;\
					varying vec2 vTexCoord;\
					uniform vec4 uColor;\
					void main(void){\
						gl_FragColor = uColor*texture2D(uImage, vTexCoord);\
					}",
				uniforms:{
					textureUniform:{
						name: "uImage",
						load: function(){loadUniform1i(this, 0);}
					},
					textureTintColorUniform:{
						name: "uColor",
						load: function(){loadUniform4fv(this, shaderVars.textureTintEnabled?shaderVars.textureTintColor:whiteColor );}
					}
				}
			},
			fShaderCircleClip:{
				type: gl.FRAGMENT_SHADER,
				source:
					"precision mediump float;\
					const float cAntialiasBlur="+Shaders.AA_BLUR+";"+
					"uniform vec4 uColor;\
					uniform vec3 uClipCircle;\
					void main(void){\
						float alpha = smoothstep( uClipCircle.z+cAntialiasBlur, uClipCircle.z, distance( vec2(gl_FragCoord), vec2(uClipCircle) ) );"+SHADER_DISCARD()+"\
						gl_FragColor = vec4(uColor.rgb,alpha);\
					}",
				uniforms:{
					fillColorUniform:{
						name: "uColor",
						load: function(){loadUniform4fv(this, shaderVars.currentColor );}
					},
					clipCircleUniform:{
						name: "uClipCircle",
						load: function(){loadUniform3f(this, shaderVars.clipCircle[0], shaderVars.clipCircle[1], shaderVars.clipCircle[2]-Shaders.AA_BLUR/2 );}
					}
				}
			},
			fShaderCircleClipVertexColor:{
				type: gl.FRAGMENT_SHADER,
				sourceTest:
					"precision mediump float;\
					const float cAntialiasBlur="+Shaders.AA_BLUR+";"+
					"uniform vec4 uColor;\
					uniform vec3 uClipCircle;\
					void main(void){\
						float l = 16.0*distance( vec2(gl_FragCoord)/16.0, vec2(uClipCircle)/16.0 )/uClipCircle.z;\
						gl_FragColor = vec4( mod(5.0*l,1.0), mod(l,1.0), mod(l/5.0,1.0), 1);\
					}",
				source:
					"precision mediump float;\
					const float cAntialiasBlur="+Shaders.AA_BLUR+";"+
					"varying vec4 vColor;\
					uniform vec3 uClipCircle;\
					void main(void){\
						float alpha = vColor.a*smoothstep( uClipCircle.z+cAntialiasBlur, uClipCircle.z, 16.0*distance( vec2(gl_FragCoord)*(1.0/16.0), vec2(uClipCircle)*(1.0/16.0) ) );"+SHADER_DISCARD()+"\
						gl_FragColor = vec4(vColor.rgb,alpha);\
					}",
				uniforms:{
					clipCircleUniform:{
						name: "uClipCircle",
						load: function(){loadUniform3f(this, shaderVars.clipCircle[0], shaderVars.clipCircle[1], shaderVars.clipCircle[2]-Shaders.AA_BLUR/2 );}
					}
				}
			},
			fShaderTextureRingClip:{
				type: gl.FRAGMENT_SHADER,
				source:
					"precision mediump float;\
					const float cAntialiasBlur="+Shaders.AA_BLUR+"/2.0;"+
					"const float M_PI=3.1415926535897932384626433832795;\
					uniform sampler2D uImage;\
					varying vec2 vTexCoord;\
					uniform vec4 uClipConcentric;\
					void main(void){\
						float dist = 16.0*distance( vec2(gl_FragCoord)*(1.0/16.0), vec2(uClipConcentric)*(1.0/16.0) );\
						float alpha = smoothstep( uClipConcentric[2]-cAntialiasBlur, uClipConcentric[2]+cAntialiasBlur, dist ) - smoothstep( uClipConcentric[3]-cAntialiasBlur, uClipConcentric[3]+cAntialiasBlur, dist );"+SHADER_DISCARD()+"\
						gl_FragColor = texture2D(uImage, vTexCoord);\
						gl_FragColor.a*=alpha;\
					}",
				uniforms:{
					clipConcentricUniform:{
						name: "uClipConcentric",
						load: function(){loadUniform4fv(this, shaderVars.clipConcentric);}
					},
					textureUniform:{
						name: "uImage",
						load: function(){loadUniform1i(this, 0);}
					},
				}
			},
			fShaderPie:{
				/*This is the only pie shader that handles the condition when pieRange[0]>pieRange[1]*/
				type: gl.FRAGMENT_SHADER,
				source:
					"precision mediump float;\
					const float cAntialiasBlur="+Shaders.AA_BLUR+"/2.0;"+
					"const float M_PI=3.1415926535897932384626433832795;\
					uniform vec4 uColor;\
					uniform vec2 uRange;\
					uniform vec4 uClipConcentric;\
					void main(void){\
						vec2 rvec = vec2(gl_FragCoord)-vec2(uClipConcentric);\
						float dist = 16.0*length( rvec/16.0 );\
						float alpha = smoothstep( uClipConcentric[2]-cAntialiasBlur, uClipConcentric[2]+cAntialiasBlur, dist ) - smoothstep( uClipConcentric[3]-cAntialiasBlur, uClipConcentric[3]+cAntialiasBlur, dist );\
						float angle = atan(rvec.x,rvec.y);\
						float angleBlur = cAntialiasBlur/(2.0*dist);\
						alpha *= sign( uRange[1]-uRange[0] )*(smoothstep( uRange[0]-angleBlur, uRange[0]+angleBlur, angle ) - smoothstep( uRange[1]-angleBlur, uRange[1]+angleBlur, angle ) );\
						alpha *= uColor.a;"+SHADER_DISCARD()+"\
						gl_FragColor = vec4(uColor.rgb,alpha);\
					}",
				uniforms:{
					fillColorUniform:{
						name: "uColor",
						load: function(){loadUniform4fv(this, shaderVars.currentColor );}
					},
					rangeUniform:{
						name:"uRange",
						load: function(){loadUniform2fv(this, shaderVars.pieRange );}
					},
					clipConcentricUniform:{
						name: "uClipConcentric",
						load: function(){loadUniform4fv(this, shaderVars.clipConcentric);}
					}
				}
			},
			fShaderPieTicks:{
				/*This is the only pie shader that handles the condition when pieRange[0]>pieRange[1]*/
				type: gl.FRAGMENT_SHADER,
				source:
					"precision mediump float;\
					const float cAntialiasBlur="+Shaders.AA_BLUR+"/2.0;"+
					"const float M_PI=3.1415926535897932384626433832795;\
					const float tickWidth = 0.75*M_PI/360.0;/*0.75 degrees*/\
					uniform float uHalfTickAngle;\
					uniform vec4 uColor;\
					uniform vec2 uRange;\
					uniform vec4 uClipConcentric;\
					void main(void){\
						vec2 rvec = vec2(gl_FragCoord)-vec2(uClipConcentric);\
						float dist = length( rvec );\
						float alpha = uColor.a*(smoothstep( uClipConcentric[2]-cAntialiasBlur, uClipConcentric[2]+cAntialiasBlur, dist ) - smoothstep( uClipConcentric[3]-cAntialiasBlur, uClipConcentric[3]+cAntialiasBlur, dist ));\
						float angle = atan(rvec.x,rvec.y);\
						float angleBlur = cAntialiasBlur/(2.0*dist);\
						alpha *= sign( uRange[1]-uRange[0] )*(smoothstep( uRange[0]-angleBlur, uRange[0]+angleBlur, angle ) - smoothstep( uRange[1]-angleBlur, uRange[1]+angleBlur, angle ) );"+SHADER_DISCARD()+"\
						gl_FragColor = vec4(uColor.rgb,alpha);\
						float trwave = (1.0-abs( 1.0 - mod((angle-uRange[0])/uHalfTickAngle,2.0) ))*uHalfTickAngle;/*a triangle wave, from zero to half the degrees_between_two_ticks*/\
						gl_FragColor.rgb *= (smoothstep(tickWidth-angleBlur,tickWidth+angleBlur,trwave)+1.0)*0.5;/*Add the tick*/\
					}",
				uniforms:{
					fillColorUniform:{
						name: "uColor",
						load: function(){loadUniform4fv(this, shaderVars.currentColor );}
					},
					rangeUniform:{
						name:"uRange",
						load: function(){loadUniform2fv(this, shaderVars.pieRange );}
					},
					clipConcentricUniform:{
						name: "uClipConcentric",
						load: function(){loadUniform4fv(this, shaderVars.clipConcentric);}
					},
					ticksUniform:{
						name: "uHalfTickAngle",
						load: function(){
							loadUniform1f(this, shaderVars.tickAngle/2.0);
						}
					}
				}
			},
			fShaderPieTexturedTicks:{
				type: gl.FRAGMENT_SHADER,
				source:
					//add 1 pixel ticks
					"precision mediump float;\
					const float cAntialiasBlur="+Shaders.AA_BLUR+"/2.0;"+
					"const float M_PI=3.1415926535897932384626433832795;\
					const float tickWidth = M_PI/360.0;/*1 degree*/\
					uniform float uHalfTickAngle;\
					uniform sampler2D uImage;\
					uniform vec2 uRange;\
					uniform vec4 uClipConcentric;\
					void main(void){\
						vec2 rvec = vec2(gl_FragCoord)-vec2(uClipConcentric);\
						float dist = length( rvec );\
						float alpha = smoothstep( uClipConcentric[2]-cAntialiasBlur, uClipConcentric[2]+cAntialiasBlur, dist ) - smoothstep( uClipConcentric[3]-cAntialiasBlur, uClipConcentric[3]+cAntialiasBlur, dist );\
						float angle = atan(rvec.x,rvec.y);\
						float angleBlur = cAntialiasBlur/(2.0*dist);\
						alpha *= sign( uRange[1]-uRange[0] )*(smoothstep( uRange[0]-angleBlur, uRange[0]+angleBlur, angle ) - smoothstep( uRange[1]-angleBlur, uRange[1]+angleBlur, angle ));"+SHADER_DISCARD()+"\
						float trwave = (1.0-abs( 1.0 - mod((angle-uRange[0])/uHalfTickAngle,2.0) ))*uHalfTickAngle;/*a triangle wave, from zero to half the degrees_between_two_ticks*/\
						float utex = (uClipConcentric[3]-dist)/(uClipConcentric[3]-uClipConcentric[2]);\
						gl_FragColor = texture2D(uImage, vec2(utex,0.0));\
						gl_FragColor.rgb *= (smoothstep(tickWidth-angleBlur,tickWidth+angleBlur,trwave)+1.0)*0.5;/*Add the tick*/\
						gl_FragColor.a *= alpha;\
					}",
				uniforms:{
					textureUniform:{
						name: "uImage",
						load: function(){loadUniform1i(this, 0);}
					},
					rangeUniform:{
						name:"uRange",
						load: function(){loadUniform2fv(this, shaderVars.pieRange );}
					},
					clipConcentricUniform:{
						name: "uClipConcentric",
						load: function(){loadUniform4fv(this, shaderVars.clipConcentric);}
					},
					ticksUniform:{
						name: "uHalfTickAngle",
						load: function(){
							loadUniform1f(this, shaderVars.tickAngle/2.0);
						}
					}
				}
			},
			fShaderPieGradient:{
				type: gl.FRAGMENT_SHADER,
				source:
					"precision mediump float;\
					const float cAntialiasBlur="+Shaders.AA_BLUR+"/2.0;"+
					"const float M_PI=3.1415926535897932384626433832795;\
					uniform vec4 uColor1;\
					uniform vec4 uColor2;\
					uniform vec2 uRange;\
					uniform vec4 uClipConcentric;\
					void main(void){\
						vec2 rvec = vec2(gl_FragCoord)-vec2(uClipConcentric);\
						float dist = length( rvec );\
						float alpha = smoothstep( uClipConcentric[2]-cAntialiasBlur, uClipConcentric[2]+cAntialiasBlur, dist ) - smoothstep( uClipConcentric[3]-cAntialiasBlur, uClipConcentric[3]+cAntialiasBlur, dist );\
						float angle = atan(rvec.x,rvec.y);\
						float angleBlur = cAntialiasBlur/(2.0*dist);\
						alpha *= smoothstep( uRange[0]-angleBlur, uRange[0]+angleBlur, angle ) - smoothstep( uRange[1]-angleBlur, uRange[1]+angleBlur, angle );"+SHADER_DISCARD()+"\
						gl_FragColor = mix( uColor1, uColor2, (angle-uRange[0])/(uRange[1]-uRange[0]) );\
						gl_FragColor.w *= alpha;\
					}",
				uniforms:{
					fillColor1Uniform:{
						name: "uColor1",
						load: function(){loadUniform4fv(this, shaderVars.color1 );}
					},
					fillColor2Uniform:{
						name: "uColor2",
						load: function(){loadUniform4fv(this, shaderVars.color2 );}
					},
					rangeUniform:{
						name:"uRange",
						load: function(){loadUniform2fv(this, shaderVars.pieRange );}
					},
					clipConcentricUniform:{
						name: "uClipConcentric",
						load: function(){loadUniform4fv(this, shaderVars.clipConcentric);}
					}
				}
			},
			fShaderPieTextured:{//vtex is actually not needed in our use-cases
				type: gl.FRAGMENT_SHADER,
				source:
					"precision mediump float;\
					const float cAntialiasBlur="+Shaders.AA_BLUR+"/2.0;"+
					"const float M_PI=3.1415926535897932384626433832795;\
					uniform sampler2D uImage;\
					uniform vec2 uRange;\
					uniform vec4 uClipConcentric;\
					void main(void){\
						vec2 rvec = vec2(gl_FragCoord)-vec2(uClipConcentric);\
						float dist = length( rvec );\
						float alpha = smoothstep( uClipConcentric[2]-cAntialiasBlur, uClipConcentric[2]+cAntialiasBlur, dist ) - smoothstep( uClipConcentric[3]-cAntialiasBlur, uClipConcentric[3]+cAntialiasBlur, dist );\
						float angle = atan(rvec.x,rvec.y);\
						float angleBlur = cAntialiasBlur/(2.0*dist);\
						alpha *= smoothstep( uRange[0]-angleBlur, uRange[0]+angleBlur, angle ) - smoothstep( uRange[1]-angleBlur, uRange[1]+angleBlur, angle );"+SHADER_DISCARD()+"\
						float utex = (uClipConcentric[3]-dist)/(uClipConcentric[3]-uClipConcentric[2]);\
						float vtex = (angle-uRange[0])/(uRange[1]-uRange[0]);\
						gl_FragColor = texture2D(uImage, vec2(utex,vtex));\
						gl_FragColor.a *= alpha;\
					}",
				uniforms:{
					textureUniform:{
						name: "uImage",
						load: function(){loadUniform1i(this, 0);}
					},
					rangeUniform:{
						name:"uRange",
						load: function(){loadUniform2fv(this, shaderVars.pieRange );}
					},
					clipConcentricUniform:{
						name: "uClipConcentric",
						load: function(){loadUniform4fv(this, shaderVars.clipConcentric);}
					}
				}
			},
			fShaderPieRanges:{
				type: gl.FRAGMENT_SHADER,
				source:
					"precision mediump float;\
					const float cAntialiasBlur="+Shaders.AA_BLUR+"/2.0;"+
					"const float M_PI=3.1415926535897932384626433832795;\
					uniform sampler2D uImage;\
					uniform vec4 uRanges;\
					uniform vec4 uClipConcentric;\
					void main(void){\
						vec2 rvec = vec2(gl_FragCoord)-vec2(uClipConcentric);\
						float dist = length( rvec );\
						float alpha = smoothstep( uClipConcentric[2]-cAntialiasBlur, uClipConcentric[2]+cAntialiasBlur, dist ) - smoothstep( uClipConcentric[3]-cAntialiasBlur, uClipConcentric[3]+cAntialiasBlur, dist );\
						float angle = atan(rvec.x,rvec.y);\
						float angleBlur = cAntialiasBlur/(2.0*dist);\
						alpha *= smoothstep( uRanges[0]-angleBlur, uRanges[0]+angleBlur, angle ) - smoothstep( uRanges[3]-angleBlur, uRanges[3]+angleBlur, angle );"+SHADER_DISCARD()+"\
						float utex = (uClipConcentric[3]-dist)/(uClipConcentric[3]-uClipConcentric[2]);\
						float vtex = 0.33/2.0 + smoothstep( uRanges[1]-angleBlur, uRanges[1]+angleBlur, angle )*0.33 + smoothstep( uRanges[2]-angleBlur, uRanges[2]+angleBlur, angle )*0.33;\
						gl_FragColor = texture2D(uImage, vec2(utex,vtex));\
						gl_FragColor.a *= alpha;\
					}",
				uniforms:{
					textureUniform:{
						name: "uImage",
						load: function(){loadUniform1i(this, 0);}
					},
					rangeUniform:{
						name:"uRanges",
						load: function(){loadUniform4fv(this, shaderVars.pieMultiRanges );}
					},
					clipConcentricUniform:{
						name: "uClipConcentric",
						load: function(){loadUniform4fv(this, shaderVars.clipConcentric);}
					}
				}
			},
			fShaderPieRangesTicks:{
				type: gl.FRAGMENT_SHADER,
				source:
					//add 1 pixel ticks
					"precision mediump float;\
					const float cAntialiasBlur="+Shaders.AA_BLUR+"/2.0;"+
					"const float M_PI=3.1415926535897932384626433832795;\
					const float tickWidth = M_PI/360.0;/*1 degree*/\
					uniform float uHalfTickAngle;\
					uniform sampler2D uImage;\
					uniform vec4 uRanges;\
					uniform vec4 uClipConcentric;\
					void main(void){\
						vec2 rvec = vec2(gl_FragCoord)-vec2(uClipConcentric);\
						float dist = length( rvec );\
						float alpha = smoothstep( uClipConcentric[2]-cAntialiasBlur, uClipConcentric[2]+cAntialiasBlur, dist ) - smoothstep( uClipConcentric[3]-cAntialiasBlur, uClipConcentric[3]+cAntialiasBlur, dist );\
						float angle = atan(rvec.x,rvec.y);\
						float angleBlur = cAntialiasBlur/(2.0*dist);\
						alpha *= smoothstep( uRanges[0]-angleBlur, uRanges[0]+angleBlur, angle ) - smoothstep( uRanges[3]-angleBlur, uRanges[3]+angleBlur, angle );"+SHADER_DISCARD()+"\
						float trwave = (1.0-abs( 1.0 - mod((angle-uRanges[0])/uHalfTickAngle,2.0) ))*uHalfTickAngle;/*a triangle wave, from zero to half the degrees_between_two_ticks*/\
						float utex = (uClipConcentric[3]-dist)/(uClipConcentric[3]-uClipConcentric[2]);\
						float vtex = 0.33/2.0 + smoothstep( uRanges[1]-angleBlur, uRanges[1]+angleBlur, angle )*0.33 + smoothstep( uRanges[2]-angleBlur, uRanges[2]+angleBlur, angle )*0.33;\
						gl_FragColor = texture2D(uImage, vec2(utex,vtex));\
						gl_FragColor.rgb *= (smoothstep(tickWidth-angleBlur,tickWidth+angleBlur,trwave)+1.0)*0.5;/*Add the tick*/\
						gl_FragColor.a *= alpha;\
					}",
				uniforms:{
					textureUniform:{
						name: "uImage",
						load: function(){loadUniform1i(this, 0);}
					},
					rangeUniform:{
						name:"uRanges",
						load: function(){loadUniform4fv(this, shaderVars.pieMultiRanges );}
					},
					clipConcentricUniform:{
						name: "uClipConcentric",
						load: function(){loadUniform4fv(this, shaderVars.clipConcentric);}
					},
					ticksUniform:{
						name: "uHalfTickAngle",
						load: function(){
							loadUniform1f(this, shaderVars.tickAngle/2.0);
						}
					}
				}
			},
			fShaderPie4Ranges:{
				type: gl.FRAGMENT_SHADER,
				source:
					//add 1 pixel ticks
					"precision mediump float;\
					const float cAntialiasBlur="+Shaders.AA_BLUR+"/2.0;"+
					"const float M_PI=3.1415926535897932384626433832795;\
					const float tickWidth = M_PI/360.0;/*1 degree*/\
					uniform sampler2D uImage;\
					uniform vec2 uRange;\
					uniform vec3 uColorRanges;\
					uniform vec4 uClipConcentric;\
					void main(void){\
						vec2 rvec = vec2(gl_FragCoord)-vec2(uClipConcentric);\
						float dist = length( rvec );\
						float alpha = smoothstep( uClipConcentric[2]-cAntialiasBlur, uClipConcentric[2]+cAntialiasBlur, dist ) - smoothstep( uClipConcentric[3]-cAntialiasBlur, uClipConcentric[3]+cAntialiasBlur, dist );\
						float angle = atan(rvec.x,rvec.y);\
						float angleBlur = cAntialiasBlur/(2.0*dist);\
						alpha *= smoothstep( uRange[0]-angleBlur, uRange[0]+angleBlur, angle ) - smoothstep( uRange[1]-angleBlur, uRange[1]+angleBlur, angle );"+SHADER_DISCARD()+"\
						float utex = (uClipConcentric[3]-dist)/(uClipConcentric[3]-uClipConcentric[2]);\
						float vtex = 0.25/2.0 + smoothstep( uColorRanges[0]-angleBlur, uColorRanges[0]+angleBlur, angle )*0.25 + smoothstep( uColorRanges[1]-angleBlur, uColorRanges[1]+angleBlur, angle )*0.25 + smoothstep( uColorRanges[2]-angleBlur, uColorRanges[2]+angleBlur, angle )*0.25;\
						gl_FragColor = texture2D(uImage, vec2(utex,vtex));\
						gl_FragColor.a *= alpha;\
					}",
				uniforms:{
					textureUniform:{
						name: "uImage",
						load: function(){loadUniform1i(this, 0);}
					},
					rangeUniform:{
						name:"uRange",
						load: function(){loadUniform2fv(this, shaderVars.pieRange );}
					},
					colorRangesUniform:{
						name:"uColorRanges",
						load: function(){loadUniform3fv(this, shaderVars.pieColor3Ranges );}
					},
					clipConcentricUniform:{
						name: "uClipConcentric",
						load: function(){loadUniform4fv(this, shaderVars.clipConcentric);}
					}
				}
			},
			fShaderPie4RangesTicks:{
				type: gl.FRAGMENT_SHADER,
				source:
					//add 1 pixel ticks
					"precision mediump float;\
					const float cAntialiasBlur="+Shaders.AA_BLUR+"/2.0;"+
					"const float M_PI=3.1415926535897932384626433832795;\
					const float tickWidth = M_PI/360.0;/*1 degree*/\
					uniform float uHalfTickAngle;\
					uniform sampler2D uImage;\
					uniform vec2 uRange;\
					uniform vec3 uColorRanges;\
					uniform vec4 uClipConcentric;\
					void main(void){\
						vec2 rvec = vec2(gl_FragCoord)-vec2(uClipConcentric);\
						float dist = length( rvec );\
						float alpha = smoothstep( uClipConcentric[2]-cAntialiasBlur, uClipConcentric[2]+cAntialiasBlur, dist ) - smoothstep( uClipConcentric[3]-cAntialiasBlur, uClipConcentric[3]+cAntialiasBlur, dist );\
						float angle = atan(rvec.x,rvec.y);\
						float angleBlur = cAntialiasBlur/(2.0*dist);\
						alpha *= smoothstep( uRange[0]-angleBlur, uRange[0]+angleBlur, angle ) - smoothstep( uRange[1]-angleBlur, uRange[1]+angleBlur, angle );"+SHADER_DISCARD()+"\
						float trwave = (1.0-abs( 1.0 - mod((angle-uRange[0])/uHalfTickAngle,2.0) ))*uHalfTickAngle;/*a triangle wave, from zero to half the degrees_between_two_ticks*/\
						float utex = (uClipConcentric[3]-dist)/(uClipConcentric[3]-uClipConcentric[2]);\
						float vtex = 0.25/2.0 + smoothstep( uColorRanges[0]-angleBlur, uColorRanges[0]+angleBlur, angle )*0.25 + smoothstep( uColorRanges[1]-angleBlur, uColorRanges[1]+angleBlur, angle )*0.25 + smoothstep( uColorRanges[2]-angleBlur, uColorRanges[2]+angleBlur, angle )*0.25;\
						gl_FragColor = texture2D(uImage, vec2(utex,vtex));\
						gl_FragColor.rgb *= (smoothstep(tickWidth-angleBlur,tickWidth+angleBlur,trwave)+1.0)*0.5;/*Add the tick*/\
						gl_FragColor.a *= alpha;\
					}",
				uniforms:{
					textureUniform:{
						name: "uImage",
						load: function(){loadUniform1i(this, 0);}
					},
					rangeUniform:{
						name:"uRange",
						load: function(){loadUniform2fv(this, shaderVars.pieRange );}
					},
					colorRangesUniform:{
						name:"uColorRanges",
						load: function(){loadUniform3fv(this, shaderVars.pieColor3Ranges );}
					},
					clipConcentricUniform:{
						name: "uClipConcentric",
						load: function(){loadUniform4fv(this, shaderVars.clipConcentric);}
					},
					ticksUniform:{
						name: "uHalfTickAngle",
						load: function(){
							loadUniform1f(this, shaderVars.tickAngle/2.0);
						}
					}
				}
			},
			
			fShaderTexturedTicks:{
				type: gl.FRAGMENT_SHADER,
				source:
					//add 1 pixel ticks
					"precision mediump float;\
					uniform sampler2D uImage;\
					varying vec2 vTexCoord;\
					uniform float uTickUnit;\
					uniform float uTickWidth;\
					uniform float uAntialiasBlur;\
					void main(void){\
						float trwave = (1.0-abs( 1.0 - mod(vTexCoord[0]*2.0/uTickUnit,2.0) ))*1.0/*uTickUnit*/;/*a triangle wave, from zero to degrees_between_two_ticks*/\
						trwave = (smoothstep(uTickWidth-uAntialiasBlur,uTickWidth+uAntialiasBlur,trwave));/*the tick*/\
						gl_FragColor = trwave*texture2D(uImage, vec2(vTexCoord[1],vTexCoord[0]));\
					}",
				uniforms:{
					textureUniform:{
						name: "uImage",
						load: function(){loadUniform1i(this, 0);}
					},
					tickUnitUniform:{
						name: "uTickUnit",//as a fraction of the bar
						load: function(){loadUniform1f(this, shaderVars.tickUnit );}
					},
					tickWidthUniform:{
						name: "uTickWidth",//as a fraction of 1 unit
						load: function(){loadUniform1f(this, shaderVars.tickWidth );}
					},
					aaBlurUniform:{
						name: "uAntialiasBlur",//as a fraction of 1 unit
						load: function(){loadUniform1f(this, shaderVars.pixelSize*Shaders.AA_BLUR/(2.0*shaderVars.tickUnit) );}
					}
				}
			},
			fShaderTexturedRangesTicks:{
				type: gl.FRAGMENT_SHADER,
				source:
					"precision mediump float;\
					uniform sampler2D uImage;\
					varying vec2 vTexCoord;\
					uniform vec2 uRange;\
					uniform float uTickUnit;\
					uniform float uTickWidth;\
					uniform float uAntialiasBlur;\
					void main(void){\
						float trwave = (1.0-abs( 1.0 - mod(vTexCoord[0]*2.0/uTickUnit,2.0) ));/*a triangle wave, from zero to degrees_between_two_ticks*/\
						trwave = (smoothstep(uTickWidth-uAntialiasBlur/uTickUnit,uTickWidth+uAntialiasBlur/uTickUnit,trwave));/*the tick*/\
						float vtex = 0.33/2.0 + smoothstep( uRange[0]-uAntialiasBlur, uRange[0]+uAntialiasBlur, vTexCoord[0] )*0.33 + smoothstep( uRange[1]-uAntialiasBlur, uRange[1]+uAntialiasBlur, vTexCoord[0] )*0.33;\
						gl_FragColor = trwave*texture2D(uImage, vec2(vTexCoord[1],vtex));\
					}",
				uniforms:{
					textureUniform:{
						name: "uImage",
						load: function(){loadUniform1i(this, 0);}
					},
					rangeUniform:{
						name:"uRange",
						load: function(){loadUniform2fv(this, shaderVars.barRange );}
					},
					tickUnitUniform:{
						name: "uTickUnit",//as a fraction of the bar
						load: function(){loadUniform1f(this, shaderVars.tickUnit );}
					},
					tickWidthUniform:{
						name: "uTickWidth",//as a fraction of 1 unit
						load: function(){loadUniform1f(this, shaderVars.tickWidth );}
					},
					aaBlurUniform:{
						name: "uAntialiasBlur",
						load: function(){loadUniform1f(this, shaderVars.pixelSize*Shaders.AA_BLUR/2.0 );}
					}
				}
			},
			fShaderTexturedRangesTicks2:{//major and minor ticks
				type: gl.FRAGMENT_SHADER,
				source:
					"precision mediump float;\
					uniform sampler2D uImage;\
					varying vec2 vTexCoord;\
					uniform vec2 uRange;\
					uniform float uTickUnit;\
					uniform float uTickWidth;\
					uniform float uTickUnit2;\
					uniform float uTickWidth2;\
					uniform float uAntialiasBlur;\
					void main(void){\
						float trwave1 = (1.0-abs( 1.0 - mod(vTexCoord[0]*2.0/uTickUnit,2.0) ));/*a triangle wave, from zero to degrees_between_two_ticks*/\
						trwave1 = (smoothstep(uTickWidth-uAntialiasBlur/uTickUnit,uTickWidth+uAntialiasBlur/uTickUnit,trwave1));/*the tick*/\
						float trwave2 = (1.0-abs( 1.0 - mod(vTexCoord[0]*2.0/uTickUnit2,2.0) ));/*a triangle wave, from zero to degrees_between_two_ticks*/\
						trwave2 = (smoothstep(uTickWidth2-uAntialiasBlur/uTickUnit2,uTickWidth2+uAntialiasBlur/uTickUnit2,trwave2));/*the tick*/\
						float vtex = 0.33/2.0 + smoothstep( uRange[0]-uAntialiasBlur, uRange[0]+uAntialiasBlur, vTexCoord[0] )*0.33 + smoothstep( uRange[1]-uAntialiasBlur, uRange[1]+uAntialiasBlur, vTexCoord[0] )*0.33;\
						gl_FragColor = trwave1*trwave2*texture2D(uImage, vec2(vTexCoord[1],vtex));\
					}",
				uniforms:{
					textureUniform:{
						name: "uImage",
						load: function(){loadUniform1i(this, 0);}
					},
					rangeUniform:{
						name:"uRange",
						load: function(){loadUniform2fv(this, shaderVars.barRange );}
					},
					tickUnitUniform:{
						name: "uTickUnit",//as a fraction of the bar
						load: function(){loadUniform1f(this, shaderVars.tickUnit );}
					},
					tickWidthUniform:{
						name: "uTickWidth",//as a fraction of 1 unit
						load: function(){loadUniform1f(this, shaderVars.tickWidth );}
					},
					tickUnitUniform2:{
						name: "uTickUnit2",//as a fraction of the bar
						load: function(){loadUniform1f(this, shaderVars.tickUnit2 );}
					},
					tickWidthUniform2:{
						name: "uTickWidth2",//as a fraction of 1 unit
						load: function(){loadUniform1f(this, shaderVars.tickWidth2 );}
					},
					aaBlurUniform:{
						name: "uAntialiasBlur",
						load: function(){loadUniform1f(this, shaderVars.pixelSize*Shaders.AA_BLUR/2.0 );}
					}
				}
			}
		};

		//not updating the uniforms that have not changed gives a nice speed increase. Stupid drivers
		function loadUniform1i(uniform,val){
			if( uniform.value!=val ){
				gl.uniform1i(uniform.location, val);
				uniform.value = val;
			}
		}
		function loadUniform1f(uniform,val){
			if( uniform.value!=val ){
				gl.uniform1f(uniform.location, val);
				uniform.value = val;
			}
		}
		function loadUniform2fv(uniform,val){
			if( !uniform.value ){
				uniform.value = vec2.clone(val);
				gl.uniform2fv(uniform.location, val);
			}
			else{
				if( uniform.value[0]!=val[0] || uniform.value[1]!=val[1] ){
					vec2.copy(uniform.value,val);
					gl.uniform2fv(uniform.location, val);
				}
			}
		}
		function loadUniform3fv(uniform,val){
			if( !uniform.value ){
				uniform.value = vec3.clone(val);
				gl.uniform3fv(uniform.location, val);
			}
			else{
				if( uniform.value[0]!=val[0] || uniform.value[1]!=val[1] || uniform.value[2]!=val[2] ){
					vec3.copy(uniform.value,val);
					gl.uniform3fv(uniform.location, val);
				}
			}
		}
		function loadUniform3f(uniform,val1,val2,val3){
			if( !uniform.value ){
				uniform.value = vec3.create();
				vec3.set(uniform.value,val1,val2,val3);
				gl.uniform3f(uniform.location, val1,val2,val3);
			}
			else{
				if( uniform.value[0]!=val1 || uniform.value[1]!=val2 || uniform.value[2]!=val3 ){
					vec3.set(uniform.value,val1,val2,val3);
					gl.uniform3f(uniform.location, val1,val2,val3);
				}
			}
		}
		function loadUniform4fv(uniform,val){
			if( !uniform.value ){
				uniform.value = vec4.clone(val);
				gl.uniform4fv(uniform.location, val);
			}
			else{
				if( uniform.value[0]!=val[0] || uniform.value[1]!=val[1] || uniform.value[2]!=val[2] || uniform.value[3]!=val[3]){
					vec4.copy(uniform.value,val);
					gl.uniform4fv(uniform.location, val);
				}
			}
		}
		function loadUniform4f(uniform,val1,val2,val3,val4){
			if( !uniform.value ){
				uniform.value = vec4.create();
				vec4.set(uniform.value,val1,val2,val3,val4);
				gl.uniform4f(uniform.location, val1,val2,val3,val4);
			}
			else{
				if( uniform.value[0]!=val1 || uniform.value[1]!=val2 || uniform.value[2]!=val3 || uniform.value[3]!=val4 ){
					vec4.set(uniform.value,val1,val2,val3,val4);
					gl.uniform4f(uniform.location, val1,val2,val3,val4);
				}
			}
		}
		function loadUniformMatrix3fv(uniform,val){
			//the matrix is never the same
			gl.uniformMatrix3fv(uniform.location, false, val);
		}

		var programs = {
			shaderProgramColorFill:{
				vShader:shaders.vShader,
				fShader:shaders.fShaderColorFill,
			},
			shaderProgramCircleClip:{
				vShader:shaders.vShader,
				fShader:shaders.fShaderCircleClip,
			},
			shaderProgramCircleClipVertexColor:{
				vShader:shaders.vShaderVertexColor,
				fShader:shaders.fShaderCircleClipVertexColor,
			},
			shaderProgramColorFillPolar:{
				vShader:shaders.vShaderPolar,
				fShader:shaders.fShaderColorFill,
			},
			shaderProgramTexture:{
				vShader:shaders.vShaderTex,
				fShader:shaders.fShaderTexture,
				useTexture:true,
			},
			shaderProgramTextureTick:{
				vShader:shaders.vShaderTex,
				fShader:shaders.fShaderTexturedTicks,
				useTexture:true,
			},
			shaderProgramTexTextransformedTick:{
				vShader:shaders.vShaderTexTextransformed,
				fShader:shaders.fShaderTexturedTicks,
				useTexture:true,
			},
			shaderProgramTextureRangesTick:{
				vShader:shaders.vShaderTex,
				fShader:shaders.fShaderTexturedRangesTicks,
				useTexture:true,
			},
			shaderProgramTexTextransformedRangesTick:{
				vShader:shaders.vShaderTexTextransformed,
				fShader:shaders.fShaderTexturedRangesTicks,
				useTexture:true,
			},
			shaderProgramTextureRangesTick2:{
				vShader:shaders.vShaderTex,
				fShader:shaders.fShaderTexturedRangesTicks2,
				useTexture:true,
			},
			shaderProgramTexTextransformedRangesTick2:{
				vShader:shaders.vShaderTexTextransformed,
				fShader:shaders.fShaderTexturedRangesTicks2,
				useTexture:true,
			},
			shaderProgramTextureRingClip:{
				vShader:shaders.vShaderTex,
				fShader:shaders.fShaderTextureRingClip,
				useTexture:true,
			},
			shaderProgramTexCoordRingClip:{
				vShader:shaders.vShaderTexCoord,
				fShader:shaders.fShaderTextureRingClip,
				useTexture:true,
			},
			shaderProgramTexturePolar:{
				vShader:shaders.vShaderTexPolar,
				fShader:shaders.fShaderTexture,
				useTexture:true,
			},
			shaderProgramTextureCoord:{
				vShader:shaders.vShaderTexCoord,
				fShader:shaders.fShaderTexture,
				useTexture:true,
			},
			shaderProgramTextureCoordPolar:{
				vShader:shaders.vShaderTexCoordPolar,
				fShader:shaders.fShaderTexture,
				useTexture:true,
			},
			shaderProgramCroppedTexture:{
				vShader:shaders.vShaderTexTextransformed,
				fShader:shaders.fShaderTexture,
				useTexture:true,
			},
			shaderProgramCroppedTexturePolar:{
				vShader:shaders.vShaderTexTextransformedPolar,
				fShader:shaders.fShaderTexture,
				useTexture:true,
			},
			shaderProgramVertexColor:{
				vShader:shaders.vShaderVertexColor,
				fShader:shaders.fShaderVertexColor,
			},
			shaderProgramVertexColorPolar:{
				vShader:shaders.vShaderVertexColorPolar,
				fShader:shaders.fShaderVertexColor,
			},
			shaderProgramInterpolatedVertexColor:{
				vShader:shaders.vShaderInterpolatedVertexColor,
				fShader:shaders.fShaderVertexColor,
			},
			shaderProgramInterpolatedVertexColorPolar:{
				vShader:shaders.vShaderInterpolatedVertexColorPolar,
				fShader:shaders.fShaderVertexColor,
			},
			shaderProgramPie:{
				vShader:shaders.vShader,
				fShader:shaders.fShaderPie,
			},
			shaderProgramPieTicks:{
				vShader:shaders.vShader,
				fShader:shaders.fShaderPieTicks,
			},
			shaderProgramPieTextured:{
				vShader:shaders.vShader,
				fShader:shaders.fShaderPieTextured,
				useTexture:true,
			},
			shaderProgramPieTexturedTicks:{
				vShader:shaders.vShader,
				fShader:shaders.fShaderPieTexturedTicks,
				useTexture:true,
			},
			shaderProgramPieRanges:{//@todo - rename pie3Ranges
				vShader:shaders.vShader,
				fShader:shaders.fShaderPieRanges,
				useTexture:true,
			},
			shaderProgramPieRangesTicks:{
				vShader:shaders.vShader,
				fShader:shaders.fShaderPieRangesTicks,
				useTexture:true,
			},
			shaderProgramPie4RangesTicks:{
				vShader:shaders.vShader,
				fShader:shaders.fShaderPie4RangesTicks,
				useTexture:true,
			},
			shaderProgramPie4Ranges:{
				vShader:shaders.vShader,
				fShader:shaders.fShaderPie4Ranges,
				useTexture:true,
			},
			shaderProgramPieGradient:{
				vShader:shaders.vShader,
				fShader:shaders.fShaderPieGradient,
			},
		};		

		for( var s in shaders ){
			if( !compileShader( shaders[s] ) )
				return false;
		}

		for( var p in programs ){
			var prog = programs[p];
			if( !createProgram( prog ) )
				return false;
			if( !prog.uniforms )
				prog.uniforms = {};
			var u;
			for( u in prog.vShader.uniforms ){
				if( !prog.uniforms[u] )
					prog.uniforms[u] = copyObject(prog.vShader.uniforms[u]);
			}
			for( u in prog.fShader.uniforms ){
				if( !prog.uniforms[u] )
					prog.uniforms[u] = copyObject(prog.fShader.uniforms[u]);
			}
			for( u in prog.uniforms ){
				var uniform = prog.uniforms[u];
				uniform.location = gl.getUniformLocation(prog.glProgram, uniform.name );
			}
			var va;
			if( !prog.vertexAttributes && prog.vShader.vertexAttributes )
				prog.vertexAttributes = {};
			for( va in prog.vShader.vertexAttributes ){
				if( !prog.vertexAttributes[va] )
					prog.vertexAttributes[va] = copyObject(prog.vShader.vertexAttributes[va]);
			}
			prog.vertexAttribCount = 0;
			if( prog.vertexAttributes ){
				for( var a in prog.vertexAttributes ){
					var attr = prog.vertexAttributes[a];
					attr.location = gl.getAttribLocation(prog.glProgram, attr.name );
					if( attr.location>=prog.vertexAttribCount )
						prog.vertexAttribCount = attr.location+1;
				}
			}
		}
				
		function copyObject(o){
			var copy = {};
			for( var p in o )
				copy[p] = o[p];
			return copy;
		}

		function compileShader(s){
			var shader = gl.createShader(s.type);
			gl.shaderSource(shader, s.source);
			gl.compileShader(shader);
			if (!gl.getShaderParameter(shader, gl.COMPILE_STATUS)) {
				error( "Failed to compile shader" + gl.getShaderInfoLog(shader));
				return false;
			}
			s.compiled = shader;
			return true;
		}

		function createProgram( p ){
			var vertexShader = p.vShader.compiled,
				fragmentShader = p.fShader.compiled;
			var program = gl.createProgram();
			gl.attachShader(program, vertexShader);
			gl.attachShader(program, fragmentShader);
			gl.linkProgram(program);
			if (!gl.getProgramParameter(program, gl.LINK_STATUS)) {
				log("Could not initialise shaders");
				return false;
			}
			if( !p.vertexAttributes )
				p.vertexAttributes = {};
			p.vertexAttributes.vertexPositionAttribute = {name:"aVertexPosition"};
			//gl.enableVertexAttribArray(p.vertexAttributes.vertexPositionAttribute.location);
			//p.trMatrixUniform = gl.getUniformLocation(program, "uTransformMatrix");
			p.glProgram = program;
			return true;
		}
		
		return programs;
	}
};