/* global graphics:true, ALIGN, log, instruments, widgetsZoom, FontFace, DEFAULT_FONT */
/* exported TextManager */
/* jslint browser: true*/
/* jshint curly: false */

var TextManager = {
	init : function ( callback ){
		var numberTexture,//s = {},
			canvas = document.createElement('canvas'),
			fontName = DEFAULT_FONT,
			fontSettings = {
				exo:{
					vCenter: 0.45,
					scale: 1
				},
				firasans:{
					vCenter: 0.55,
					scale: 1.05
				},
				default:{
					vCenter: 0.5,
					scale: 1.025
				}
			},
			fontVCenter = fontSettings[fontName]?fontSettings[fontName].vCenter:fontSettings.default.vCenter,
			fontScale = fontSettings[fontName]?fontSettings[fontName].scale:fontSettings.default.scale,
			numberString = '0123456789-.',
			forcePow2NumberTexture = true,
			maxFontSize = getMaxFontSize( instruments );
		canvas.width = 256;
		canvas.height = 32;
		
		function getMaxFontSize( instr ){
			var max = 0;
			for( var i in instr ){
				if( instr[i].fontSize && instr[i].fontSize>max )
					max = instr[i].fontSize;
				if( instr[i].instruments ){
					var mfs = getMaxFontSize( instr[i].instruments );
					if( mfs>max )
						max = mfs;
				}
			}
			return max;
		}
		this.onContextLost = function(){
			numberTexture = null;
		};
		function onResize(){
			//for( var n in numberTextures )
			if( numberTexture ){
				//log('TextManager: Deleting number image');
				graphics.deleteTexture(numberTexture/*s[n]*/);
				numberTexture = undefined;
			}
		}
		this.getNumberTexture = function(){
			return numberTexture;
		};
		this.getPrecisionValue = function( num, precision ){
			if( precision<0 )
				return num;
			return (num*(precision?(Math.pow(10, precision?precision:0)|0):1));
		};
		this.printNumber = function( num, x,y, textSize, precision, alignFlags, textColor, slideDigits, slideClip, disableBlend ){
			if( !numberTexture ){
				log('Create number image @size='+ maxFontSize + ', zoom='+widgetsZoom );
				numberTexture = drawStringToCanvas( numberString, Math.round(maxFontSize*2*100*widgetsZoom), true, forcePow2NumberTexture );
			}
			
			textSize *= fontScale;
			textSize = Math.round(textSize);//make it same size as the numbers, which are rounded for better centering inside the canvas
			var texture, cw,ch;
			texture = numberTexture;
			cw = texture.charWidth;
			ch = texture.stringHeight;
			var scale = textSize/texture.stringHeight;
			
			var slide = 0, sdg,
				slideHeight;
			
			num = this.getPrecisionValue(num,precision);//(num*(precision?(Math.pow(10, precision?precision:0)|0):1))|0;
			
			if( slideDigits ){
				var pow = Math.pow(10, slideDigits-1 );
				slideHeight = textSize*1.1;//texture.stringHeight*0.5;
				slide = slideHeight*(num%pow)/pow;
				sdg = slideDigits;
			}
			
			num = slideDigits? (num|0) : Math.round(num);
			
			var numDigits = (num===0)?1:Math.floor(Math.log10(Math.abs(num))) + 1;
			if( precision ){
				if( precision>0 ){
					if( numDigits<precision+1 )
						numDigits = precision+1;
				}
				else
					numDigits = -precision;
			}
			var minus = num<0,
				numWidth = (numDigits-(precision?-0.75:0)+(minus?1:0))*cw*scale;//account for smaller decimal char
			
			x+= numWidth;//start printing from right to left
			x-=cw*scale;
			x += alignHorizontal( texture, numWidth, alignFlags );
			y += alignVertical( texture, ch*scale, alignFlags );
			
			graphics.activateTextureTint();
			graphics.setTextureTintColor( textColor );
			
			var stopShift = false;
			for( var i=0; i<numDigits; i++ ){
				if( i==precision && precision!==0 ){
					//decimal point - not monospaced here
					var ptw = cw*0.75;
					graphics.drawImage( texture, cw*10.88,0, ptw,texture.stringHeight, x,y, ptw*scale,texture.stringHeight*scale, undefined, disableBlend );
					x-=(ptw*scale)|0;
				}
				var digit = (i>=0)?(Math.abs(num)%10):10;//number or minus sign(#10)	
				if( slideDigits ){
					if( i<slideDigits ){
						var s = slide,
							sdigit = digit,
							prevDigit, nextDigit,
							shiftSlide = i<sdg && !stopShift && slide>0.5*slideHeight;
							
						if( shiftSlide )
							s -= slideHeight;
						if( i<sdg-1 )
							digit = sdigit = nextDigit = prevDigit = 0;
						else{
							if( shiftSlide )
								sdigit = digit===9?0:sdigit+1;
							prevDigit = sdigit===0?9:sdigit-1;
							nextDigit = sdigit===9?0:sdigit+1;
						}
						this.drawDigit(nextDigit,s-slideHeight,x,y,texture,cw,scale,disableBlend);
						this.drawDigit(sdigit,s,x,y,texture,cw,scale,disableBlend);
						if( i<sdg )
							this.drawDigit(prevDigit,s+slideHeight,x,y,texture,cw,scale,disableBlend);
						if( i>=sdg-1 && digit===9 ){
							slideDigits++;
							stopShift=true;
							if( slideClip ){
								graphics.setClipScreenBox(slideClip);
								slideClip = null;
							}
						}
					}
					else{
						slide = 0;
						this.drawDigit(digit,0,x,y,texture,cw,scale,disableBlend);
					}
					
				}
				else
					this.drawDigit(digit,0,x,y,texture,cw,scale,disableBlend);
				
				x-=cw*scale;
				num = (num/10)|0;
			}
			if( minus )
				graphics.drawImage( texture, cw*10,0, cw,texture.stringHeight, x,y, cw*scale,texture.stringHeight*scale, undefined, disableBlend );//minus sign
			
			/*function drawDigit(digit,slide){
				graphics.drawImage( texture, cw*digit, 0, cw,texture.stringHeight, x,y+slide, cw*scale,texture.stringHeight*scale );
			}*/
			graphics.deactivateTextureTint();
		};
		this.drawDigit = function(digit,slide,x,y,texture,cw,scale, disableBlend ){
			graphics.drawImage( texture, cw*digit, 0, cw,texture.stringHeight, x,y+slide, cw*scale,texture.stringHeight*scale, undefined, disableBlend );
		};
		/* If the text is an array - it will be reprinted only if dirtyArray is true*/
		this.printText = function( text, x,y, textSize, texture, textColor, alignFlags, dirtyArray, lineSpacing, disableBlend ){
			textSize *= fontScale;
			textSize = Math.round(textSize);//for better centering inside the canvas
			if( texture && (texture.textSize!=textSize || (!Array.isArray(text) && texture.text!=text) || dirtyArray ) ){
				//log('TextManager: Image size for text "' + text + '" has changed, deleting');
				graphics.deleteTexture(texture);
				texture = null;
			}
			if( !texture ){
				texture = drawStringToCanvas( text, textSize, false, true, lineSpacing );
				texture.textSize = textSize;
				texture.text = text;
				//log('TextManager: Create image for text "' + text + '", size:' + textSize +', dimentions:' + texture.width + 'x' + texture.height );
			}
			var multiline = Array.isArray(text);
			x += alignHorizontal( texture, texture.stringWidth, alignFlags );
			var h = multiline?texture.stringHeight*text.length/texture.lines:texture.stringHeight;//when printing multiline we have the option to print less lines that what has been rendered in the texture
			y += alignVertical( texture, h, alignFlags, multiline );
			graphics.activateTextureTint();
			graphics.setTextureTintColor( textColor );
			graphics.drawImage( texture, 0,0, texture.stringWidth, h, x,y, texture.stringWidth, h, null, disableBlend );
			graphics.deactivateTextureTint();
			return texture;
		};
		function alignVertical( texture, h, alignFlags, multiline ){
			if( multiline ){
				switch(alignFlags&ALIGN.MaskVertical){	
					case ALIGN.Top:
						return 0;
					case ALIGN.Middle:
						return -h*0.5;
					default://ALIGN.Bottom:
						return -h;
				}
			}
			else{
				switch(alignFlags&ALIGN.MaskVertical){	
					case ALIGN.Top:
						return -h*0.2;
					case ALIGN.Middle:
						return -h*0.5;
					default://ALIGN.Bottom:
						return -h*0.8;
				}
			}
		}
		function alignHorizontal( texture, w, alignFlags ){
			switch(alignFlags&ALIGN.MaskHorizontal){	
				case ALIGN.Left:
					return 0;
				case ALIGN.Center:
					return -w/2;
				default://ALIGN.Right:
					return -w;
			}
		}
		function drawStringToCanvas( text, textSize, monospace, pow2, lineSpacing ){
			var ctx = canvas.getContext('2d');
			ctx.save();
			
			ctx.textBaseline = 'middle';
			ctx.font =  textSize + 'px ' + fontName;
			
			var multiline = Array.isArray(text),
				w=0,h,cw;
			
			if( multiline ){
				if( !lineSpacing )
					lineSpacing = textSize;
				for( var i=0; i<text.length; i++ ){
					var lw;
					if( monospace )
						lw = text.length;
					else
						lw = ctx.measureText(text[i]).width;
					if( lw>w )
						w = lw;
				}
				if( monospace )
					w *= ctx.measureText('oooo').width/4;
				h = lineSpacing*text.length;
			}
			else{
				if( monospace )
					w = ctx.measureText('oooo').width*text.length/4;
				else
					w = ctx.measureText(text).width;
				h = textSize;
			}
			
			initCanvas( ctx, w, h, textSize, pow2 );
			
			//log( 'drawString (font=' + ctx.font + ') :' + text + ' <=> ' + textSize + 'px ' + fontName );
			
			/*ctx.fillStyle = "yellow";
			ctx.fillRect(0,0,canvas.width,canvas.height);
			ctx.fillStyle = "green";
			ctx.fillRect(0,0,w,h);
			ctx.fillStyle = "magenta";
			ctx.fillRect(0,0,w,h/2);
			ctx.fillStyle = "black";*/
			
			if( monospace ){
				//@todo - multiline - not needed for now
				cw = w/text.length;
				ctx.textAlign = 'center';//center each character
				for( var n=0; n<text.length; n++ )
					ctx.fillText( text[n], cw*(n+0.5), h*fontVCenter);//draw one by one to make sure the chars are exactly where we expect them
				ctx.textAlign = 'left';//as it was
			}
			else{
				if( multiline ){
					var y = textSize*fontVCenter;
					for( var l=0; l<text.length; l++,y+=lineSpacing )
						ctx.fillText( text[l], 0, y );
				}
				else
					ctx.fillText( text, 0, textSize*fontVCenter );
			}

			
			var texture = graphics.createTexture(canvas,pow2);//mipmap when pow2
				
			if( monospace )
				texture.charWidth = cw;
			if( multiline )
				texture.lines = text.length;
			texture.stringHeight = h;
			texture.stringWidth = w;
			texture.width = canvas.width;
			texture.height = canvas.height;
			
			ctx.restore();
			return texture;
		}
		function initCanvas( ctx, w,h, textSize, pow2 ){	
			//if( canvas.width<dim.width )
			if( pow2 ){
				w = Math.pow(2, Math.ceil(Math.log(w)/Math.log(2)));
				h = Math.pow(2, Math.ceil(Math.log(h)/Math.log(2)));
			}
			else{
				w = Math.ceil(w);
				h = Math.ceil(h);
			}
			if( canvas.width!=w )
				canvas.width=w;
			if( canvas.height!=h )
				canvas.height=h;
			//ctx.font = '12px Consolas,Monaco,Lucida Console,Liberation Mono,DejaVu Sans Mono,Bitstream Vera Sans Mono,Courier New, monospace';
			ctx.fillStyle = '#FFFFFF';
			ctx.font = textSize + 'px ' + fontName;
			ctx.textAlign = 'left';
			ctx.textBaseline = 'middle';
			ctx.clearRect(0,0,w,h);
			
			ctx.shadowColor = "rgba(0,0,0,0.5)";
			ctx.shadowBlur = textSize/10;
		}

		window.addEventListener('resize', onResize );
		//onResize();
		
		if( callback ){
			if(  (typeof(FontFace) !== 'undefined') ){
				log("====================== Found font loading API! ======================");
				//wait for the font to load
				var f = new FontFace("newfont", "url(fonts/" + fontName + ".ttf)", {});
				f.load().then(function () {
					log("====================== Success loading font! ======================");
					callback();
					onResize();
				});
			}
			else{
				//@todo - use polyfill
				//an ugly hack - wait 2 seconds
				log("====================== No font loading API! ======================");
				setTimeout( function(){
					callback();
					onResize();
				}, 2000);
			}
		}
	}
};