/* global controls, coordScale, getInstrumentValue, setInstrumentValue, netLockLocalValue, netExpireLocalValue */
/* exported getControlAt*/
/* jslint browser: true*/
/* jshint proto: true */
/* jshint curly: false */

var controlTypes = {};

controlTypes.base = {
	getAbsoluteX : function(){
		return this.x+this.parentInstrument.getAbsoluteX();
	},
	getAbsoluteY : function(){
		return this.y+this.parentInstrument.getAbsoluteY();
	},
	init:function(){
		if( coordScale!=1 )
			this.adjustCoordinates(coordScale);
	},
	adjustCoordinates: function( scale ){
		this.x *= scale;
		this.y *= scale;
		if( this.w )
			this.w *= scale;
		if( this.h )
			this.h *= scale;
	}
};

/*
 * Finds the first listened by the parent instrument property and modifies it when dragging.
 * Override onDrag to return the changed value
 * The changed value is sent over the net, and modified locally
 * The local modification stays for a limited time second, in that time the new value should be updated from the server   
 */ 
controlTypes.drag = {//sendDelta, deltaScale, intValue
	__proto__: controlTypes.base,	
	timeout:0, initialValue: 0,
	onMouseDown : function( /*x,y*/ ){
		var prop = this.getValueKey();
		this.initialValue = getInstrumentValue( this.parentInstrument, prop );
		if( this.sendDelta ){
			this.prevValue = this.initialValue;
			//this.sum = 0;
		}
		else
			setInstrumentValue( getInstrumentValue( this.parentInstrument, prop ), prop );
		netLockLocalValue( this.getValueKey() );
	},
	onMouseUp: function(){
		netExpireLocalValue( this.getValueKey() );
	},
	onMouseDrag : function( x,y, dx,dy ){
		var key = this.getValueKey(),
			min, max, repeat,
			inputRange = this.parentInstrument.getValueProperty("inputValueRange",key);
		//todo find min max
		if( inputRange ){
			min = Math.min(inputRange[0],inputRange[1]);
			max = Math.max(inputRange[0],inputRange[1]);
			repeat = this.parentInstrument.getValueProperty("repeat",key);   
		}
		else{
			min = -Math.PI;
			max = Math.PI;
			repeat = true;
		}
		var val = this.onDrag( x,y,dx,dy, this.initialValue, min, max );
		if( !repeat )
			val = Math.min( Math.max(val,min),max );
		if( !this.sendDelta ){
			if( this.intValue )
				val = Math.round(val);
			setInstrumentValue( val, key );
		}
		else{
			setInstrumentValue( val, key, true );
			var delta = val-this.prevValue;
			if( this.deltaScale )
				delta *= this.deltaScale;
			if( this.intValue ){
				delta = Math.round(delta);
				this.prevValue += delta/(this.deltaScale?this.deltaScale:1);
			}
			else
				this.prevValue = val;
			//this.sum+=delta;
			//console.log('sum='+this.sum);
			if( delta )
				//setInstrumentValue( 90, key+'_delta', false, true );
				setInstrumentValue( delta, key+'_delta', false, true );
		}
	},
	getValueKey: function(){//private
		//the knob value must be the first listened one in the parent instrument
		if( this.parentInstrument.listen instanceof Array )
			return this.parentInstrument.listen[0];
		else
			return this.parentInstrument.listen; 
	}
};

function getControlAt(x,y){
	var bestDist, bestControl;
	for( var i=0; i<controls.length; i++ ){
		var control = controls[i];
		var cx = control.getX?control.getX():control.x,
			cy = control.getY?control.getY():control.y;
		var parent = control.parentInstrument;
		while( parent ){
			cx += parent.x;
			cy += parent.y;
			parent = parent.parentInstrument;
		}
		if( cx-control.w/2<x && cx+control.w/2>x && cy-control.h/2<y && cy+control.h/2>y ){
			var dist = Math.sqrt( (cx-x)*(cx-x) + (cy-y)*(cy-y) );
			if( bestControl===undefined || dist<bestDist ){
				bestDist = dist;
				bestControl = control;
			}  
		}
	}
	return bestControl;
}