/* global graphics, coordScale, time, canvas, config, instrumentTypes:true, convergeValue, instruments, netUpdate, useOffscreenCanvas */
/* exported drawInstruments*/
/* jslint browser: true*/
/* jshint proto: true */
/* jshint curly: false */

function drawInstruments( time, connected, paused ){
	if( graphics.beginDraw() ){
		for( var i=0; i<instruments.length; i++ ){
			var ent = instruments[i];
			if( connected && (!paused || ent.doNotPause ) && time!==undefined )
				netUpdate(ent);
			if( !ent.isHidden || !ent.isHidden() )
				ent.draw();
		}
	}
}

instrumentTypes.static = {//[x,y]/[top/left],[xOffset],[yOffset]
	__proto__:instrumentTypes.base, 
	draw: function(){		
		this.setInitialTransform();
		if( this.top!==undefined )
			graphics.drawImage( this.image, this.getLeft()+(this.xOffset?this.xOffset:0), this.getTop()+(this.yOffset?this.yOffset:0) );//we have top/left coordinates
		else{
			try{
				graphics.drawImage( this.image, this.getX()-((this.image.width/2)|0)-(this.xOffset?this.xOffset:0), this.getY()-((this.image.height/2)|0)-(this.yOffset?this.yOffset:0) );
			}
			catch( err ){
				console.log( "drawImage error in " + this.image.src + " " + err );
				console.log( "x:"+ (this.getX()-((this.image.width/2)|0)-(this.xOffset?this.xOffset:0)),
							 "y:"+ (this.getY()-((this.image.height/2)|0)-(this.yOffset?this.yOffset:0))
						   );
			}
		}
	},
	getTop: function(){return this.top;},
	getLeft: function(){return this.left;}
};

instrumentTypes.rotate = {//x,y,[rotateCenterX],[rotateCenterY],[inputValueRange],[outputValueRange],[repeat],[nocycle]the x/y positioning is relative to the center of rotation, if no rotateCenterX/Y is specified - the center of rotation is the image center
	__proto__:instrumentTypes.base,
	getRotate: function(){
		var rot = this.getDisplayedValue( this.getValue()/*valueObject.value*/, this.getValueProperty('repeat')?360:undefined )*2*Math.PI/360;
		if( this.invert != (this.control && this.control.invert) )
			return -rot;
		else
			return rot;
	},
	draw: function(){
		this.setInitialTransform();
		var xCenter = this.image.width/2,
			yCenter = this.image.height/2,
			xFract = xCenter-(xCenter|0),//the toInt offset when centering if this was a static instrument, do this to align static and rotated instruments 
			yFract = yCenter-(yCenter|0);
		graphics.translate( this.getX()+xFract, this.getY()+yFract );
		graphics.rotate( this.getRotate() );
		this.drawImage( this.image, (this.rotateCenterX!==undefined)?-this.rotateCenterX:-xCenter,
											(this.rotateCenterY!==undefined)?-this.rotateCenterY:-yCenter );
	},
	drawImage: function( image,x,y ){//override if for instance you need another transform after rotate
		graphics.drawImage( image, x,y ); 
	},
	init:function(){
		var valueObject;
		if( this.defaultValue )
			valueObject = this[this.defaultValue];
		else
			valueObject = this;
		if( valueObject.inputValueRange ){                                                           
			var offset, scale, delta = valueObject.inputValueRange[1]-valueObject.inputValueRange[0];
			if( valueObject.repeat ){
				//the inputValueRange[0] and inputValueRange[1] are the same value - a full turn 	
				scale = 360/delta;
				offset = -valueObject.inputValueRange[0]*valueObject.valueScale/2;
			}
			else{                                          
				//inputValueRange[0/1] correspond to the min/max input values
				//outputValueRange[0/1] corresponds to the min/max angles
				if( valueObject.outputValueRange ){
					scale = (valueObject.outputValueRange[1]-valueObject.outputValueRange[0])/delta;
					offset = -valueObject.inputValueRange[0]*scale+valueObject.outputValueRange[0];
				}
				else{
					offset = 0;
					scale = 1;
				}
			}
			if( valueObject.valueOffset )
				valueObject.valueOffset += offset;
			else
				valueObject.valueOffset = offset;
				
			if( valueObject.valueScale )
				valueObject.valueScale *= scale;
			else
				valueObject.valueScale = scale;	
		}
		else if( valueObject.repeat ){
			valueObject.inputValueRange = [ 0, 2*Math.PI ];//the net extrapolation code needs to know
		}
		instrumentTypes.base.init.call(this);			
	}
};

instrumentTypes.rotateRadians = {
	__proto__:instrumentTypes.rotate,
	repeat:true,
	inputValueRange:[0,2*Math.PI],
	outputValueRange:[0,360]
};

instrumentTypes.slide = {//x,y,[horizontal], [inputValueRange, outputValueRange]
	__proto__:instrumentTypes.static,
	getSlide: function(){
		return this.getDisplayedValue(this.getValue())*coordScale;
	},
	getX: function(){
		return this.x + (this.horizontal? this.getSlide() : 0);
	},
	getY: function(){
		return this.y + (this.horizontal?0 : this.getSlide());
	},
	init:function(){
		if( this.inputValueRange && this.outputValueRange ){                                                           
			var delta = this.inputValueRange[1]-this.inputValueRange[0];                                        
			//inputValueRange[0/1] correspond to the min/max input values
			//outputValueRange[0/1] corresponds to the min/max slide in pixels
			this.valueScale = (this.outputValueRange[1]-this.outputValueRange[0])/delta;
			this.valueOffset = -this.inputValueRange[0]*this.valueScale+this.outputValueRange[0] + (this.valueOffset ? this.valueOffset : 0 ); 		
		}		               
		instrumentTypes.base.init.call(this);			
	}
};


instrumentTypes.slideTexture = {//x,y,peekSize,[horizontal],[textureNotCentered],[repeat],inputValueRange, edgeRange(the range starts from the edges, not the dial center)
	__proto__:instrumentTypes.base,
	getSlide: function(){
		//var fullRange = this.repeat || this.edgeRange;//this.getValueProperty("repeat") || this.getValueProperty("repeat");
		return this.getDisplayedValue(this.getValue()+(this.repeat?(this.peekSize/2):0), this.repeat? ((this.horizontal?this.image.width:this.image.height) ) :undefined ); 
		//return this.getDisplayedValue(this.getValue(), this.getValueProperty("repeat")? ((this.horizontal?this.image.width:this.image.height)+(this.repeat?(this.peekSize/2):0) ) :undefined );
		//var size = (this.horizontal?this.image.width:this.image.height); 						
		//return this.getDisplayedValue( this.getValue(), this.getValueProperty("repeat")?( size+(this.repeat?(this.peekSize/2):0) ) :undefined );
	},
	draw: function(){
		this.setInitialTransform();
		if( this.horizontal )
			this.drawImage(this.image,
					-this.getSlide(), 0,//clipping coords x,y
					this.getX()-this.peekSize/2, this.getY()-this.image.height/2,//x,y coord of the image
					this.peekSize, this.image.height );//w/h
		else
			this.drawImage(this.image,
					0,-this.getSlide(),//clipping coords x,y
					this.getX()-this.image.width/2, this.getY()-this.peekSize/2, //x,y coord of the image
					this.image.width, this.peekSize );//w/h
	},
	drawImage: function( image, clipX, clipY, x,y, w,h ){
		if( !this.repeat )
			this.clipDrawImage( image, clipX, clipY, x,y, w,h );
		else{
			if( !config.useWebGL )
				graphics.drawTiledImage( image, clipX, clipY, x,y, w,h );
			else{
				//This draws only 2 tiles: current and prev or next, draws the tiles only in the slider direction
				if( this.horizontal )
					clipX = clipX%image.width;
				else
					clipY = clipY%image.height;
				this.clipDrawImage( image, clipX, clipY, x,y, w,h );
				if( this.horizontal ){
					if( clipX<0 )
						this.clipDrawImage( image, clipX+image.width, clipY, x,y, w,h );
					else if( clipX+w>=image.width )
						this.clipDrawImage( image, clipX-image.width, clipY, x,y, w,h );
				}
				else{
					if( clipY<0 )
						this.clipDrawImage( image, clipX, clipY+image.height, x,y, w,h );
					if( clipY+h>=image.height )
						this.clipDrawImage( image, clipX, clipY-image.height, x,y, w,h );
				}
			}
		}
	},
	//a workaround for webkit bug 116991 , 66574 : https://bugs.webkit.org/show_bug.cgi?id=116991 :
	//when the drawRect is not completely filled by the image - it draws nothing
	//do my own clipping - change the position and size of the drawn rectangle	
	//@todo - only for tiled, for non-tiled - keep the image inside the rect
	clipDrawImage: function( image, clipX, clipY, x,y, w,h ){
	    //clip x & width
		if( clipX<0 ){
			x-=clipX;
			w+=clipX;		
			clipX=0;
		}
		if( clipX+w>=image.width )
			w-= clipX+w-image.width;
			
		if( w<0 )
			return;
			
		//clip y & height
		if( clipY<0 ){
			y-=clipY;
			h+=clipY;		
			clipY=0;
		}
		if( clipY+h>=image.height )
			h-= clipY+h-image.height;
			
		if( h<0 )
			return;
			                                                                                                                                                                          
		graphics.drawImage( image, clipX, clipY, w, h, x,y, w,h );
	},
	init:function(){			
		if( this.inputValueRange ){
			var lenPx = this.horizontal? this.image.width : this.image.height,
				delta = this.inputValueRange[1]-this.inputValueRange[0],
				offset, scale,
				psize = this.peekSize*coordScale;
			if( this.repeat ){
				//the inputValueRange[0] and inputValueRange[1] are the same value - a full turn 	
				scale = lenPx/delta;
				offset = -lenPx/2 -(this.inputValueRange[0]+this.inputValueRange[1])*scale/2;
			}
			else if( this.edgeRange ){
				scale = -(lenPx-psize)/delta;
				offset = -this.inputValueRange[0]*scale;							
			}
			else{						
				scale = -(lenPx-psize)/delta;	
				if( this.textureNotCentered )
					offset = this.inputValueRange[0]*scale;
				else
					offset = -lenPx/2 -(this.inputValueRange[1]+this.inputValueRange[0])*scale/2 + psize/2;
			}
			if( this.valueOffset )
				this.valueOffset += offset;
			else
				this.valueOffset = offset;
				
			if( this.valueScale )
				this.valueScale *= scale;
			else
				this.valueScale = scale; 		
		}
		instrumentTypes.base.init.call(this);			
	},
	adjustCoordinates: function( scale, atInit ){
		if( this.edgeRange )
			this.edgeRange = this.edgeRange;
				
		this.peekSize *= scale; 

		if( !atInit ){		
			if( this.valueOffset )
				this.valueOffset *= scale;
			if( this.valueScale )
				this.valueScale *= scale;
		}									
		instrumentTypes.base.adjustCoordinates.call( this, scale, atInit );
	}  	
}; 
                                                    
instrumentTypes.rollCounter = {//height, numDigits,[base=10],[leading],[fract],[interpolateSpeed] if we have multiple src images (imageFrames) - uses the second image for the fract digit
	__proto__:instrumentTypes.base,
	draw: function(){
		var base = this.base?this.base:10,
			leading = this.leading? this.leading : this.image.width,
			interpolateSpeed = (this.interpolateSpeed? this.interpolateSpeed : defaultIntepolateSpeed)/coordScale,
			w = this.image.width,
			h = this.height,
			x = (this.getX() + (leading*this.numDigits/2) )|0,
			y = (this.getY()-h/2)|0,
			value = this.getDisplayedValue(this.value);
		this.setInitialTransform();
		var image = this.imageFrames? this.imageFrames[1] : this.image;
		if( !config.useWebGL )
			graphics.activateTiledImage( image );
		value = this.fract ? ((value*10)|0) : (value|0);//this.fract ? value*10 : (value|0);  	 
		for( var i=0; i<this.numDigits; i++ )
		{
			var roll = this.image.height*(value%base)/base;
			if( true )//i!=0 || !this.fract )
			{
				if( !this.rollValues )
					this.rollValues = new Array(this.numDigits);
				if( !this.rollValues[i] )
					this.rollValues[i] = {};
				else if( !config.dontConvergeValues )
					roll = convergeValue( roll, time, this.rollValues[i].prevValue, this.rollValues[i].timestamp, /*this.*/interpolateSpeed, true, this.image.height );
				this.rollValues[i].prevValue = roll;
				this.rollValues[i].timestamp = time;
			}   
			value = (value/base)|0;
			x -= leading;
		
			if( !config.useWebGL )
				graphics.drawActiveTiledImage( 0, roll, x, y, w,h );
			else{
				roll = roll%image.height;
				graphics.drawImage( image, 0, roll, w, h, x, y, w, h );
				if( roll<0 )
					graphics.drawImage( image, 0, roll+image.height, w, h, x, y, w, h );
				else
					graphics.drawImage( image, 0, roll-image.height, w, h, x, y, w, h );
			}
			if( i===0 && this.imageFrames ){
				if( !config.useWebGL )
					graphics.activateTiledImage( this.image );
				image = this.image;
			}
		}
		if( !config.useWebGL )
			graphics.deactivateTiledImage();
	},
	adjustCoordinates: function( scale, atInit ){
		this.height *= scale;
		this.leading *= scale;
		instrumentTypes.base.adjustCoordinates.call( this, scale, atInit);
	}  
};

instrumentTypes.switchImages = {// src must be an array
	__proto__:instrumentTypes.static,
	interpolateSpeed: 0,
	getFrame: function(){
		return this.getDisplayedValue( this.value ) % this.imageFrames.length;
	},
	draw: function(){
		this.image = this.imageFrames[this.getFrame()];
		instrumentTypes.static.draw.call(this);
	}
};

/*
 * A text instrument, renders a string from a texture with the necessary symbols
 * Supports multiline text, you may specify leading and line spacing, a mapping(glyphRemap) of
 * individual symbols or a grouo of symbols(digits, big/small letters from alphabet) to an index 
 * in the texure so that it both works with normal strings and the texture contains only what's needed. 
 * If necesary - you can specify kerning/offset for individual letters(specialKerning.leading/offset),
 * for instance the symbols '1', 'I' and ':' can be narrower    
 */   
instrumentTypes.text = {//glyphCount,[glyphRemap],[specialKerning],[leading],[lineSpacing]
	__proto__:instrumentTypes.base,
	draw: function(){
		var text = this.getDisplayedValue().toString(),
			   h = (this.image.height/this.glyphCount)|0,
			   xpos=0;
		this.setInitialTransform();
		for( var i=0; i<text.length; i++ ){
			var index,
				char = text.charAt(i),
				charCode = char.charCodeAt();
			if( char=='\n' ){
				graphics.translate( -xpos, this.lineSpacing?this.lineSpacing:h );//start of the new line
				xpos = 0;
				continue;
			} 
			if( this.glyphRemap ){
				if( this.glyphRemap.hasOwnProperty(char) )
					index = this.glyphRemap[char]; 
				else if( char>'0' && char<='9' && this.glyphRemap.hasOwnProperty('0') )
					index = this.glyphRemap['0']+(charCode-'0'.charCodeAt());
				else if( char>'a' && char<='z' && this.glyphRemap.hasOwnProperty('a') )
					index = this.glyphRemap['a']+(charCode-'a'.charCodeAt());
				else if( char>'A' && char<='Z' && this.glyphRemap.hasOwnProperty('A') )
					index = this.glyphRemap['A']+(charCode-'A'.charCodeAt());
				else
					index = charCode-'0'.charCodeAt();
			}
			else
				index = charCode-'0'.charCodeAt();
			var sk = (this.specialKerning && this.specialKerning[char])? this.specialKerning[char] : null;
			if( index>=0 && index<this.glyphCount )
				graphics.drawImage( this.image,
						0,h*index,//clipping coords x,y
						this.image.width, h,//clipping w/h
						(this.getX()-this.image.width/2 + (sk ? sk.offset : 0))|0, (this.getY()-h/2)|0,//x,y coord of the image, @todo correct center
						this.image.width, h );//w/h
			if( i!=text.length-1 && text[i+1]!='\n' ){
				var tx = sk ? sk.leading : (this.image.width + (this.leading?this.leading:0) );
				xpos+=tx; 
				graphics.translate( tx, 0 );
			}  
		}
	},
	adjustCoordinates: function( scale, atInit ){
		if( this.leading )
			this.leading *= scale;
		if( this.lineSpacing )
			this.lineSpacing *= scale;
		if( this.specialKerning ){
			for(var prop in this.specialKerning){
				if( this.specialKerning[prop].leading )
					this.specialKerning[prop].leading *= scale;
				if( this.specialKerning[prop].offset )
					this.specialKerning[prop].offset *= scale;					
			}
		}
		instrumentTypes.base.adjustCoordinates.call( this, scale, atInit);
	} 
};

/* Used for testing purposes */
instrumentTypes.area = {//w,h,[x,y]/[top/left],[xOffset],[yOffset]
	__proto__:instrumentTypes.base, 
	draw: function(){
		//do nothing
	},
	adjustCoordinates: function( scale, atInit ){
		this.w *= scale;
		this.h *= scale;
		if( this.xOffset )
			this.xOffset *= scale;
		if( this.yOffset )
			this.yOffset *= scale;
		instrumentTypes.base.adjustCoordinates.call( this, scale, atInit );
	}
};


instrumentTypes.rect = {//w,h,[x,y]/[top/left],[xOffset],[yOffset]
	__proto__:instrumentTypes.area, 
	draw: function(){
		graphics.drawRect(this.x-this.w/2, this.y-this.h/2, this.w, this.h,"yellow");
	}
};

instrumentTypes.canvas = {
	__proto__:instrumentTypes.base,
	invalidated:true,
	init:function(){
		this.pInstr = this.parentInstrument; 
	},
	draw: function(){
		if( useOffscreenCanvas && this.parentInstrument ){
			delete this.parentInstrument;
			this.invalidated = true;
		}
		if( !useOffscreenCanvas && !this.parentInstrument ){
			this.parentInstrument = this.pInstr;
			delete this.canvas;				
		}
		if( useOffscreenCanvas ){
			//@todo - custom canvas size????
			if( this.canvas===undefined )
				this.canvas = graphics.createOffscreenCanvas();
			if( this.canvas.width!=canvas.width ){
				this.canvas.width = canvas.width;
				this.invalidated = true; 
			}
			if( this.canvas.height!=canvas.height ){
				this.canvas.height = canvas.height;
				this.invalidated = true; 
			}
			if( this.invalidated )
				graphics.invalidateOffscreen(this.canvas);
		}
		
		if( this.invalidated || !useOffscreenCanvas ){		
			for( var i=0; i<this.instruments.length; i++ )
				if( !this.instruments[i].isHidden || !this.instruments[i].isHidden() )
					this.instruments[i].draw();
			this.invalidated = false;
		}
			
		if( useOffscreenCanvas ){
			if( this.invalidated )
				graphics.endOffscreenCanvas();
			
			graphics.drawOffscreenCanvas(this.canvas);
		}
	}
};