
/* global config, graphics, imagesFolder, Image, drawProgressIteration:true, window */
/* exported doDrawProgressBar, drawProgressBar */
/* jshint curly: false */

var lastProgress=0, drawProgressIteration = 0,
	progressSettings = {
		width : 0.8,//relative to the canvas width
		height : 0.015,//0.025,//relative to the canvas width
		top : 0.9,
		barColor: '#00b8e0',//commend this line to enable red to green coloring ftom 0% to 100%
		barBorderColor: '#aaa',
		barBorderWidth: 0,//0.5,//relative to the progress bar height, set to 0 to disable
		background1: '#fff',//'#bbb',
		background2: '#fff',//'#444',
		progressBackground1: '#005387',//'#666',
		progressBackground2: '#005387',//'#999',
		radialBackground: true,
		repeatRedraw: 3,//work around a cocoonjs bug with jerky progress bar, disable(set to 0) on iOS
		logo: "logo.png",
		logoWidth: 0.33,
		logoY: 0.35,
		radius: 0.3//0.5//relative to the progress bar height
};
	
function drawProgressBar( progress ){
	if( progress-lastProgress<0.033 )
		return;///do not redraw if value has not changed much
	lastProgress = progress;
	doDrawProgressBar( progress );
	
	if( progressSettings.repeatRedraw ){
		//work around on cocoonjs bug on android - it swaps the buffers even if no frame will be drawn resulting in jerky animation
		if( !drawProgressIteration ){
			window.requestAnimFrame( repeatDraw );
		}
		else
			drawProgressIteration = 0;//repeat again, we have a frame pending
	}
	
	function repeatDraw(){
		drawProgressIteration++;
		doDrawProgressBar( lastProgress );
		if( drawProgressIteration!=progressSettings.repeatRedraw )
			window.requestAnimFrame( repeatDraw );
		else
			drawProgressIteration = 0;
	}
}

/* Sandbox version: http://jsfiddle.net/yUpVL/7/ */
function doDrawProgressBar( progress ){
	if( config.useWebGL )
		return;//@todo - webgl version
	
	var canvas = graphics.getCanvas(),
		context = canvas.getContext('2d');
	if(context){
		var x = canvas.width*(1-progressSettings.width)/2,
			y = canvas.height*progressSettings.top,
			w = canvas.width*progressSettings.width,
			h = canvas.height*progressSettings.height,
            radius = progressSettings.radius*h,
			barBorderWidth = progressSettings.barBorderWidth*h,
			gr;
        
        context.save();
		context.setTransform(1, 0, 0, 1, 0, 0);
        
		if( progressSettings.background1==progressSettings.background2 )
			context.fillStyle = progressSettings.background1;
		else{
			if( progressSettings.radialBackground ){
				gr = context.createRadialGradient(canvas.width/2,canvas.height*4/6,0,canvas.width/2,0,canvas.height);
				gr.addColorStop(0,progressSettings.background1);
				gr.addColorStop(1,progressSettings.background2); 
			}
			else{
				gr = context.createLinearGradient(0,canvas.height,0,0);
				gr.addColorStop(0,progressSettings.background1);
				gr.addColorStop(1,progressSettings.background2); 
			}
			context.fillStyle = gr;
		}
		context.fillRect( 0, 0, canvas.width, canvas.height );
		
		var logo = progressSettings.logoImage;
		if( !logo ){
			logo = progressSettings.logoImage = new Image();
			logo.src = imagesFolder + '/' + progressSettings.logo;
		}
		
		if( logo.complete ){
			if( !progressSettings.logoLoaded ){
				progressSettings.logoLoaded = true;
				drawProgressIteration = 0;
			}
			var lw = canvas.width*progressSettings.logoWidth,
				lh = lw*logo.height/logo.width,
				ly = canvas.height*progressSettings.logoY;
			context.drawImage( logo, (canvas.width-lw)/2, ly, lw, lh );
		}
        
		if( barBorderWidth ){
			context.fillStyle = progressSettings.barBorderColor;
			var borderRadius = radius+barBorderWidth/2;
			roundRect( context, x-barBorderWidth,y-barBorderWidth,w+2*barBorderWidth,h+2*barBorderWidth, borderRadius, true, true ).fill();
		}
		    
		if( progressSettings.progressBackground1==progressSettings.progressBackground2 ){
			context.fillStyle = progressSettings.progressBackground1;
		}
		else{
			gr=context.createLinearGradient(0,y,0,y+h);
			gr.addColorStop(0,progressSettings.progressBackground1);
			gr.addColorStop(1,progressSettings.progressBackground2);
			context.fillStyle = gr;
		}
		roundRect( context, x,y,w,h, radius, true, true ).fill();
		
		var scale, r = 1-progress,	g = progress, r1, g1;
		if( progressSettings.barColor )
			context.fillStyle = progressSettings.barColor;
		else{
			if( config.useGradients ){        
				gr=context.createLinearGradient(0,y,0,y+h);

				//due to a cocoonjs bug with interpreting HSL values the below code does not work:
				//gr.addColorStop(0,'hsl('+100*progress+',100%,'+(50-20*progress)*1.5+'%)');
				//gr.addColorStop(1,'hsl('+100*progress+',100%,'+(50-20*progress)*0.5+'%)');
				//so use this instead:
				scale = 1/(Math.sqrt(r*r+g*g));
				r1=Math.min(r*scale*1.5*255,255);
				g1=Math.min(g*scale*0.75*255,255);
				var r2=Math.round(r1*0.5).toString(16),
					g2=Math.round(g1*0.5).toString(16);
				r1 = Math.round(r1).toString(16);
				g1 = Math.round(g1).toString(16);
				if( r1.length<2 ) r1 = '0'+r1;
				if( g1.length<2 ) g1 = '0'+g1;
				if( r2.length<2 ) r2 = '0'+r2;
				if( g2.length<2 ) g2 = '0'+g2;
				//log( progress + " " + '#'+r1+g1+'00' + '#'+r2+g2+'00' )
				gr.addColorStop(0,'#'+r1+g1+'00');
				gr.addColorStop(1,'#'+r2+g2+'00');
				context.fillStyle = gr; 		
			}
			else{
				scale = 1/(Math.sqrt(r*r+g*g));
				r1=Math.min(r*scale*1.5*255,255);
				g1=Math.min(g*scale*0.75*255,255);
				r1 = Math.round(r1).toString(16);
				g1 = Math.round(g1).toString(16);
				if( r1.length<2 ) r1 = '0'+r1;
				if( g1.length<2 ) g1 = '0'+g1;
				context.fillStyle = '#'+r1+g1+'00';
			}
		}
		roundRect( context, x,y,w*progress,h, radius, true, true ).fill();
	
        context.restore();
	}
}   

//CanvasRenderingContext2D.prototype.roundRect = function( x, y, w, h, r, left, right ) {
function roundRect( context, x, y, w, h, r, left, right ) {
	if( w < 2*r )
		r = w / 2;
	if( h < 2*r )
		r = h / 2;
	var left_r = left? r:0,
		right_r = right? r:0;
	context.beginPath();
	context.moveTo( x+left_r, y );
	context.arcTo( x+w, y, x+w, y+h, right_r );
	context.arcTo( x+w, y+h, x, y+h, right_r );
	context.arcTo( x, y+h, x, y, left_r );
	context.arcTo( x, y, x+w, y, left_r );
	context.closePath();
	return context;
}