/* global graphics, config, Cocoon, ImageLoader, log, viewLimits, postponeResize:true, hiRez:true, zoomIn:true, drawXOrigin:true, drawYOrigin:true, showZoomIcon:true, widgetsZoom:true, coordScale:true, assert, log, draw, hiRezPostfix, loRezPostfix, imagePostfix:true, clearTop:true, instruments, connected */
/* exported resizeCanvas, transformInputCoordinates, drawXOrigin, drawYOrigin, widgetsZoom */
/* jslint browser: true*/
/* jshint curly: false */

/*
 * Set the optimal size of the canvas, try to keep the pixel aspect ratio to fixed values ( 1:2 1:1 2:1 )
 */

var zoomIn = (localStorage.getItem("zoomIn")!==null ) ? ( localStorage.zoomIn=="true" ) : null,//false,
	showZoomIcon = false,
	addBlackToCanvasTop = typeof Cocoon!=="undefined",//use black pixels in the canvas top
	coordScale = hiRez?viewLimits.hiRezCoordScale:1,
	widgetsZoom = 1,
	drawXOrigin=0, drawYOrigin=0,
	transformInputCoordinates = true,
	postponeResize = false,
	currentWidth, currentHeight;


function resizeCanvas(event) {
	if (ImageLoader.isLoadingImages()) {
		log("postponing resize");
		postponeResize = true;
		return;
	}

	var epsilon = 0.001;

	//window size
	var windowWidth = window.innerWidth*devicePixelRatio,
		windowHeight = window.innerHeight*devicePixelRatio;

	log("------------Calculate viewport size------------");
	log("window size: ("+window.innerWidth+"x"+window.innerHeight+")*"+devicePixelRatio);

	//if( windowWidth==currentWidth && windowHeight==currentHeight )
	//	return;

	currentWidth = windowWidth;
	currentHeight = windowHeight;

	//we have a frame defined by the inner and outer sizes of viewLimits.min/max width/height
	//see if the view fits inside that frame when zoomed out 1, 2, 3, 4 times
	var canvasZoom = 0.5,
		zoomLimit = config.useHiRez?8:5;
	//try to zoom to fit inside the frame, if there are black pixels on top - we don't care if the view is higher than the frame		
	while (canvasZoom < zoomLimit && !(windowWidth <= viewLimits.maxWidth * canvasZoom && (windowHeight <= viewLimits.maxHeight * canvasZoom || viewLimits.blackPixelsOnTop) && windowWidth >= viewLimits.minWidth * canvasZoom && windowHeight >= viewLimits.minHeight * canvasZoom)){
		if( canvasZoom<1 )
			canvasZoom = 1;
		else if( canvasZoom<2 )
			canvasZoom = 2;
		else{
			if( config.useHiRez )
				canvasZoom += 2;
			else
				canvasZoom += 1;
		}
	}

	var viewWidth, viewHeight, fitWidth, fitHeight;
	if (canvasZoom < zoomLimit) {
		//the view fits within the viewLimits frame with 1:canvasZoom scale	
		viewWidth = windowWidth / canvasZoom;
		viewHeight = Math.min(windowHeight / canvasZoom, viewLimits.maxHeight);
		fitWidth = false,
		fitHeight = false;
		showZoomIcon = false;

		assert( ( viewWidth >= viewLimits.minWidth && viewWidth <= viewLimits.maxWidth  ) ||
				( viewHeight >= viewLimits.minHeight && viewHeight <= viewLimits.maxHeight  ),
				"view does not fit in" );
		log( 'Canvas zoom is 1:' + canvasZoom );
	}
	else{
		//we can't acheive 1:0.5, 1:1, 1:2, etc mapping from image to screen pixels, choose the best-aspect size and rescale
		log('Can not use exact pixel mapping, using best-aspect size');
		var viewSize = calculateViewSize( windowWidth / windowHeight, zoomIn );
		viewWidth = viewSize.viewWidth,
		viewHeight = viewSize.viewHeight,
		fitWidth = viewSize.fitWidth,
		fitHeight = viewSize.fitHeight;
		showZoomIcon = viewSize.showZoomIcon;
		if( showZoomIcon )
			localStorage.zoomIn = zoomIn = viewSize.zoomIn;
		else
			delete localStorage.zoomIn;
		canvasZoom = viewWidth/windowWidth;

		assert( fitWidth || fitHeight, "neighter witdth or height are fit to window" );
	}

	var canvas = graphics.getCanvas();
	var canvasWidth, canvasHeight;
	if (fitWidth) {
		widgetsZoom = windowWidth / viewWidth;
		canvasHeight = viewHeight * windowWidth / viewWidth;
		canvasWidth = windowWidth;
	} else if (fitHeight) {
		widgetsZoom = windowHeight / viewHeight;
		canvasWidth = viewWidth * windowHeight / viewHeight;
		canvasHeight = windowHeight;
	} else {
		widgetsZoom = canvasZoom;
		canvasWidth = canvasZoom * viewWidth;
		canvasHeight = canvasZoom * viewHeight;
	}

	var reloadScale;
	if ( config.useHiRez && !hiRez && widgetsZoom > 1 + epsilon) { //viewLimits.hiRezCoordScale-epsilon ){
		//switch from low to high resolution textures
		hiRez = true;
		imagePostfix = hiRezPostfix;
		coordScale = viewLimits.hiRezCoordScale;
		//widgetsZoom /= viewLimits.hiRezCoordScale;
		log("switch to hi-Rez textures");
		reloadScale = viewLimits.hiRezCoordScale;
	} else if (hiRez && widgetsZoom <= 1 / viewLimits.hiRezCoordScale + epsilon) {
		//switch from high to low resolution textures
		hiRez = false;
		imagePostfix = loRezPostfix;
		coordScale = 1;
		//widgetsZoom *= viewLimits.hiRezCoordScale;
		log("switch to lo-Rez textures");
		reloadScale = 1 / viewLimits.hiRezCoordScale;
	}

	if( hiRez )
		widgetsZoom /= 2;

	var stretchCanvas = 1;//TODO: get rid of
	/*if (widgetsZoom > 1) {
		//will stretch the whole canvas, not individual widgets
		stretchCanvas = widgetsZoom;
		canvasWidth *= coordScale/widgetsZoom;
		canvasHeight *= coordScale/widgetsZoom;
		widgetsZoom = 1;//devicePixelRatio;-found-out emperically
		log( "Will expand canvas on screen: " + Math.round( canvasWidth ) +"x" + Math.round( canvasHeight ) + ", widgets zoom is now " + devicePixelRatio );	
	}*/

	log("Canvas size: " + Math.round(canvasWidth) + "x" + Math.round(canvasHeight) + (canvasZoom != 1 ? (", zoomed " + canvasZoom) : ""));
	log("View size: " + Math.round(viewWidth) + "x" + Math.round(viewHeight));

	if (!Cocoon) {
		var style = "";
		if (Math.round(canvasWidth * stretchCanvas) < windowWidth)
			style = "left:" + Math.round((windowWidth - canvasWidth * stretchCanvas) / 2) + "px;"; //center horizontally
		if (stretchCanvas != 1)
			style += fitHeight ? "height:100%" : "width:100%";
		canvas.setAttribute("style", style);
	}

	//now that we know the canvas dimensions - find an optimal centering/trimming of the picture
	drawXOrigin = (viewWidth - viewLimits.normalWidth) / 2;
	if (addBlackToCanvasTop && canvasHeight * stretchCanvas < windowHeight) { //canvasHeight>canvas.height+2 ){
		//align the canvas to the bottom
		//the canvas must extend over the whole screen because it can't be aligned to bottom in cocoonjs
		drawYOrigin = (windowHeight/ stretchCanvas  - canvasHeight /*+ viewLimits.addTop*/) / (widgetsZoom*coordScale) ;
		log("canvas black pixels: ", drawYOrigin);
		clearTop = drawYOrigin + 2; //windowHeight-canvasHeight+2;
		canvasHeight = windowHeight / stretchCanvas;
	} else {
		clearTop = 0;
		//the canvas is aligned to the the bottom and may have black non-canvas space above it
		if (viewHeight >= viewLimits.normalHeight)
			drawYOrigin = viewLimits.addTop * (viewHeight - viewLimits.normalHeight) / (viewLimits.maxHeight - viewLimits.normalHeight); //canas.height==viewLimits.normalHeight => drawYOrigin=0, canvas.height==viewLimits.maxHeight => drawYOrigin=viewLimits.addTop
		else
			drawYOrigin = viewLimits.trimTop * (viewHeight - viewLimits.normalHeight) / (viewLimits.minHeight - viewLimits.normalHeight); //canas.height==viewLimits.normalHeight => drawYOrigin=0, canvas.height==viewLimits.maxHeight => drawYOrigin=viewLimits.trimTop
	}
	drawXOrigin = Math.round(drawXOrigin * widgetsZoom * coordScale);
	drawYOrigin = Math.round(drawYOrigin * widgetsZoom * coordScale);

	canvas.width = Math.round(canvasWidth);
	canvas.height = Math.round(canvasHeight);
	canvas.setAttribute("style", 'height:'+window.innerHeight+'px;');

	log("y origin:" + drawYOrigin + ", x origin:" + drawXOrigin);

	if (!ImageLoader.loadedImages())
		ImageLoader.loadImages(instruments);
	else if (reloadScale) {
		log("Reload images, scale:" + reloadScale);
		ImageLoader.reloadImages(reloadScale);
	}

	log("------------------------");

	if (!config.constantDrawLoop && event !== undefined && !connected)
		draw(); //when not connected - the canvas is not constantly redrawn, so draw it to avoid black screen*/
}

function calculateViewSize( aspect, doZoomIn )
{
	var desiredAspect = viewLimits.maxWidth / viewLimits.maxHeight, // 1280/960 = 1024/768 = 1.333333333
		maxAspect = viewLimits.maxWidth / viewLimits.minHeight,
		minAspect = viewLimits.minWidth / viewLimits.maxHeight,
		desiredZoomOutAspect = viewLimits.maxWidth/viewLimits.maxHeight,// 1280/960 = 1024/768 = 1.333333333
		desiredZoomInAspect = viewLimits.minWidth/viewLimits.minHeight,
		fitWidth, fitHeight;

	if( aspect > maxAspect ){
		//too wide
		//fit minimal height, black borders on left/right
		viewHeight = viewLimits.minHeight;
		viewWidth = viewLimits.maxWidth;
		fitHeight = true;
		logtext = 'Too wide, black pixels on left/right, fit at: minHeight';
	}
	else if( aspect < minAspect ){
		//too high
		//fit minimal width, black area on top
		viewWidth = viewLimits.minWidth;
		viewHeight = viewLimits.maxHeight;
		fitWidth = true;
		logtext = 'Too high, black pixels on top, fit at: minWidth';
	}
	else{
		//the aspect is ok, try to resize as close to normaWidth/normalHeight as possible
		if (aspect > desiredAspect) {
			//wider than desired
			viewHeight = viewLimits.normalHeight;
			viewWidth = aspect * viewHeight;
			if (viewWidth <= viewLimits.maxWidth) {
				//it fits at viewLimits.normalHeight
				fitHeight = true;
				logtext = 'Fit at: normalHeight';
			} else {
				//still too wide, shrink it to maxWidth
				viewWidth = viewLimits.maxWidth;
				viewHeight = viewWidth / aspect;
				fitWidth = true;
				logtext = 'Fit at: maxWidth';
			}
		} else {
			//higher than desired 
			viewWidth = viewLimits.normalWidth;
			viewHeight = viewWidth / aspect;
			if (viewHeight <= viewLimits.maxHeight) {
				//it fits at normalWidth
				fitWidth = true;
				logtext = 'Fit at: normalWidth' + (viewHeight > viewLimits.maxHeight ? ", some black pixels on top" : "");
			} else {
				//still too high, shrink to maxHeight
				viewHeight = viewLimits.maxHeight;
				viewWidth = viewHeight * aspect;
				fitWidth = true;
				logtext = 'Fit at: maxHeight + some black pixels on top';
			}
		}
	}

	assert( fitWidth || fitHeight, "Eighter width or height must fit");

	//see if zoom-in or out is feasible
	var showZoom,
		zoomInSize = getZoomInSize(aspect, desiredZoomInAspect),
		zoomInRatio = zoomInSize.fitWidth ? viewWidth / zoomInSize.viewWidth : viewHeight / zoomInSize.viewHeight,
		newSize;
	if (zoomInRatio > viewLimits.zoomInThreshold) {
		//always zoom-in if feasible
		log("Zoom-in is possible, " + (zoomInSize.fitWidth ? "fit minWidth" : "fit minHeight"));
		if (doZoomIn === null)
			doZoomIn = false; //default to the scenario with 1:1 pixels, the one that's not zoomed-in
		if (doZoomIn)
			newSize = zoomInSize;
		showZoom = true;
	} else {
		//zoom-in not needed, see if zoom-out is feasible
		var zoomOutSize = getZoomOutSize(aspect, desiredZoomOutAspect),
			zoomOutRatio = zoomOutSize.fitWidth ? zoomOutSize.viewWidth / viewWidth : zoomOutSize.viewHeight / viewHeight;
		if (zoomOutRatio > viewLimits.zoomOutThreshold) {
			log("Zoom-out is possible, " + (zoomInSize.fitWidth ? "fit normalWidth" : ", fit normalHeight"));
			if (doZoomIn === null)
				doZoomIn = true; //default to the scenario with 1:1 pixels, the one that's not zoomed-out
			if (!doZoomIn)
				newSize = zoomOutSize;
			showZoom = true;
		} else{
			showZoom = false;
			log("Zoom in/out is not needed");
		}
		if (newSize)
			log((newSize == zoomOutSize) ? "Zooming out" : "Zooming in");
	}

	if (newSize) {
		viewWidth = newSize.viewWidth;
		viewHeight = newSize.viewHeight;
		fitWidth = newSize.fitWidth;
		fitHeight = newSize.fitHeight;
	} else
		log(logtext);

	return {
		viewWidth : viewWidth,
		viewHeight : viewHeight,
		fitWidth : fitWidth,
		fitHeight : fitHeight,
		zoomIn: doZoomIn,
		showZoomIcon: showZoom
	}
}

function getZoomOutSize( aspect, desiredZoomAspect ){//, maxZoomAspect ){
	var size = {};
	if( aspect < desiredZoomAspect ){
		//higher than desired, fit normal width
		log("Zoom-out size: higher than desired, fit normal width");
		size.viewWidth = viewLimits.normalWidth;
		//size.viewHeight = Math.min( size.viewWidth/aspect, viewLimits.maxHeight );
		size.viewHeight = size.viewWidth/aspect;
		if( size.viewHeight > viewLimits.maxHeight ){
			//higher than allowed
			size.viewWidth *= viewLimits.maxHeight/size.viewHeight;//decrease w and h proportionaly to fit in maxHeight
			size.viewHeight = viewLimits.maxHeight;
			if( size.viewWidth < viewLimits.minWidth )
				size.viewWidth = viewLimits.minWidth
		}
		size.fitWidth = true;
		size.fitHeight = false;
	}
	else{
		//wider than desired, fit normal height
		log("Zoom-out size: wider than desired, fit normal height");
		size.viewHeight = viewLimits.normalHeight;
		//size.viewWidth = Math.min( aspect*size.viewHeight, viewLimits.maxWidth );
		size.viewWidth = aspect*size.viewHeight;
		if( size.viewWidth > viewLimits.maxWidth ){
			//wider than allowed
			size.viewHeight *= viewLimits.maxWidth/size.viewWidth;//decrease w and h proportionaly to fit in maxHeight
			size.viewWidth = viewLimits.maxWidth;
			if( size.viewHeight < viewLimits.minHeight )
				size.viewHeight = viewLimits.minHeight
		}
		size.fitHeight = true;
		size.fitWidth = false;
	}
	return size;
}

function getZoomInSize( aspect, desiredZoomAspect ){
	var size = {};                                          
	if( aspect < desiredZoomAspect ){
		//higher than desired, fit min width
		log("Zoom-in size: higher than desired, fit min width");
		size.viewWidth = viewLimits.minWidth;
		//size.viewHeight = Math.min( size.viewWidth/aspect, viewLimits.maxHeight );
		size.viewHeight = size.viewWidth/aspect;
		if( size.viewHeight > viewLimits.maxHeight ){
			//higher than allowed
			size.viewWidth *= viewLimits.maxHeight/size.viewHeight;//decrease w and h proportionaly to fit in maxHeight
			size.viewHeight = viewLimits.maxHeight;
			if( size.viewWidth < viewLimits.minWidth )
				size.viewWidth = viewLimits.minWidth
		}
		//@todo? thinner than allowed?
		size.fitWidth = true;
		size.fitHeight = false;
	}
	else{
		//wider than desired, fit min height
		log("Zoom-in size: wider than desired, fit min height");
		size.viewHeight = viewLimits.minHeight;
		//size.viewWidth = Math.min( aspect*size.viewHeight, viewLimits.maxWidth );
		size.viewWidth = aspect*size.viewHeight;
		if( size.viewWidth > viewLimits.maxWidth ){
			//wider than allowed
			size.viewHeight *= viewLimits.maxWidth/size.viewWidth;//decrease w and h proportionaly to fit in maxHeight
			size.viewHeight = viewLimits.maxWidth;
			if( size.viewHeight < viewLimits.minHeight )
				size.viewHeight = viewLimits.minHeight
		}
		//@todo? shorter than allowed?
		size.fitHeight = true;
		size.fitWidth = false;
	}
	return size;
}		